#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <string>
 
#include "ImageTag.h"
#include "DescribeImage.h"

#ifndef WIN32
#	include <unistd.h>
#	include <sys/stat.h>
#	include <sys/types.h>
#
#	define stricmp strcasecmp
#	define strnicmp strncasecmp
#else
#	include <io.h>
#	include <direct.h>
#
#	define pipe(X) _pipe(X, 0, O_BINARY)
#	define R_OK 4
#endif

#if MagickLibVersion >= 0x662
#	include <magick/methods.h>
#endif

#define STAT_IMAGE(X) if (access((const char*)X, R_OK) != 0) { lasso_setResultMessage(token, "Invalid image filepath."); return osErrInvalidPathname; }


static const char* kImage		= "image";
static const char* kPrivateData = "_private_";
static const char* kImageSem	= "blobber";

static const char* kAddComment	= "addcomment";
static const char* kAnnotate	= "annotate";
static const char* kBlur		= "blur";
static const char* kBytes		= "bytes";
static const char* kComposite	= "composite";
static const char* kCommand		= "command";
static const char* kComments	= "comments";
static const char* kConvert		= "convert";
static const char* kCount		= "count";
static const char* kContrast	= "contrast";
static const char* kCrop		= "crop";
static const char* kData		= "data";
static const char* kDepthTag	= "depth";
static const char* kDescribe	= "describe";
static const char* kEnhance		= "enhance";
static const char* kExecute		= "execute";
static const char* kFile		= "file";
static const char* kFlipV		= "flipv";
static const char* kFlipH		= "fliph";
static const char* kFormat		= "format";
static const char* kHeightTag	= "height";
static const char* kHistogram	= "histogram";
static const char* kLength		= "length";
static const char* kModulate	= "modulate";
static const char* kPixel		= "pixel";
static const char* kResolutionV = "resolutionv";
static const char* kResolutionH = "resolutionh";
static const char* kRotate		= "rotate";
static const char* kSave		= "save";
static const char* kScale		= "scale";
static const char* kSharpen		= "sharpen";
static const char* kWidthTag	= "width";
static const char* kSetColorSpace	= "setcolorspace";
static const char* kColorSpace	= "colorspace";
static const char* kDebugTag	= "debug";

static const char* kAliased		= "-aliased";
static const char* kAmount		= "-amount";
static const char* kAngle		= "-angle";
static const char* kBase64		= "-base64";
static const char* kBgColor		= "-bgcolor";
static const char* kBinary		= "-binary";
static const char* kColor		= "-color";
static const char* kDepth		= "-depth";
static const char* kFont		= "-font";
static const char* kGaussian	= "-gaussian";
static const char* kHeight		= "-height";
static const char* kHex			= "-hex";
static const char* kInfo		= "-info";
static const char* kLeft		= "-left";
static const char* kMaxArea		= "-maxarea";
static const char* kOp			= "-op";
static const char* kOpacity		= "-opacity";
static const char* kPosition	= "-position";
static const char* kQuality		= "-quality";
static const char* kRadius		= "-radius";
static const char* kSample		= "-sample";
static const char* kShort		= "-short";
static const char* kSigma		= "-sigma";
static const char* kSize		= "-size";
static const char* kThumbnail	= "-thumbnail";
static const char* kThreshold	= "-threshold";
static const char* kTop			= "-top";
static const char* kV			= "-v";
static const char* kWidth		= "-width";
static const char* kPrivatePrm	= "-private";
static const char* kRGB	= "-rgb";
static const char* kCMYK	= "-cmyk";
static const char* kGRAY	= "-gray";

static const double kDefaultTextSize = 14.0;
static const char * kDefaultFont =

#ifdef WIN32
	"Arial";
#elif MACHINE_MACOSX
	"/Library/Fonts/Chalkboard.ttf" // "/usr/X11R6/lib/X11/fonts/luxisr.ttf"
#else
	"System"
#endif
;

static const char* kCompositeOps[] =
{
	"Undefined",
	"Over",
	"In",
	"Out",
	"Atop",
	"Xor",
	"Plus",
	"Minus",
	"Add",
	"Subtract",
	"Difference",
	"Multiply",
	"Bumpmap",
	"Copy",
	"CopyRed",
	"CopyGreen",
	"CopyBlue",
	"CopyOpacity",
	"Clear",
	"Dissolve",
	"Displace",
	"Modulate",
	"Threshold",
	"None",
	"Darken",
	"Lighten",
	"Hue",
	"Saturate",
	"Colorize",
	"Luminize",
	"Screen",
	"Overlay",
	"CopyCyan",
	"CopyMagenta",
	"CopyYellow",
	"CopyBlack",
	NULL
};
  
static const char * channelParams[4] = { 
	"-red", 
	"-green", 
	"-blue", 
	"-alpha" 
};

static const ChannelType channels[4] = { 
	RedChannel, 
	GreenChannel, 
	BlueChannel, 
	OpacityChannel 
};
	
bool hasX11 = true;

enum filters
{
	filterNone,
	filterBlur,
	filterGaussianBlur,
	filterMotionBlur,
	filterSharpen,
	filterUnsharpMask
};

enum commands
{
	cmdMogrify = 1,
	cmdMontage,
	cmdComposite,
	cmdConvert,
	cmdNone
};

		
static char * my_strmove(char * s1, const char * s2)
{
	return (char *) memmove(s1, s2, strlen(s2) + 1);
}

static const char * severityToString(ExceptionType type)
{
	switch (type)
	{
		case UndefinedException: 		return "UndefinedException";
		//case WarningException: 		return "WarningException";
		case ResourceLimitWarning: 		return "ResourceLimitWarning";
		case TypeWarning: 				return "TypeWarning";
		case OptionWarning: 			return "OptionWarning";
		case DelegateWarning: 			return "DelegateWarning";
		case MissingDelegateWarning: 	return "MissingDelegateWarning";
		case CorruptImageWarning: 		return "CorruptImageWarning";
		case FileOpenWarning: 			return "FileOpenWarning";
		case BlobWarning: 				return "BlobWarning";
		case StreamWarning: 			return "StreamWarning";
		case CacheWarning: 				return "CacheWarning";
		case CoderWarning: 				return "CoderWarning";
		case ModuleWarning: 			return "ModuleWarning";
		case DrawWarning: 				return "DrawWarning";
		case ImageWarning: 				return "ImageWarning";
		case XServerWarning: 			return "XServerWarning";
		case MonitorWarning: 			return "MonitorWarning";
		case RegistryWarning: 			return "RegistryWarning";
		case ConfigureWarning: 			return "ConfigureWarning";
		//case ErrorException: 			return "ErrorException";
		case ResourceLimitError: 		return "ResourceLimitError";
		case TypeError: 				return "TypeError";
		case OptionError: 				return "OptionError";
		case DelegateError: 			return "DelegateError";
		case MissingDelegateError: 		return "MissingDelegateError";
		case CorruptImageError: 		return "CorruptImageError";
		case FileOpenError: 			return "FileOpenError";
		case BlobError: 				return "BlobError";
		case StreamError: 				return "StreamError";
		case CacheError: 				return "CacheError";
		case CoderError: 				return "CoderError";
		case ModuleError: 				return "ModuleError";
		case DrawError: 				return "DrawError";
		case ImageError: 				return "ImageError";
		case XServerError: 				return "XServerError";
		case MonitorError: 				return "MonitorError";
		case RegistryError: 			return "RegistryError";
		case ConfigureError: 			return "ConfigureError";
		//case FatalErrorException: 	return "FatalErrorException";
		case ResourceLimitFatalError: 	return "ResourceLimitFatalError";
		case TypeFatalError: 			return "TypeFatalError";
		case OptionFatalError: 			return "OptionFatalError";
		case DelegateFatalError: 		return "DelegateFatalError";
		case MissingDelegateFatalError: return "MissingDelegateFatalError";
		case CorruptImageFatalError: 	return "CorruptImageFatalError";
		case FileOpenFatalError: 		return "FileOpenFatalError";
		case BlobFatalError: 			return "BlobFatalError";
		case StreamFatalError: 			return "StreamFatalError";
		case CacheFatalError: 			return "CacheFatalError";
		case CoderFatalError: 			return "CoderFatalError";
		case ModuleFatalError: 			return "ModuleFatalError";
		case DrawFatalError: 			return "DrawFatalError";
		case ImageFatalError: 			return "ImageFatalError";
		case XServerFatalError: 		return "XServerFatalError";
		case MonitorFatalError: 		return "MonitorFatalError";
		case RegistryFatalError: 		return "RegistryFatalError";
		case ConfigureFatalError: 		return "ConfigureFatalError";
	}
	return NULL;
}


osError
updateErrorInfo(lasso_request_t token, ExceptionInfo * exception)
{
	if (exception != NULL)
	{
		char buf[1024];
		const char * str = severityToString(exception->severity);
		sprintf(buf, "ImageMagick %s (%d): %s. %s", 
			str ? str : "error", 
			exception->severity,
			exception->reason ? exception->reason : "(Reason unknown)", 
			exception->description ? exception->description : "");
		lasso_setResultCode( token, (osError)exception->severity );
		lasso_setResultMessage( token, buf );
		return (osError)exception->severity;
	}
	
	lasso_setResultMessage( token, "ImageMagick: Unknown Exception." );
	return (osError)-1;

}

void
im_dtor( lasso_type_t self )
{
	im_type_ptr data = NULL;
	lasso_getPtrMember (NULL, self, kPrivateData, (void**)&data);
	
	if ( data != NULL && --data->refcount < 1)
	{
		if (data->image)	DestroyImageList(data->image);
		if (data->info )	DestroyImageInfo(data->info);
		free(data);
	}
}


osError 
im_callback( lasso_request_t token, tag_action_t action )
{
	lasso_type_t	lhs, rhs = NULL;
	im_type_ptr		data = NULL;
	
	lasso_getTagParam2 (token, 0, &rhs);	
	if ( rhs == NULL )
	{
		lasso_setResultMessage(token, "Could not retrieve param type instance.");
		return osErrInvalidMemoryObject;
	}

/*	
	auto_lasso_value_t tagName, typeName;
	lasso_getTagName (token, &tagName);
	lasso_typeGetName (token, rhs, &typeName);	
	printf ("image->%s: %s\n", tagName.name, typeName.name);
*/	
	if ( lasso_typeIsA2(token, rhs, kImage) == osErrNoErr )
	{
		lasso_getPtrMember(token, rhs, kPrivateData, (void**)&data);
		if ( data == NULL || data->image == NULL )
		{
			lasso_setResultMessage(token, "Could not retrieve custom type data.");
			return osErrInvalidMemoryObject;
		}
		
		if ( lasso_getTagSelf (token, &lhs) == osErrNoErr )
		{
			data->refcount++;
			lasso_setPtrMember(token, lhs, kPrivateData, data);
		}	
	}

	return osErrNoErr;
}

osError
ResolveFileName(lasso_request_t token, char ** argv, secure_operation_t perm)
{
	if (argv == NULL || *argv == NULL || strlen(*argv) == 0)
		return (osError)-1;
		
	osPathname path;
	lasso_fullyQualifyPath( token, *argv, path );
	lasso_resolvePath( token, path, path );

	//printf ("Resolved path: %s\n", path);

	if ( lasso_operationAllowed ( token, perm, path ) != osErrNoErr )
	{
		return osErrNoPermission;
	}
		
#if MACHINE_PC
	//lasso_getPlatformSpecificPath( path, path );
#else
	my_strmove(path, path+2); // hack for OSX LCAPI bug
#endif
	
	STAT_IMAGE(path);
	
	if ( strnicmp(path, *argv, strlen(*argv)) != 0 )
	{
		LiberateMemory((void **)argv);
		*argv = AllocateString(path);
	}
	return osErrNoErr;
}

osError 
im_init( lasso_request_t token, tag_action_t )
{
	auto_lasso_value_t	v;
	im_type_ptr			data = NULL;
	int					count = 0;
	osError				err = osErrNoErr;
	lasso_type_t		self = NULL, 
						param = NULL, 
						member = NULL;

	lasso_getTagParamCount (token, &count); 
	if (count < 1)
	{
		lasso_setResultMessage(token, "Image type initializer requires at least one parameter (filepath, image type, or image data)");
		return osErrInvalidMemoryObject;
	}
	
	Image *image = NULL;
	Image *list = NULL;
	ImageInfo *info = NULL;
	ChannelType channelType = UndefinedChannel;
	
	ExceptionInfo exception;
	GetExceptionInfo(&exception);
	
	bool combine = (count >= 3) && (lasso_tagParamIsDefined(token, channelParams[0]) == osErrNoErr);
	bool ping = (lasso_tagParamIsDefined(token, kInfo) == osErrNoErr);	

	for (int idx = 0; idx < 4; ++idx)
	{
		if ( combine )
		{
			if ( lasso_findTagParam2(token, channelParams[idx], &param) != osErrNoErr || param == NULL)
				continue;
			channelType = (ChannelType)(channelType | channels[idx]);
		}
		else
		{
			lasso_getTagParam2(token, idx, &param);
		}	
		
		if ( lasso_typeIsA2W( token, param, LPW(L"file") ) == osErrNoErr ) // file
		{
			static const UChar * kPathW = LPW(L"path");
			
			lasso_type_t tag = NULL;
			if ( lasso_typeGetMemberW(token, param, kPathW, &tag) == osErrNoErr )
				lasso_typeRunTagW(token, kPathW, tag, 0, NULL, &param, param);
		}
		
		if ( lasso_typeIsA( token, param, lpTypeString ) == osErrNoErr ) // filepath
		{
			lasso_typeGetString ( token, param, &v );
			if ( v.nameSize == 0 )
			{
				return osErrInvalidPathname;			
			}
		
			osPathname path;	
			lasso_fullyQualifyPath( token, v.name, path );
			lasso_resolvePath( token, path, path );
			if ( lasso_operationAllowed ( token, kReadFiles, path ) != osErrNoErr )
			{
				lasso_setResultMessage ( token, "Operation not allowed: read file." );
				return osErrNoPermission;
			}
	
#if MACHINE_PC
			//lasso_getPlatformSpecificPath( path, path );
#else
			my_strmove(path, path+2); // hack for OSX LCAPI bug
#endif	
			/*
			struct rlimit rlimit;
		 	if ( getrlimit(RLIMIT_STACK, &rlimit) == 0 )
				printf("RLIMIT_STACK: %d/%d\n", rlimit.rlim_cur / 1024, rlimit.rlim_max / 1024);
			*/
		 
			STAT_IMAGE(path);

			if (!info)
				info = CloneImageInfo ( (ImageInfo *)NULL );
			CopyMagickString( info->filename, path, MaxTextExtent);
			if (!combine)
				info->ping = (MagickBooleanType)ping;
			lasso_acquireSem(token, kImageSem);
			image = ReadImage ( info, &exception );
			lasso_releaseSem(token, kImageSem);
		}
		else if ( lasso_typeIsA( token, param, lpTypePair ) == osErrNoErr ) // data
		{	
			lasso_type_t p = NULL;
			lasso_pairGetFirst( token, param, &p );
			lasso_typeGetString( token, p, &v );
			
			if ( stricmp(v.name, kBinary) == 0 )
			{
				lasso_pairGetSecond( token, param, &p );
				lasso_typeGetString( token, p, &v );
				
				if (!info)
					info = CloneImageInfo ( (ImageInfo *)NULL );
				
				lasso_acquireSem(token, kImageSem); 
				image = BlobToImage( info, v.name, v.nameSize, &exception );
				lasso_releaseSem(token, kImageSem);
			}
			else if ( stricmp(v.name, kBase64) == 0 )
			{
				lasso_pairGetSecond( token, param, &p);
				lasso_typeGetString( token, p, &v );
				
				lasso_acquireSem(token, kImageSem);
				image = ReadInlineImage ( info, v.name, &exception );
				lasso_releaseSem(token, kImageSem);
			}
			else if ( stricmp(v.name, kPrivatePrm) == 0 )
			{
				osInt64 imgPtr = 0;
				lasso_pairGetSecond( token, param, &p );
				lasso_typeGetInteger( token, p, &imgPtr );
				if (imgPtr)
					image = (Image*)imgPtr;
			}
			else
			{
				lasso_setResultMessage(token, "Unknown parameter");
				err = osErrInvalidParameter;
				goto bail;		
			}
		}
		else if ( lasso_typeGetName(token, param, &v) == osErrNoErr && stricmp(v.name, kImage) == 0 )
		{	
			im_type_ptr img2 = NULL;	
			lasso_getPtrMember(token, param, kPrivateData, (void**)&img2);
			if ( img2 == NULL )
			{
				lasso_setResultMessage(token, "Could not obtain custom type data.");
				return osErrInvalidMemoryObject;		
			}
	
			image = ReferenceImage ( img2->image ); //increment refcount
			if ( info == (ImageInfo*)NULL )
				info = CloneImageInfo ( img2->info ); 
	
			if ( info == (ImageInfo*)NULL )
			{
				lasso_setResultMessage(token, "Could not create a copy of the image info");
				goto bail;
			}
		}
		
		if (image == (Image *)NULL)
		{	
			if ( exception.severity == FileOpenError )
			{
				lasso_setResultMessage(token, "Could not load image. Check the file pathname and permissions.");
				err = (osError)FileOpenError;
			}
			else
			{
				err = updateErrorInfo(token, &exception);
			}		
			goto bail;
		}
		
		if ( !combine )
		{
			list = NULL;
			break;
		}
		
		if (list == (Image *)NULL)
		{
			list = image;
			list->list = list; // head
		}
		else
		{
			image->previous = list;
			image->list = list->list;
			list->next = image;
			list = list->next;
		}
		image = NULL;
	}
	
	if (combine)
	{
		list = list->list;
		image = CombineImages(list->list, channelType, &exception);
		if (image == (Image *)NULL)
		{
			updateErrorInfo(token, &exception);
			goto bail;
		}
		DestroyImageList(list);
	}
	
	data = (im_type_t*)malloc(sizeof(im_type_t));
	if ( data == NULL )
	{
		lasso_setResultMessage(token, "Could not allocate memory for internal data");
		err = osErrOutOfMemory;
		goto bail;			
	}
	
	memset(data, 0, sizeof(im_type_t));
	data->image = image;
	data->info = info;
	
	err = lasso_typeAllocCustom( token, &self, kImage );
	if ( err != osErrNoErr || self == NULL )
	{
		lasso_setResultMessage(token, "Could not allocate new custom type instance.");
		goto bail;
	}
	
	if ( lasso_typeAllocTag(token, &member, im_transform) != osErrNoErr || member == NULL )
	{
		lasso_setResultMessage(token, "Could not add image transform member tags.");
		goto bail;
	}

	lasso_typeAddMember(token, self, kAnnotate,		member);	
	lasso_typeAddMember(token, self, kBlur,			member);
	lasso_typeAddMember(token, self, kCommand,		member);
	lasso_typeAddMember(token, self, kComposite,	member);
	lasso_typeAddMember(token, self, kContrast,		member);
	lasso_typeAddMember(token, self, kConvert,		member);
	lasso_typeAddMember(token, self, kCrop,			member);
	lasso_typeAddMember(token, self, kExecute,		member);
	lasso_typeAddMember(token, self, kEnhance,		member);
	lasso_typeAddMember(token, self, kFlipV,		member);
	lasso_typeAddMember(token, self, kFlipH,		member);
	lasso_typeAddMember(token, self, kModulate,		member);
	lasso_typeAddMember(token, self, kRotate,		member);
	lasso_typeAddMember(token, self, kSave,			member);
	lasso_typeAddMember(token, self, kScale,		member);
	lasso_typeAddMember(token, self, kSharpen,		member);
	lasso_typeAddMember(token, self, kSharpen,		member);
	
	if ( lasso_typeAllocTag(token, &member, im_info) != osErrNoErr )
	{
		lasso_setResultMessage(token, "Could not add image info member tags.");
		goto bail;
	}
	
	lasso_typeAddMember(token, self, kAddComment,	member);
	lasso_typeAddMember(token, self, kBytes,		member);
	lasso_typeAddMember(token, self, kComments,		member);
	lasso_typeAddMember(token, self, kCount,		member);
	lasso_typeAddMember(token, self, kData,			member);
	lasso_typeAddMember(token, self, kDepthTag,		member);
	lasso_typeAddMember(token, self, kDescribe,		member);
	lasso_typeAddMember(token, self, kFile,			member);
	lasso_typeAddMember(token, self, kFormat,		member);
	lasso_typeAddMember(token, self, kHeightTag,	member);
	lasso_typeAddMember(token, self, kLength,		member);
	lasso_typeAddMember(token, self, kPixel,		member);
	lasso_typeAddMember(token, self, kResolutionV,	member);
	lasso_typeAddMember(token, self, kResolutionH,	member);
	lasso_typeAddMember(token, self, kWidthTag,		member);
	lasso_typeAddMember(token, self, kSetColorSpace, member);
	lasso_typeAddMember(token, self, kColorSpace, member);
	lasso_typeAddMember(token, self, kDebugTag,		member);
	lasso_typeAddMember(token, self, kHistogram,	member);
	
	if ( lasso_typeAllocTag(token, &member, im_callback) != osErrNoErr )
	{
		lasso_setResultMessage(token, "Could not add image callback tags.");
		goto bail;
	}

	lasso_typeAddMember(token, self, kLPOnAssign,	member);
				
	err = lasso_setPtrMember(token, self, kPrivateData, data);
	if (err != osErrNoErr)
	{
		lasso_setResultMessage(token, "Error saving custom data pointer.");
		goto bail;
	}
	
	lasso_acquireSem(token, kImageSem);
	DestroyExceptionInfo(&exception);
	lasso_releaseSem(token, kImageSem);
	
	lasso_typeSetCustomDtor(self, im_dtor);
	lasso_returnTagValue(token, self);
	
	return osErrNoErr;
	
bail:
	lasso_acquireSem(token, kImageSem);
	DestroyExceptionInfo(&exception);
	
	if (image)	DestroyImageList(image);
	if (info)	DestroyImageInfo(info);		
	if (data)	free(data);
	if (combine)
		DestroyImageList(image);
	lasso_releaseSem(token, kImageSem);
	
	return err;
}


osError 
im_transform( lasso_request_t token, tag_action_t )
{
	lasso_type_t	self = NULL;
	im_type_ptr		data = NULL;
	
	lasso_getTagSelf (token, &self);	
	if ( self == NULL )
	{
		lasso_setResultMessage(token, "Could not retrieve custom type instance.");
		return osErrInvalidMemoryObject;
	}
	
	lasso_getPtrMember(token, self, kPrivateData, (void**)&data);
	if ( data == NULL || data->image == NULL )
	{
		lasso_setResultMessage(token, "Could not retrieve custom type data.");
		return osErrInvalidMemoryObject;
	}
		
	auto_lasso_value_t tag, param;
	int count = 0;
		
	lasso_getTagName(token, &tag);
	lasso_getTagParam(token, 0, &param);
	lasso_getTagParamCount(token, &count);
		
	if ( stricmp(tag.name, kCommand) == 0 || stricmp(tag.name, kExecute) == 0 )
	{
		if (count == 0)
		{
			lasso_setResultMessage(token, "Command member tag takes single string parameter (command line arguments).");
			return (osError)-1;		
		}

		// rtrim
		char * ch = (char*)(param.name + param.nameSize);
		while ( isspace(*--ch) ) *ch = '\0';
		// ltrim
		char * c = (char*)param.name;
		while ( isspace(*c) ) c++;
		
		int cmd = 0;
		if ( strnicmp(c, "mogrify", 7) == 0 )
			cmd = cmdMogrify;
		else if ( strnicmp(c, "montage", 7) == 0 )
			cmd = cmdMontage;
		//else if ( strnicmp(c, "convert", 7) == 0 )
		//	cmd = cmdConvert;
		else if ( strnicmp(c, "composite", 9) == 0 )
			cmd = cmdComposite;
	
		if ( cmd == 0 )
		{
			lasso_setResultMessage(token, "Execute tag supports the following commands: mogrify, montage and composite.");
			return (osError)-1;
		}
		
		int argc = 0;
		char ** argv = NULL;
		char * destination = tmpnam(NULL);
				
		if ( cmd != cmdMogrify )
		{
			char temp[1024];
			if ( cmd == cmdConvert )
			{
				if ( strlen(data->image->filename) == 0 )
				{
					tmpnam( data->image->filename );
					if ( WriteImage( data->info, data->image ) != 0 )
					{
						lasso_setResultMessage(token, "Could not save temporary image data.");
						return updateErrorInfo(token, &data->image->exception);				
					}
				}
				sprintf (temp, "convert %s %s", data->image->filename, c + 8);
			}
			else
			{
				sprintf (temp, "%s //%s.miff", param.name, destination);
			}
			argv = StringToArgv ( temp, &argc );
		}
		else
		{									
			argv = StringToArgv ( param.name, &argc );
		}
		
		if (argv == NULL)
		{
			lasso_setResultMessage(token, "Could not convert command string to command line arguments.");
			return osErrOutOfMemory;		
		}

		unsigned int status = ExpandFilenames(&argc, &argv);
		if (status == 0)
		{
			MagickFatalError(ResourceLimitFatalError,"MemoryAllocationFailed", (char *)NULL);
			return (osError)-1;
		}
					
		register int i,j;
		char * option;
		int filecount = 0;
		osError err = osErrNoErr;
				
		for (i = 2; i < argc; i++)
		{
			option=argv[i];
			j = i+1;

			if (((cmd==cmdMogrify || cmd == cmdConvert) && (LocaleCompare("random-threshold", option+1) == 0 || LocaleCompare("ordered-dither", option+1) == 0) )
			|| (  cmd==cmdConvert && (LocaleCompare("limit", option+1) == 0 || LocaleCompare("ordered-dither", option+1) == 0) )
			|| (  cmd==cmdComposite && (LocaleCompare("limit", option+1) == 0 || LocaleCompare("ordered-dither", option+1) == 0) ) )
			{
				i++;
			}
			
			if (strlen(option) == 1 || ((*option != '-') && (*option != '+')))
			{
				filecount++;
				if ( cmd == cmdMogrify )
				{
					// file name - remove it
					LiberateMemory((void **) &option);
					for ( ; j <= argc; j++ )
						argv[j-1] = argv[j];
					argc--;
				}
				else
				{
					err = ResolveFileName( token, &argv[i], kReadFiles );
					if ( err == osErrNoPermission )
					{
						lasso_setResultMessage ( token, "Operation not allowed: read file." );
						return err;
					}
				}		
			}
			else if (cmd == cmdMogrify || cmd == cmdConvert)
			{		
				i++;
			}
		}

		if (cmd == cmdComposite && (filecount < 3 || filecount > 4))
		{
			lasso_setResultMessage(token, "Composite command accepts minimum 2 and maximum 3 file arguments.");
		}		
		else if (argc < 2)
		{
			lasso_setResultMessage(token, "Execute tag command-line parameters should contain at least two arguments.");
		}
		else
		{
			ExceptionInfo exception;
			GetExceptionInfo(&exception);
			
			if (cmd == cmdMogrify /*|| cmd == cmdConvert*/)
			{
				status = MogrifyImageCommand(data->info, argc-1, &argv[1], (char **)NULL, &exception);
				if (status == 1 || exception.severity == OptionError) // hack; previous call will always throw error 410
				{
#if MagickLibVersion > 0x623
#	if MagickLibVersion == 0x660 || MagickLibVersion >= 0x662 || MagickLibVersion == 0x654 // ???
					status = MogrifyImages(data->info, MagickFalse, argc-1, (const char**)&argv[1], &data->image, &exception);
#	else
					status = MogrifyImages(data->info, argc-1, (const char**)&argv[1], &data->image, &exception);
#	endif
#else
					status = MogrifyImages(data->info, argc-1, (char**)&argv[1], &data->image);
#endif
				}
			}
			else
			{
				Image *image = NULL;
				ImageInfo * info = CloneImageInfo ( (ImageInfo *)NULL );
				if (cmd == cmdConvert)
				{
					//printf ("Convert destination: %s\n", argv[argc-1]);
					strcpy( info->filename, argv[argc-1] );
				}
				else
				{
					strcpy( info->filename, destination );
					strcpy( info->filename + strlen(destination), ".miff" );
				}
				
				//printf ("Load img: %s\n", info->filename);				
				if (cmd == cmdConvert)
				{
					status = ConvertImageCommand(data->info, argc-1, &argv[1], (char **)NULL, &exception);
				}
				else if (cmd == cmdMontage)
				{
					status = MontageImageCommand(info, argc-1, &argv[1], (char **)NULL, &exception);
				}
				else if (cmd == cmdComposite)
				{
					status = CompositeImageCommand(info, argc-1, &argv[1], (char **)NULL, &exception);				
				}

				if (status)
				{
					image = ReadImage ( info, &exception );
					remove(info->filename);
				}
				
				if (image != NULL)
				{			
					DestroyImageList(data->image);
					data->image = image;
				}
				//printf ("Status: %d\n", status);			
				DestroyImageInfo(info);
			}
			
			if (exception.severity != UndefinedException && (cmd != cmdMogrify || exception.severity != OptionError) )
			{
				err = updateErrorInfo(token, &exception);
			}
			else if (data->image->exception.severity != UndefinedException)
			{
				err = updateErrorInfo(token, &data->image->exception);
			}
			DestroyExceptionInfo(&exception);
		}
				
		for (int i = 0; i < argc; i++)
			LiberateMemory( (void **) &argv[i] );
		LiberateMemory( (void **) &argv );
		
		return err;
	}
	else if ( stricmp(tag.name, kSave) == 0 )
	{
		osPathname path;	
		lasso_fullyQualifyPath( token, param.name, path );
		lasso_resolvePath( token, path, path );
				
		if ( lasso_operationAllowed( token, kCreateFiles, path ) != osErrNoErr )
		{
			lasso_setResultMessage( token, "Operation not allowed: create file." );
			return osErrNoPermission;
		}

#if MACHINE_PC
		//lasso_getPlatformSpecificPath( path, path );
#else
		strcpy(path, path+2); // hack for OSX LCAPI bug
#endif

		strcpy(data->image->filename, path);

		FILE * file = NULL;				
		if ( data->info->file )
		{
			file = data->info->file;
			data->info->file = NULL;
		}	

		if (count > 1)
		{
			if ( lasso_findTagParam(token, kQuality, &param) == osErrNoErr )
			{
				int q = strtol( param.data, NULL, 10 );
				if (q)
					data->info->quality = ((q - 1) % 100) + 1;
			}
			if ( lasso_findTagParam(token, "-dpi", &param) == osErrNoErr)
			{
				double dpi = strtod( param.data, NULL );
				if (dpi)
   					data->info->pointsize = dpi;
			}
		}

		char oldMagick[32];
		strncpy(oldMagick, data->image->magick, 31);
		if ( WriteImage( data->info, data->image ) == MagickFalse )
		{
			return updateErrorInfo(token, &data->image->exception);
		}
		strncpy(data->image->magick, oldMagick, 31);

#if MACHINE_UNIX
		chmod(path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
#endif	
				
		if (file)
			data->info->file = file;
  
		return osErrNoErr;
	}	
	else if ( stricmp(tag.name, kComposite) == 0 )
	{
		lasso_type_t img = NULL;
		lasso_getTagParam2( token, 0, &img );
		
		if ( img == NULL )
		{
			lasso_setResultMessage(token, "First tag parameter must be Image type or variable.");
			return osErrInvalidMemoryObject;
		}
		
		im_type_ptr data2 = NULL;	
		lasso_getPtrMember( token, img, kPrivateData, (void**)&data2 );
		if ( data2 == NULL )
		{
			lasso_setResultMessage(token, "Could not retrieve second image data.");
			return osErrInvalidMemoryObject;
		}
	
		auto_lasso_value_t	v;		
		CompositeOperator 
			compose = OverCompositeOp;
		Image 
			*image2 = NULL; 
		long 
			x_offset = 0, 
			y_offset = 0;
		
		if ( lasso_findTagParam(token, kLeft, &v) == osErrNoErr )
			x_offset = strtol(v.data, NULL, 10);
		if ( lasso_findTagParam(token, kTop, &v) == osErrNoErr )
			y_offset = strtol(v.data, NULL, 10);
			
		if ( lasso_findTagParam(token, kOp, &v) == osErrNoErr )
		{
#if MagickLibVersion < 0x637
			long op = ParseMagickOption(MagickCompositeOptions, MagickTrue, v.data);
			if (op != UndefinedCompositeOp)
				compose = (CompositeOperator)op;
#else
			compose = (CompositeOperator)atoi(v.data);
#endif
		}
		
		if ( lasso_findTagParam(token, kOpacity, &v) == osErrNoErr )
		{
			ExceptionInfo exception;
			GetExceptionInfo(&exception);
			image2 = CloneImage ( data2->image, 0, 0, MagickFalse, &exception);
			DestroyExceptionInfo(&exception);
			
			//0 is fully opaque and MaxRGB is fully transparent.
			SetImageOpacity ( (image2 != NULL) ? image2 : data2->image, (unsigned int)(strtod(v.data, NULL) * MaxRGB) );
		}
			
		if ( image2 != (Image *) NULL ) 
		{
			CompositeImage ( data->image, compose, image2, x_offset, y_offset );
			DestroyImage(image2);
		}
		else
		{
			CompositeImage ( data->image, compose, data2->image, x_offset, y_offset );
		}
		
		if ( data->image->exception.severity != UndefinedException )
		{
			return updateErrorInfo( token, &data->image->exception );
		}
		return osErrNoErr;
	}
	else if ( stricmp(tag.name, kContrast) == 0 )
	{
		ContrastImage(data->image, (MagickBooleanType)(count==0 || stricmp(param.name, "false") != 0));
		if ( data->image->exception.severity != UndefinedException )
		{
			return updateErrorInfo( token, &data->image->exception );
		}
		return osErrNoErr;
	}
	else if ( stricmp(tag.name, kModulate) == 0 )
	{	
		if ( count < 3 )
		{
			lasso_setResultMessage(token, "Modulate tag takes three integer parameters: brightness, saturation and hue (original = 100).");
			return osErrInvalidParameter;
		}
		
		int p[3];
		for (int i = 0; i < 3; i++)
		{
			lasso_getTagParam(token, i, &param);
			p[i] = strtol(param.name, NULL, 10) /* + 100 */;
		}
		
		char bsh[32];
		sprintf(bsh, "%d,%d,%d", p[0], p[1], p[2] );
		
		ModulateImage( data->image, bsh );
		if ( data->image->exception.severity !=	 UndefinedException )
		{
			return updateErrorInfo( token, &data->image->exception );
		}
		return osErrNoErr;
	}
	else if ( stricmp(tag.name, kAnnotate) == 0 )
	{
		if ( !hasX11 )
		{
			lasso_setResultMessage(token, "Library not found: /usr/X11R6/lib/libXext.6.dylib");
			return (osError)-1;
		}
		
		if (count < 2 || param.name == NULL)
		{
			lasso_setResultMessage(token, "Annotate tag takes at least two parameters (text and position).");
			return osErrInvalidParameter;
		}
		
		DrawInfo * draw_info = CloneDrawInfo(data->info, NULL);

		if ( draw_info->text != (char *) NULL )
		{
			RelinquishMagickMemory(draw_info->text);
			draw_info->text = NULL;
		}
		draw_info->text = AllocateString(param.name);
	
		if (draw_info->font != (char *) NULL)
		{
			RelinquishMagickMemory(draw_info->font);
			draw_info->font = NULL;
		}
		if ( lasso_findTagParam(token, kFont, &param) == osErrNoErr )
			draw_info->font = AllocateString(param.data);
		else
			draw_info->font = AllocateString(kDefaultFont);		
 
		if ( lasso_findTagParam(token, kSize, &param) == osErrNoErr )
			draw_info->pointsize = strtod(param.data, NULL);
		else
			draw_info->pointsize = kDefaultTextSize;

		if ( lasso_findTagParam(token, kColor, &param) == osErrNoErr )
		{
			ExceptionInfo exception;
			GetExceptionInfo(&exception);
			if ( QueryColorDatabase(param.data, &draw_info->fill, &exception) == 0 )
			{
				lasso_setResultMessage(token, "Warning: could not set text color."); 
				//printf("EXCEPTION: %d, %s, %s\n", exception.severity, exception.reason, exception.description);
			}
			DestroyExceptionInfo(&exception);
		}
		else
		{
			draw_info->fill.red = draw_info->fill.green = draw_info->fill.blue = (Quantum)0xffff;
		} 
 		
		draw_info->text_antialias = (MagickBooleanType)(lasso_tagParamIsDefined(token, kAliased) == osErrNoErr);
		
		if (draw_info->geometry != (char *) NULL)
		{
			RelinquishMagickMemory(draw_info->geometry);
			draw_info->geometry = NULL;
		}
			
		char geom[32] = "";
		if ( lasso_findTagParam(token, kLeft, &param) != osErrNoErr )
		{
			strcat(geom, "+0");
		}
		else
		{
			if ( *param.data != '+' && *param.data != '-' )
				strcat(geom, "+");
			strcat(geom, param.data);
		}
		
		if ( lasso_findTagParam(token, kTop, &param) != osErrNoErr )
		{
			strcat(geom, "+0");
		}
		else
		{
			if ( *param.data != '+' && *param.data != '-' )
				strcat(geom, "+");
			strcat(geom, param.data);
		}
			
		draw_info->geometry = AllocateString( geom );
				
		if ( AnnotateImage(data->image, draw_info) == 0 )
		{		
			printf("ImageMagick error: Could not annotate the image.\n");
		}
		
		DestroyDrawInfo(draw_info);
		
		if ( data->image->exception.severity != UndefinedException )
		{
			return updateErrorInfo(token, &data->image->exception);
		}
		return osErrNoErr;
	}
		
	Image *image = NULL;
	ExceptionInfo exception;	
	GetExceptionInfo(&exception);
		
	if ( stricmp(tag.name, kFlipV) == 0 )
	{
		image = FlipImage( data->image, &exception );
	}
	else if ( stricmp(tag.name, kFlipH) == 0 )
	{
		image = FlopImage( data->image, &exception );
	}
	else if ( stricmp(tag.name, kCrop) == 0 )
	{
		if ( count == 0 )
		{
			lasso_setResultMessage( token, "Crop member takes at least one parameter." );
			return (osError)-1;
		}

		RectangleInfo geometry;
		geometry.width = data->image->columns;
		geometry.height = data->image->rows;
		
		if ( lasso_findTagParam(token, kWidth, &param) == osErrNoErr )
			geometry.width = strtol(param.data, NULL, 10);
		if ( lasso_findTagParam(token, kHeight, &param) == osErrNoErr )
			geometry.height = strtol(param.data, NULL, 10);
		if ( lasso_findTagParam(token, kLeft, &param) == osErrNoErr )
			geometry.x = strtol(param.data, NULL, 10);
		if ( lasso_findTagParam(token, kTop, &param) == osErrNoErr )
			geometry.y = strtol(param.data, NULL, 10);
		
		image = CropImage ( data->image, &geometry, &exception ); 
	}
	else if ( stricmp(tag.name, kRotate) == 0 )
	{
		PixelPacket oldcolor = data->image->background_color;
		double degrees = strtod(param.name, NULL);
		
		if ( lasso_findTagParam(token, kBgColor, &param) == osErrNoErr && param.dataSize > 0 )
		{
			if ( QueryColorDatabase(param.data, &data->image->background_color, &exception) == 0 )
			{
				//invalid color spec
				lasso_setResultMessage(token, "Warning: could not set image background color.");
			}
		}
		
		// IM rotates images counter-clockwise, while Lasso uses the opposite direction
		image = RotateImage ( data->image, -degrees, &exception );
		data->image->background_color = oldcolor;
	}
	else if ( stricmp(tag.name, kScale) == 0 )
	{		
		long columns = 0;
		long rows = 0;
		bool percentage = false;
		
		if ( lasso_findTagParam(token, kWidth, &param) == osErrNoErr )
		{
			columns = strtol(param.data, NULL, 10);
			percentage = strchr(param.data, '%') != NULL;
		}
		
		if ( lasso_findTagParam(token, kHeight, &param) == osErrNoErr )
		{
			rows = strtol(param.data, NULL, 10);
			if ( strchr(param.data, '%') != NULL )
			{
				columns = (long)(data->image->columns * (columns == 0 ? rows:columns) * 0.01);
				rows = (long)(rows * data->image->rows * 0.01);
			}
			else if ( columns == 0 )
			{
				columns = (long)((double)rows / data->image->rows * data->image->columns);
			}
		}
		else if ( columns > 0 )
		{
			if ( percentage )
			{
				rows = (long)(data->image->rows * columns * 0.01);
				columns = (long)(data->image->columns * columns * 0.01);			
			}
			else
			{
				rows = (long)((double)columns / data->image->columns * data->image->rows);
			}	
		}
		else // error
		{
			lasso_setResultMessage(token, "Scale member tag takes at least one dimension parameter (-height and/or -width)");
		}		
	
		if ( columns > 0 && rows > 0 )
		{
			if ( lasso_tagParamIsDefined(token, kSample) == osErrNoErr )
				image = SampleImage ( data->image, columns, rows, &exception );
			else if ( data->image->colorspace != CMYKColorspace && lasso_tagParamIsDefined(token, kThumbnail) == osErrNoErr )
				image = ThumbnailImage ( data->image, columns, rows, &exception );
			else
				image = ScaleImage ( data->image, columns, rows, &exception ); 
		}
	}
	else if ( stricmp(tag.name, kEnhance) == 0 )
	{	
		image = EnhanceImage(data->image, &exception );
	}
	else if ( stricmp(tag.name, kBlur) == 0 || stricmp(tag.name, kSharpen) == 0 )
	{
		int mode = filterBlur;
		double 
			radius = 0.0, 
			sigma = 1.0, 
			amount = 0.0, // or angle
			threshold = 0.0;

		if ( stricmp(tag.name, kSharpen) == 0 )
			mode = filterSharpen;
			
		for (int i = 0; i < count; i++)
		{
			lasso_getTagParam(token, i, &param);
			if ( stricmp(param.name, kGaussian) == 0 )
			{
				mode = filterGaussianBlur;
			}
			else if ( stricmp(param.name, kRadius) == 0 )
			{
				radius = strtod( param.data, NULL );
			}
			else if ( stricmp(param.name, kSigma) == 0 )
			{
				sigma = strtod( param.data, NULL );
			}
			else if ( stricmp(param.name, kAmount) == 0 )
			{
				amount = strtod( param.data, NULL );
				if (mode == filterSharpen)
					mode = filterUnsharpMask;
			}
			else if ( stricmp(param.name, kAngle) == 0 )
			{
				amount = strtod( param.data, NULL ); // angle
				if (mode != filterSharpen && mode != filterUnsharpMask)
					mode = filterMotionBlur;
			}
			else if ( stricmp(param.name, kThreshold) == 0 )
			{
				threshold = strtod( param.data, NULL );
				if (mode == filterSharpen)
					mode = filterUnsharpMask;
			}
		}
	
		switch (mode)
		{
			case filterSharpen:
				image = SharpenImage ( data->image, radius, sigma, &exception ); 
				break;
			case filterGaussianBlur:
				image = GaussianBlurImage ( data->image, radius, sigma, &exception );
				break;
			case filterMotionBlur:
				image = MotionBlurImage ( data->image, radius, sigma, amount, &exception ); 
				break;
			case filterBlur:
				image = BlurImage ( data->image, radius, sigma, &exception ); 
				break;
			case filterUnsharpMask:
				image = UnsharpMaskImage ( data->image, radius, sigma, amount, threshold, &exception ); 
				break;
		}
	}
	else if ( stricmp(tag.name, kConvert) == 0 )
	{
		if (count < 1 || param.nameSize < 2 || param.nameSize > 31 || param.name == NULL)
		{	
			lasso_setResultMessage( token, "Convert member tag takes at least 1 parameter (image format string)" );
			return osErrInvalidParameter;
		}

		const MagickInfo * fmt = GetMagickInfo( param.name, &exception );
		if (fmt == NULL)
		{
			char temp[128];
			sprintf(temp, "Unsupported ImageMagick format: %s", param.name?param.name:"<null>"); 
			lasso_setResultMessage( token, temp );
			return osErrInvalidParameter;		
		}
				
		ImageInfo * info = CloneImageInfo( (ImageInfo*)NULL );		
		if (info == NULL)
		{
			return osErrOutOfMemory;
		}
		
		size_t len; 
		char oldMagick[32];
		oldMagick[31] = '\0';
		
		strcpy(oldMagick, data->image->magick);
		strncpy(data->image->magick, param.name, param.nameSize+1);
	
		if (count > 1)
		{
			lasso_getTagParam( token, 1, &param );
			if ( stricmp( param.name, kQuality ) == 0 )
			{
				int q = strtol( param.data, NULL, 10 );
				info->quality = ((q - 1) % 100) + 1;
			}
		}
		
		lasso_acquireSem(token, kImageSem);
		void * blob = ImageToBlob(info, data->image, &len, &exception);
		
		if (blob == NULL)
		{
			DestroyImageInfo(info);
			strncpy(data->image->magick, oldMagick, 31);
		}
		else
		{
			image = BlobToImage( info, blob, len, &exception );
			if (image != NULL)
			{
				DestroyImageInfo(data->info);
				data->info = info;
			}
		}
		lasso_releaseSem(token, kImageSem);
	}
				
	if (image == (Image *) NULL)
	{
		osError err = updateErrorInfo(token, &exception);	
		DestroyExceptionInfo(&exception);
		return err;
	}	
				
	DestroyImageList(data->image);
	DestroyExceptionInfo(&exception);

	data->image = image;

	return osErrNoErr;
}

static inline unsigned short _ScaleQuantumToShort(const Quantum quantum)
{
  if (quantum <= 0.0)
    return(0);
  if ((257.0*quantum) >= 65535.0)
    return(65535);
  return((unsigned short) (257.0*quantum));
}

static inline unsigned long _ScaleQuantumToLong(const Quantum quantum)
{
  if (quantum <= 0.0)
    return(0UL);
  if ((65537.0*quantum) >= 4294967295.0)
    return(4294967295UL);
  return((unsigned long) (65537.0*quantum));
}
  
osError 
im_info( lasso_request_t token, tag_action_t action )
{
	lasso_type_t self = NULL;
	lasso_getTagSelf (token, &self);	
	if ( self == NULL )
	{
		lasso_setResultMessage(token, "Could not retrieve custom type instance.");
		return osErrInvalidMemoryObject;
	}
	
	im_type_ptr data = NULL;
	lasso_getPtrMember(token, self, kPrivateData, (void**)&data);
	if ( data == NULL || data->image == (Image *)NULL )
	{
		lasso_setResultMessage(token, "Could not retrieve custom type data.");
		return osErrInvalidMemoryObject;
	}
		
	lasso_type_t out = NULL;
	auto_lasso_value_t	tag;
	lasso_getTagName(token, &tag);

	if ( stricmp(tag.name, kDebugTag) == 0 )
	{
		int count = 0;
		lasso_getTagParamCount(token, &count);
		if ( count > 0 )
		{/*
			auto_lasso_value_t param;
			lasso_getTagParam(token, 0, &param);
			bool debug = param.name && stricmp("true", param.name) == 0 || stricmp("1", param.name) == 0;
			data->image->debug = (MagickBooleanType)debug;
  			
  			ExceptionInfo  exception;
  			GetExceptionInfo(&exception);
			LogInfo * log_info=(LogInfo *) GetLogInfo("*",&exception);
			DestroyExceptionInfo(&exception);
  				
			if (log_info && debug)
			{
#if 0 // 
				log_info->handler_mask = (LogHandlerType)(log_info->handler_mask | FileHandler);
				log_info->event_mask = TraceEvent; //ParseMagickOption(MagickLogEventOptions, MagicFalse, mask);
#endif
			}
			lasso_typeAllocNull(token, &out);
		*/}
		else
		{
			lasso_typeAllocBoolean(token, &out, data->image->debug == MagickTrue);
		}
	}
	else if ( stricmp(tag.name, kCount) == 0 )
	{
		lasso_typeAllocInteger(token, &out, GetImageListLength(data->image));
	}
	else if ( stricmp(tag.name, kWidthTag) == 0 )
	{
		lasso_typeAllocInteger(token, &out, data->image->columns);
	}
	else if ( stricmp(tag.name, kHeightTag) == 0 )
	{
		lasso_typeAllocInteger(token, &out, data->image->rows);
	}
	else if ( stricmp(tag.name, kDepthTag) == 0 )
	{
		lasso_typeAllocInteger(token, &out, data->image->depth);
	}
	else if ( stricmp(tag.name, kLength) == 0 )
	{
		lasso_typeAllocInteger(token, &out, GetBlobSize(data->image));
	}
	else if ( stricmp(tag.name, kResolutionV) == 0 )
	{
		lasso_typeAllocDecimal(token, &out, data->image->y_resolution);
	}
	else if ( stricmp(tag.name, kResolutionH) == 0 )
	{
		lasso_typeAllocDecimal(token, &out, data->image->x_resolution); 
	}
	else if ( stricmp(tag.name, kFormat) == 0 )
	{
		return lasso_outputTagData(token, data->image->magick);
	}
	else if ( stricmp(tag.name, kFile) == 0 )
	{
		if ( data->image->filename != (char*) NULL )
			return lasso_outputTagData(token, data->image->filename);
		lasso_typeAllocNull(token, &out);
	}
	else if ( stricmp(tag.name, kComments) == 0 )
	{
		const ImageAttribute * attr = GetImageAttribute( data->image, "Comment" );
		if ( attr != (ImageAttribute *)NULL )
			lasso_outputTagData(token, attr->value);	
		return osErrNoErr;
	}
	else if ( stricmp(tag.name, kAddComment) == 0 )
	{
		auto_lasso_value_t param;
		if ( lasso_getTagParam(token, 0, &param) == osErrNoErr && param.name != NULL )
		{
			if ( stricmp(param.name, "null") == 0 )
			{
				DeleteImageAttribute( data->image, "Comment" );
			}
			else
			{
				SetImageAttribute( data->image, "Comment", param.name );
			}
		}
		return osErrNoErr;
	}
	else if ( stricmp(tag.name, kData) == 0 || stricmp(tag.name, kBytes) == 0 )
	{
		size_t size = GetBlobSize(data->image);
				
		if (size == 0 )
		{
			size = 1024*1024;
		}
		{
			ExceptionInfo exception;
			GetExceptionInfo( &exception );
			
			lasso_acquireSem(token, kImageSem);
			unsigned char * blob = (unsigned char *)ImageToBlob(data->info, data->image, &size, &exception);
			lasso_releaseSem(token, kImageSem);
			
			if (blob != NULL)
			{
				osError err = lasso_outputTagBytes( token, (const char*) blob, (int)size );
				RelinquishMagickMemory(blob);
				return err;
				
			}
			return updateErrorInfo(token, &exception);
		}
	}
	else if ( stricmp(tag.name, kDescribe) == 0 )
	{
		int fildes[2];
		if ( pipe(fildes) != 0 )
		{
			lasso_setResultMessage(token, "Could not create pipe.");
			lasso_log(LOG_LEVEL_CRITICAL, "Error in Image->Describe: could not create pipe.");
			return osErrCouldNotCreateOrOpenFile;
		}

		FILE * f = fdopen(fildes[1], "w");
		if (f != NULL)
		{
			MagickBooleanType verbose=(MagickBooleanType)(lasso_tagParamIsDefined(token, kShort) != osErrNoErr);
#if MagickLibVersion > 0x623
			DescribeImage2(data->image, f, verbose); 
#else	
			DescribeImage2(data->image, f, verbose);
#endif
			fclose(f);
			FILE * f = fdopen(fildes[0], "r");
			if (f != NULL)
			{
				char buf[1024];
				size_t read = 0;
				while ( (read = fread(buf, 1, 1024, f)) > 0 )
				{
					lasso_outputTagData2( token, buf, (int)read );
				}
				fclose(f);
			}
		}

		return osErrNoErr;
	}
	
	else if ( stricmp(tag.name, kPixel) == 0 )
	{
		int count = 0;
		lasso_getTagParamCount(token, &count);
		if ( count < 2 )
		{
			lasso_setResultMessage(token, "Pixel member tag takes two integer parameters (left, top).");
			return (osError)-1;
		}

		char tuple[64];
		auto_lasso_value_t param;
			
		lasso_getTagParam(token, 0, &param);	
		const long x = strtol(param.name, NULL, 10);
		
		lasso_getTagParam(token, 1, &param);
		const long y = strtol(param.name, NULL, 10);
		
		unsigned int depth = 0; 
		bool hex = lasso_tagParamIsDefined(token, kHex) == osErrNoErr;
		
		if ( lasso_findTagParam(token, kDepth, &param) == osErrNoErr )
			depth = strtol(param.data, NULL, 10);
		if (depth == 0)
			depth = data->image->depth;
		
		const PixelPacket color = GetOnePixel ( data->image, x, y );
#if MagickLibVersion >= 0x608
		MagickPixelPacket mpp;
		memset(&mpp, 0, sizeof(mpp));
		mpp.matte = data->image->matte;
		mpp.depth = depth;
		mpp.red = color.red;
		mpp.green = color.green;
		mpp.blue = color.blue;
		mpp.opacity = color.opacity;
		GetColorTuple(&mpp, (MagickBooleanType)hex, tuple);
#else
		GetColorTuple(&color, depth, data->image->matte, (MagickBooleanType)hex, tuple);
#endif
		if ( hex )
			return lasso_outputTagData(token, tuple);
			
		lasso_type_t rgb[4] = {NULL, NULL, NULL, NULL};
		
		if ( depth <= 8 )
		{
			lasso_typeAllocInteger(token, &rgb[0], ScaleQuantumToChar(color.red));
			lasso_typeAllocInteger(token, &rgb[1], ScaleQuantumToChar(color.green));
			lasso_typeAllocInteger(token, &rgb[2], ScaleQuantumToChar(color.blue));
			if ( data->image->matte )
				lasso_typeAllocInteger(token, &rgb[3], ScaleQuantumToChar(color.opacity));	
		}
		else if ( depth <= 16 )
		{
			lasso_typeAllocInteger(token,  &rgb[0], _ScaleQuantumToShort(color.red));
			lasso_typeAllocInteger(token,  &rgb[1], _ScaleQuantumToShort(color.green));
			lasso_typeAllocInteger(token,  &rgb[2], _ScaleQuantumToShort(color.blue));
			if ( data->image->matte )
				lasso_typeAllocInteger(token, &rgb[3], _ScaleQuantumToShort(color.opacity)); 
		}
		else
		{
			lasso_typeAllocInteger(token,  &rgb[0], _ScaleQuantumToLong(color.red));
			lasso_typeAllocInteger(token,  &rgb[1], _ScaleQuantumToLong(color.green));
			lasso_typeAllocInteger(token,  &rgb[2], _ScaleQuantumToLong(color.blue));
			if ( data->image->matte )
				lasso_typeAllocInteger(token,  &rgb[3], _ScaleQuantumToLong(color.opacity)); 
		}
		
		lasso_typeAllocArray(token, &out, data->image->matte ? 4 : 3, rgb);
	}
	else if ( stricmp(tag.name, kHistogram) == 0 )
	{
		osError err = osErrNoErr;
		Image *histogram = NULL;
    	ExceptionInfo exception;
	
    	GetExceptionInfo(&exception);
#if MagickLibVersion > 0x628
	    histogram = UniqueImageColors(data->image, &exception);
#endif
		if (histogram == NULL)
		{
			err = updateErrorInfo(token, &exception);
		}
		else
		{
			lasso_type_t pName = NULL, pData = NULL, p = NULL;
			lasso_typeAllocString(token, &pName, kPrivatePrm, strlen(kPrivatePrm));
			lasso_typeAllocInteger(token, &pData, (osInt64)histogram);
			lasso_typeAllocPair(token, &p, pName, pData);
			
			err = lasso_typeAlloc(token, kImage, 1, &p, &out);
			if (err != osErrNoErr || out == NULL)
			{
				lasso_setResultMessage(token, "Could not allocate image type (histogram).");
				DestroyImageList(histogram);
				err = osErrInvalidMemoryObject;
			}
		}
		DestroyExceptionInfo(&exception);
	}
	else if ( stricmp(tag.name, kSetColorSpace) == 0 )
	{
		auto_lasso_value_t param;
		INITVAL(&param);
		
		if ( lasso_findTagParam(token, kRGB, &param) == osErrNoErr )
			::SetImageColorspace(data->image, RGBColorspace);
		else if ( lasso_findTagParam(token, kCMYK, &param) == osErrNoErr )
			::SetImageColorspace(data->image, CMYKColorspace);
		else if ( lasso_findTagParam(token, kGRAY, &param) == osErrNoErr )
			::SetImageColorspace(data->image, GRAYColorspace);
		else
		{
			lasso_setResultMessage(token, "Invalid colorspace");
			return osErrInvalidParameter;
		}
		lasso_typeAllocNull(token, &out);
	}
	else if ( stricmp(tag.name, kColorSpace) == 0 )
	{
		switch(data->image->colorspace)
		{
			case RGBColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"RGB"), 3);
				break;
			case GRAYColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"GRAY"), 4);
				break;
			case TransparentColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"Transparent"), 11);
				break;
			case OHTAColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"OHTA"), 4);
				break;
			case LABColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"LAB"), 3);
				break;
			case XYZColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"XYZ"), 3);
				break;
			case YCbCrColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"YCbCr"), 5);
				break;
			case YCCColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"YCC"), 3);
				break;
			case YIQColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"YIQ"), 3);
				break;
			case YPbPrColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"YPbPr"), 5);
				break;
			case YUVColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"YUV"), 3);
				break;
			case CMYKColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"CMYK"), 4);
				break;
			case sRGBColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"sRGB"), 4);
				break;
			case HSBColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"HSB"), 3);
				break;
			case HSLColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"HSL"), 3);
				break;
			case HWBColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"HWB"), 3);
				break;
			case Rec601LumaColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"Rec601Luma"), 10);
				break;
			case Rec601YCbCrColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"Rec601YCbCr"), 11);
				break;
			case Rec709LumaColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"Rec709Luma"), 10);
				break;
			case Rec709YCbCrColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"Rec709YCbCr"), 11);
				break;
			case LogColorspace:
				lasso_typeAllocStringW(token, &out, LPW(L"Log"), 3);
				break;
			default:
				lasso_typeAllocStringW(token, &out, LPW(L"UNDEFINED"), 9);
				break;
		}
	}
	
	return lasso_returnTagValue(token, out);
}


void registerLassoModule()
{
#if MACHINE_MACOSX
	int res = access("/usr/X11R6/lib/libXext.6.dylib", 0); //libX11.6.dylib
	if ( res == -1 ) 
	{
		lasso_log(LOG_LEVEL_CRITICAL, "ImageTag module disabled (libXext.6.dylib not found)");
		return;
	}
	setenv("MAGICK_HOME", "/ImageMagick-6.3.2", 0);
	setenv("MAGICK_FONT_PATH", "/System/Library/Fonts/", 0);
#endif
	
	ExceptionInfo exception;
	InitializeMagick( "Lasso8Service" );
	GetExceptionInfo( &exception );
	
	if (exception.severity == UndefinedException)
	{
		lasso_registerTagModule("magick", "image", im_init, flag_typeSubstitutionTag | flag_typeInitializer, "image type");
#ifdef WIN32
		size_t ul;
#elif MagickLibVersion >= 0x663
		size_t ul;
#else
		unsigned long ul;
#endif
		lasso_log(LOG_LEVEL_WARNING, "Using %s", GetMagickVersion(&ul));
	}
	else
	{
		lasso_log(LOG_LEVEL_CRITICAL, "Could not initialize ImageMagick library.");
	}
	
	lasso_createSem( (lasso_request_t)NULL, kImageSem );
	
	DestroyExceptionInfo(&exception);
}
