import java.io.*;
import java.util.*;
import java.util.zip.*;
import com.omnipilot.lassopro.*;

public class ZipType extends LassoTagModule
{
	ZipFile 	zip = null;
	Enumeration items = null;
	
	String[] members = { 
		"Data", 
		"Name", 
		"Comment", 
		"Time", 
		"Crc", 
		"Extra",
		"Size", 
		"CompressedSize", 
		"IsDir",
		"File",
		"Count"
	};
	
	public void registerLassoModule()
	{
		registerTagModule( "ZipType", "zip", "format", FLAG_SUBSTITUTION | FLAG_INITIALIZER, "zip file" );
	}
	
	public int format(LassoCall lasso, int action)
	{	
		int err = ERR_NOERR;

		// define the variable to hold the tag param value
		LassoValue param = new LassoValue();
		
		// zip file pathname
		String pathname = null;
		
		// get the first tag param
		if ( lasso.getTagParam( 0, param ) != ERR_NOERR || param.name().length() < 1 )
		{
			lasso.setResultMessage("[Zip] tag: invalid file pathname parameter");
			return LassoErrors.InvalidParameter;	
		}
							
		try
		{
			// define the storage for the tag param count
			IntValue count = new IntValue();
			
			// how many params this tag has?
			err = lasso.getTagParamCount( count );
			
			// proceed if the tag has at least one parameter
			if ( err == ERR_NOERR && count.intValue() > 0 )
			{					
				// convert absolute or relative path into from-the-server-root path.
				String filePath = lasso.fullyQualifyPath( param.name() );
					
				// now convert a from-the-root path into a full internal path
				filePath = lasso.resolvePath( filePath );
				
				//finally, convert internal path to a platform-specific path
				filePath = lasso.getPlatformSpecificPath( filePath );

				// try to create new instance of a ZipFile object														
				ZipFile zip = new ZipFile( filePath );	
		
				// allocate new instance of the zip data type
				LassoTypeRef self = new LassoTypeRef();
				if ( (err = lasso.typeAllocCustom( self, "zip" )) != ERR_NOERR )
				{
					lasso.setResultMessage("[Zip] tag: couldn't create zip type.");
					return err;
				}
				
				// add custom type members
				LassoTypeRef ref = new LassoTypeRef();
				String className = this.getClass().getName();				
				for ( int i = 0; i < members.length; i++ )
				{
					if ( (err = lasso.typeAllocTag( ref, className, "zipMemberFunc" )) != ERR_NOERR ||
						 (err = lasso.typeAddMember( self, members[i], ref )) != ERR_NOERR )
					{
						lasso.setResultMessage("[Zip] tag: error adding member " + members[i] );
						return err;
					}
				}
				
				// add onDestroy callback
				if ( lasso.typeAllocTag( ref, className, "zipDestroyFunc" ) == ERR_NOERR )
					lasso.typeAddMember( self, "onDestroy", ref );
				
				// attach this Java class instance to the custom type		 
				if ( (err = lasso.typeSetCustomJavaObject( self, zip )) != ERR_NOERR )
				{
					lasso.setResultMessage("[Zip] tag: couldn't attach ZipFile java object to a custom type");
					return err;
				}
				
				return lasso.returnTagValue(self);
			}
		}
		catch (ZipException e)
		{
			System.err.println( e.toString() );
			lasso.setResultMessage( e.getMessage() );
			return LassoErrors.FileNotFound;
		}
		catch ( java.io.IOException e )
		{
			System.err.print( e.toString() );
			lasso.setResultMessage( e.getMessage() );			
			return LassoErrors.FileNotFound;
		}
		return err;
	}
	
	// onDestroy callback
	public int destroyFunc(LassoCall lasso, int action)
	{
		LassoTypeRef me = new LassoTypeRef();
				
		if ( lasso.getTagSelf(me) == ERR_NOERR )
		{
			ZipFile zip = (ZipFile)lasso.typeGetCustomJavaObject(me);
			if (zip != null)
				try {
					zip.close();
				} catch (IOException e) {}
			lasso.typeFreeCustomJavaObject(me);
		}
		return ERR_NOERR;
	}
	
	// member tag method
	public int memberFunc(LassoCall lasso, int action)
	{
		int err = ERR_NOERR;
		ZipFile zip = null;
		LassoTypeRef me = new LassoTypeRef();
				
		if ( lasso.getTagSelf(me) != ERR_NOERR || 
			 me.isNull() ||
			 (zip = (ZipFile)lasso.typeGetCustomJavaObject(me)) == null )
		{
			lasso.setResultMessage("[Zip] tag: couldn't retrieve custom ZipFile java object");
			return LassoErrors.InvalidMemoryObject;
		}
	
		LassoValue tag = new LassoValue();
		
		if ( lasso.getTagName(tag) != ERR_NOERR ||
			 tag.name() == null || 
			 tag.name().length() < 1 )
		{
			return LassoErrors.InvalidParameter;
		}
	
		int idx = -1;
		String name = null;
		LassoValue param = new LassoValue();
		
		if ( lasso.getTagParam(0, param) == ERR_NOERR )
		{
			if ( param.type() == LassoValue.TYPE_INT )
			{
				try { idx = Integer.parseInt( param.name() ); }
					catch (Exception npe) {}
			}
			else
			{
				name = param.name();
			}
		}
		
		LassoTypeRef out = new LassoTypeRef();
		ZipEntry entry = null;
		
		if ( name != null )
		{ 
			entry = zip.getEntry(name);
		}
		else if (idx > -1)
		{
			//item = getIdxEntry(zip, index); 
			Enumeration enumer = zip.entries();
			int count = 0;
			
			// iterate through zip archive items, incrementing the count
			while ( enumer.hasMoreElements() && count++ < idx )
			{
				entry = (ZipEntry)enumer.nextElement();
			}
		}
		else if ( tag.name().equalsIgnoreCase("file" ) )
		{
			return lasso.outputTagData( zip.getName() );
		}
		else if ( tag.name().equalsIgnoreCase("count" ) )
		{
			lasso.typeAllocInteger( out, zip.size() );
			return lasso.returnTagValue( out );
		}		
		
		String result = null;
		
		if ( tag.name().equalsIgnoreCase("name") )
			result = entry.getName();
		else if ( tag.name().equalsIgnoreCase("comment") )
			result = entry.getComment();
		else if ( tag.name().equalsIgnoreCase("crc") )
			result = Long.toHexString( entry.getCrc() );
		else if ( tag.name().equalsIgnoreCase("method") )
			result = ( entry.getMethod() == ZipEntry.DEFLATED ? "DEFLATED" : "STORED" );		
			
		if ( result != null )
			return lasso.outputTagData( result );
		
		if ( tag.name().equalsIgnoreCase("size") )
			lasso.typeAllocInteger(out, entry.getSize());
		else if ( tag.name().equalsIgnoreCase("compressedSize") )
			lasso.typeAllocInteger(out, entry.getCompressedSize());
		else if ( tag.name().equalsIgnoreCase("time") )
			lasso.typeAllocInteger(out, entry.getTime());
		else if ( tag.name().equalsIgnoreCase("isDir") )
			lasso.typeAllocBoolean(out, entry.isDirectory());

		if ( !out.isNull() )
			return lasso.returnTagValue( out );
		
		return err;
	}

}

