#ifndef _LASSOCAPI_H_
#define _LASSOCAPI_H_


/** \mainpage LassoCAPI
 *
 * This is the Doxygen reference for LassoCAPI 8.
 */


/** \file LassoCAPI.h
 * @brief LassoCAPI Main Header File
 *   
 * This file contains all of the available LCAPI defines and functions.
 */

#include "osConfig.h"
#include "osErrors.h"
#include "RequestParams.h"
#include "LassoDefines.h"
#include <time.h>

#ifdef __cplusplus
extern "C"
{
    using namespace LPTypes;

#endif

#ifdef WIN32
#ifdef EXPORT_LCAPI /* external modules don't define anything */
#define LCAPICALL __declspec(dllexport)
#else
#define LCAPICALL __declspec(dllimport)
#endif
#else
#ifdef EXPORT_LCAPI /* external modules don't define anything */
#define LCAPICALL // automatically all exported
#else
#define LCAPICALL extern
#endif
#endif

/**
	Defines for testing the LCAPI version
 */
#define LCAPI_VERSION 4   /**< Current Lasso Version. */
#define LCAPI_VERSION_1 1 /**< Lasso Version 5. First LCAPI release */
#define LCAPI_VERSION_2 2 /**< Lasso Version 6. Second LCAPI release */
#define LCAPI_VERSION_3 3 /**< Lasso Version 7. Third LCAPI release */
#define LCAPI_VERSION_4 LCAPI_VERSION /**< Lasso Version 8. Fourth LCAPI release */

/** 
	@brief Used for retrieving data from some LCAPI functions.
	
	The lasso_value_t struct is used for shuffling data 
	into and out of LassoCAPI functions.

	Any LCAPI function that constructs one of these will insure that Lasso properly 
	disposes of the data once the user's LassoCAPI function call is complete.
	
	lasso_value_t is suitable for transferring binary data as long as the \a nameSize or \a dataSize
	members are properly set with the size of the data in bytes.
	
	Depending on the purpose of the function which is constructing the lasso_value_t, any combination 
	of the \a name or \a data members may be filled in or may be NULL. Consult the documentation of 
	the specific function for more details.
	
	Should be initialized using the INITVAL(X) macro before use.
*/
typedef struct 
{
	/**
	 *	@brief The \a name portion.
	 *
	 *	May be NULL.
	 */
	const char * name;
	/**
	 *	@brief The size of the \a name member, in bytes.
	 *
	 *
	 */
	unsigned int nameSize;
	/**
	 *	@brief The \a value portion.
	 *
	 *	May be NULL.
	 */
	const char * data;
	/**
	 *	@brief The size of the \a data member, in bytes.
	 *
	 *
	 */
	unsigned int dataSize;
	/**
	 *	@brief The type code for the \a data member.
	 *
	 *
	 */
	LP_TypeDesc type;
} lasso_value_t;

/**
	@brief Used for retrieving Unicode character data from some LCAPI functions.
	
	The lasso_value_w_t struct is used for shuffling Unicode character data 
	into and out of LassoCAPI functions.

	Any LCAPI function that constructs one of these will insure that Lasso properly 
	disposes of the data once the user's LassoCAPI function call is complete.
	
	Depending on the purpose of the function which is constructing the lasso_value_t, any combination 
	of the \a name or \a data members may be filled in or may be NULL. Consult the documentation of 
	the specific function for more details.
	
	Should be initialized using the INITVAL(X) macro before use.
*/
typedef struct 
{
	/**
	 *	@brief The \a name portion.
	 *
	 *	May be NULL.
	 */
	const UChar * name;
	/**
	 *	@brief The size of the \a name member, in bytes.
	 *
	 *
	 */
	unsigned int nameSize;
	/**
	 *	@brief The \a value portion.
	 *
	 *	May be NULL.
	 */
	const UChar * data;
	/**
	 *	@brief The size of the \a data member, in bytes.
	 *
	 *
	 */
	unsigned int dataSize;
	/**
	 *	@brief The type code for the \a data member.
	 *
	 *
	 */
	LP_TypeDesc type;
} lasso_value_w_t;

/**
	@brief Initialize a lasso_value_t or lasso_value_w_t to be blank.
	
	It is recommended that each lasso_value_t be initialized using this macro before use.
	
	Example:
	<pre>
		lasso_value_t myVal;
		INITVAL(&myVal);
	</pre>
*/
#define INITVAL(X) { (X)->name = (X)->data = NULL; (X)->nameSize = (X)->dataSize = 0; }

/**
	Special typedef so programmers know when Lasso will automatically dispose of the value.
*/
typedef lasso_value_t auto_lasso_value_t;
/**
	Special typedef so programmers know when Lasso will automatically dispose of the value.
*/
typedef lasso_value_w_t auto_lasso_value_w_t;

/**
	@brief Allocates a lasso_value_t with the indicated data.
	
	Anything allocated with this function will NOT be garbage collected by Lasso
	and must be freed using lasso_freeValue.
	
	@param result A pointer to the lasso_value_t which will be constructed.
	@param name The name portion which will be copied and set.
	@param nameSize The size of the name parameter in bytes.
	@param data The data portion which will be copied and set.
	@param dataSize The size of the data parameter in bytes.
	@param type The type code for the value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_allocValue( lasso_value_t * result, const char * name, unsigned int nameSize, const char * data, unsigned int dataSize, LP_TypeDesc type );
/**
	@brief Allocates a lasso_value_w_t with the indicated data.
	
	Anything allocated with this function will NOT be garbage collected by Lasso
	and must be freed using lasso_freeValueW.
	
	@param result A pointer to the lasso_value_w_t which will be constructed.
	@param name The name portion which will be copied and set.
	@param nameSize The number of characters in the name portion.
	@param data The data portion which will be copied and set.
	@param dataSize The number of characters in the data portion.
	@param type The type code for the value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_allocValueW( lasso_value_w_t * result, const UChar * name, unsigned int nameSize, const UChar * data, unsigned int dataSize, LP_TypeDesc type );
/**
	@brief Allocates a lasso_value_t with the indicated data and encoding method.
	
	This function can be used to convert Unicode data into any of the supported encodings.
	
	Anything allocated with this function will NOT be garbage collected by Lasso
	and must be freed using lasso_freeValue.
	
	@param result A pointer to the lasso_value_w_t which will be constructed.
	@param name The name portion which will be converted and set.
	@param nameSize The number of characters in the name portion.
	@param nameEncoding The destination encoding for the name portion.
	@param data The data portion which will be converted and set.
	@param dataSize The number of characters in the data portion.
	@param dataEncoding The destination encoding for the data portion.
	@param type The type code for the value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_allocValueConv( lasso_value_t * result, const UChar * name, unsigned int nameSize, const char * nameEncoding, 
																const UChar * data, unsigned int dataSize, const char * dataEncoding, LP_TypeDesc type );
/**
	@brief Frees a previously allocated lasso_value_t
	
	Do not pass an auto_lasso_value_t to this function or you will end up with a double free.
	
	@param result The lasso_value_t whose members will be freed.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_freeValue( lasso_value_t * result );
/**
	@brief Frees a previously allocated lasso_value_w_t
	
	Do not pass an auto_lasso_value_w_t to this function or you will end up with a double free.
	
	@param result The lasso_value_w_t whose members will be freed.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_freeValueW( lasso_value_w_t * result );


/*********************************************************
	DATASOURCE ACTIONS
**********************************************************/
/**
	@brief Types of actions datasources could be called for.
	
	One of these will be passed as the \a action parameter for the lasso_ds_func.
*/
typedef enum
{
	datasourceInit,			/**< Sent when a new "instance" of the datasource module is created. */
	datasourceTerm,			/**< Sent when the "instance" of the datasource module is destroyed. */
	datasourceExists,		/* UNUSED */

	datasourceNames,		/**< Sent when Lasso attempts to gather the names of all the databases that the datasource module supports.
							 	 Call lasso_addDataSourceResult once for each supported database.
							 */
	datasourceTableNames,   /**< Sent when Lasso attempts to gather the names of the tables available for the given database.
								 The name of the database itself will be passed in the \a param parameter.
							 	 Call lasso_addDataSourceResult once for each available table.
							 */
	datasourceSearch,		/**< Sent when the datasource module is to perform a -search action. */ 
	datasourceAdd,			/**< Sent when the datasource module is to perform a -add action. */
	datasourceUpdate,		/**< Sent when the datasource module is to perform a -update action. */
	datasourceDelete,		/**< Sent when the datasource module is to perform a -delete action. */
	datasourceInfo,			/**< Sent when the datasource module is to perform a -show action. */
	datasourceExecSQL,		/**< Sent when the datasource module is to perform a -sql action. */

	datasourceRandom,		/**< Sent when the datasource module is to perform a -random action. */

	datasourceSchemaNames, 	/**< Sent when Lasso attempts to gather the names of the schemas available for the given database.
								 The name of the database itself will be passed in the \a param parameter.
							 	 Call lasso_addDataSourceResult once for each available schema.
							 	 
							 	 This is currently only utilized for LJAPI.
							 */

	datasourceCloseConnection, /**< Sent when the datasource module should close a connection previously 
									set via the lasso_setDSConnection function. 
								*/
	datasourceTickle, 		/**< Sent to the datasource module when -database and -table are specified in an inline, 
							     but no action, or a -nothing action is used.
							     The database could, perhaps, set or reset its connection to the database via lasso_setDSConnection.
							     Or, it could do nothing.
							 */
	datasourceDuplicate,	// FileMaker module ONLY impliments this message
	datasourceScripts,		// FileMaker module ONLY impliments this message
	datasourceImage,		// FileMaker module ONLY impliments this message

	datasourceFindAll,		/**< Sent when the datasource module is to perform a -findall action. */
	datasourceMatchesName,	/**< Sent to the datasource to find out if it "goes" under the given name. 
								 The name which is being tested will be passed in the name member of the \a param parameter.
								 The data member of the \a param parameter will be NULL and the dataSize member will be zero.
								 If the name matches the name the datasource goes under, the datasource 
								 should both set the dataSize member to non-zero and return osErrNoErr. 
								 Otherwise, it is assumed that the name is not a matching name for the datasource module.
							*/
	datasourcePrepareSQL,	/**< Sent to the datasource to prepare a sql statement for later execution.
								 This is sent when the datasource action was -prepare. The statement text is sent to the
								 datasource in the \a data member of the \a param parameter that is passed to each datasource call.
							*/
	datasourceUnprepareSQL  /**< Sent to the datasource after a datasourcePrepareSQL action.
								 This permits the datasource to perform and necessary cleanup activities after executing a prepared statement.
							*/
} datasource_action_t;

#define SET_MATCHED_NAME(param) { const_cast<auto_lasso_value_t*>(param)->dataSize = 1; }

/*********************************************************
	TAG ACTIONS
**********************************************************/
/**
	@brief Types of actions tag modules could be called for.
	
	Ignore this for now, it may be put into use in a future version but is not utliized at present.
*/
typedef int tag_action_t;

const tag_action_t tagFormatSubstitution = 1;

/*********************************************************
	SESSION OPTIONS - UNUSED
**********************************************************/
/*
	Various options that can be set for the current session.
*/
typedef enum
{
	optNone			/*  no options yet. this is for future use */
		
} session_option_t;

/*********************************************************
	LASSO REQUEST TOKEN/DATASOURCE TOKEN
**********************************************************/
/**
	@brief Special value passed to modules that identify the request.
	
	The same value should be used when calling into any LassoCAPI function.
*/
typedef void * lasso_request_t;

/*********************************************************
	LASSO TYPE TOKEN - LCAPI 6
**********************************************************/
/**
	@brief Represents a type within Lasso
	
	This opaque value represents an instance of a type within Lasso
*/
typedef void * lasso_type_t;

/*********************************************************
	DATASOURCE CONNECTION TOKEN - LCAPI 7
**********************************************************/
/**
	@brief Represents a datasource module connection.
	
	This opaque value is only interpreted by the datasource module itself.
	It can be stored via lasso_setDSConnection and retrieved via lasso_getDSConnection.
	Lasso will automatically send the datasource the datasourceCloseConnection message when it is time to close
	the connection.
*/
typedef void * lasso_dsconnection_t;

/*********************************************************
	MODULE SERVICE FUNCTIONS
**********************************************************/
/**
	@brief Service function for Lasso Datasource modules
	
	A LCAPI datasource module should implement one of these and pass it to the lasso_registerDSModule or lasso_registerDSModuleW function 
	when the datasource module is registered. The function will be called by Lasso when a datasource operation is to be performed.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param action The current action for the datasource module to perform.
	@param param Any additional data associated with the action.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
typedef osError (*lasso_ds_func)(lasso_request_t token, datasource_action_t action, const auto_lasso_value_t * param);
/**
	@brief Service function for a Lasso tag
	
	An LCAPI tag should implement one of these and pass it to the lasso_registerTagModule or lasso_registerTagModuleW function
	when the LCAPI module is registered. The function will be called by Lasso every time the tag is called in a script.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param action Unused. Do not rely on this parameter to hold any particular value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
typedef osError (*lasso_tag_func)(lasso_request_t token, tag_action_t action);

/*********************************************************
	MODULE REGISTRATION
**********************************************************/
/**
	@brief LCAPI module registration callback.
	
	All Lasso modules must export a function named "registerLassoModule".
	When Lasso attempts to load an LCAPI module, it will look for this exported funtion and, if found, call it.
	Within that function, the user may register any number of datasource modules or any number of tags using the 
	lasso_registerDSModule, lasso_registerDSModuleW, lasso_registerTagModule or lasso_registerTagModuleW functions.
*/
typedef void (*register_module_func)(void);

/** @name Tag Registration Flags 

	The following flags may be OR'd together in various combinations and passed to lasso_registerTagModule or lasso_registerTagModuleW
	as the \a flags parameter to control how the tag behaves when it is called.
*/
/*@{*/
const int flag_typeInitializer =		0x00000001; /**< @brief The tag is to be treated as an initializer for a custom type. */
const int flag_typeSubstitutionTag =	0x00000002; /**< @brief The tag is to be treated as a "regular" tag. 
						
														 This is the default behavior so this flag is not required. 
													 */
const int flag_typeAsync =				0x00000004; /**< @brief The tag is to be run in its own thread. 

														 It will return no value to the caller. 
													 */
const int flag_typeContainerTag =		0x00000008; /**< @brief The tag is a container tag.

														 The tag must be called using the proper <pre>tag; ... /tag;</pre> syntax or an error is generated.
														 The result of executing any body statements can be retrieved using the lasso_childrenRun function.
													 */
const int flag_typeInterstitial =		0x00000010; /**< @brief Used internally. */
const int flag_typeSkipSecurityCheck =	0x00000040; /**< @brief Bypass any security checks for the tag. 

														 Useful to avoid any performance hits associated with security checks. 
														 Should only be used when a tag performs only trival or completely secure operations. 
													 */
const int flag_INTERNALONLY =			0x00000080; /**< @brief Used internally. */
const int flag_typeLoopingTag =			0x00000100; /**< @brief The tag is a looping container tag. 

														 For any tag registered with this flag, Lasso will automatically 
														 keep track of the loop_count and increment it each time the tag calls lasso_childrenRun.
													 */
const int flag_typeEval =				0x00000200; /**< @brief Used internally. Not useful for LCAPI tag types. */
const int flag_noGlobalImport =			0x00000400; /**< @brief The tag should not be automatically imported into the global namespace.

														 All LCAPI tags, starting with LP8, which are registered with a namespace should specify this flag.
													 */
const int flag_INTERNALONLY2 =			0x00000800; /**< @brief Used internally. */
const int flag_deprecated =				0x00001000; /**< @brief Use of the tag is deprecated. 

														 Depending on the administrator's configuration, calling the tag will 
														 automatically output a warning to the logging system.
													 */
const int flag_noDefaultEncoding =		0x00002000; /**< @brief Never apply default HTML encoding to the tag's return value. */
const int flag_INTERNALONLY3 =			0x00004000;
const int flag_prototype =				0x00008000; /**< @brief Applies only to types registered with lasso_registerTypeModule.
														
														An instance of the type will be created immediately (onCreate will not be called).
														Any further calls to this type will result in a fast copy of the prototype being created.
													*/
const int flag_atomic =					0x00010000;
const int flag_private =				0x00020000;
// feel free to use these flags with lasso_typeXXXFlag
const int flag_User1 =					0x01000000; /**< @brief User defined flag.

														 May be used with lasso_tagSetFlag, lasso_tagHasFlag, lasso_tagClearFlag OR
														 lasso_typeSetFlag, lasso_typeHasFlag, lasso_typeClearFlag
													 */
const int flag_User2 =					0x02000000; /**< @brief User defined flag.

														 May be used with lasso_tagSetFlag, lasso_tagHasFlag, lasso_tagClearFlag OR
														 lasso_typeSetFlag, lasso_typeHasFlag, lasso_typeClearFlag
													 */
const int flag_User3 =					0x03000000; /**< @brief User defined flag.

														 May be used with lasso_tagSetFlag, lasso_tagHasFlag, lasso_tagClearFlag OR
														 lasso_typeSetFlag, lasso_typeHasFlag, lasso_typeClearFlag
													 */
const int flag_User4 =					0x04000000; /**< @brief User defined flag.

														 May be used with lasso_tagSetFlag, lasso_tagHasFlag, lasso_tagClearFlag OR
														 lasso_typeSetFlag, lasso_typeHasFlag, lasso_typeClearFlag
													 */
/** 
	@brief Recommended default registration flags.
	
	For a normal tag.
*/
const int REG_FLAGS_TAG_DEFAULT = (flag_typeSubstitutionTag | flag_noGlobalImport);
/** 
	@brief Recommended default registration flags.
	
	For a container tag.
*/
const int REG_FLAGS_CONTAINER_DEFAULT = (flag_typeContainerTag | flag_noGlobalImport);
/** 
	@brief Recommended default registration flags.
	
	For a looping container tag.
*/
const int REG_FLAGS_LOOPING_DEFAULT = (flag_typeLoopingTag | flag_noGlobalImport);
/** 
	@brief Recommended default registration flags.
	
	For a type initializer tag.
*/
const int REG_FLAGS_TYPE_DEFAULT = (flag_typeInitializer | flag_noGlobalImport);
/*@}*/

/** @name Tag Registration 

	These functions are called to register a new tag.
*/
/*@{*/
/**
	@param namespaceName The namespace in which the tag will be registered.
	@param tagName The name which the tag can be called by.
	@param func The user supplied callback for the tag.
	@param flags Any special flags for the tag.
	@param description A description for the tag.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_registerTagModule( const char * namespaceName, const char * tagName, lasso_tag_func func, int flags, const char * description );
LCAPICALL osError lasso_registerTagModuleW( const UChar * namespaceName, const UChar * tagName, lasso_tag_func func, int flags, const UChar * description );
/*@}*/

/*@{*/
/**
	@param namespaceName The namespace in which the type will be registered.
	@param tagName The name which the type can be called by.
	@param func The user supplied callback for the type initializer.
	@param flags Any special flags for the type. If flag_typeInitializer is ommited, it will be automatically added. Any conflicting flags such as flag_typeContainerTag or flag_typeLoopingTag will be ignored.
	@param description A description for the type.
	@param superType Optionally, the name of the new type's super type.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_registerTypeModule( const char * namespaceName, const char * tagName, lasso_tag_func func, int flags, const char * description, const char * superType );
LCAPICALL osError lasso_registerTypeModuleW( const UChar * namespaceName, const UChar * tagName, lasso_tag_func func, int flags, const UChar * description, const UChar * superType );
/*@}*/

/** @name Datasource Module Registration 
	
	These functions are called to register a new datasource module.
*/
/*@{*/
/**
	@param moduleName The name of the datasource module.
	@param func The User supplied callback for the datasource module.
	@param flags Any special flags for the datasource module.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_registerDSModule( const char * moduleName, lasso_ds_func func, void * ud );
LCAPICALL osError lasso_registerDSModuleW( const UChar * moduleName, lasso_ds_func func, void * ud );
/*@}*/
/*********************************************************
	REQUEST PARAMS
**********************************************************/
/**
	@brief Fetches a request specific variable.
	
	Request specific variables are enumerated in RequestParams.h
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param key The key identifying the request variable to retrieve.
	@param result The resulting data for the requested variable. Some request variables require input data which is submited in this parameter.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getRequestParam( lasso_request_t token, RequestParamKeyword key, auto_lasso_value_t * result );
/**
	@brief Fetches a server specific variable.
	
	Request specific variables are enumerated in RequestParams.h
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param key The key identifying the server variable to retrieve.
	@param result The resulting data for the server variable.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getLassoParam( lasso_request_t token, LassoParamKeyword key, auto_lasso_value_t * result );
/*
	return osErrNoErr if Lasso IS LassoLite
*/
LCAPICALL osError lasso_isLassoLite( lasso_request_t token );

/*********************************************************
	ERROR/RESULT CODES
**********************************************************/
/**
	@brief Set result code number.
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param err The result code
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_setResultCode( lasso_request_t token, osError err );
/**	
	Set result message string.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param msg The result message string.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_setResultMessage( lasso_request_t token, const char * msg );
/**	
	Set result message string.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param msg The result message string.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_setResultMessageW( lasso_request_t token, const UChar * msg );


/*********************************************************
	LASSO TYPE FUNCTIONS - LCAPI 6

	LCAPI 6 gives programmers the ability to access 
	Lasso's internal type system from within their LCAPI 
	tags.
**********************************************************/
/**
	@brief Allocates a new type instance.
	
	The name of the type to allocate is signified by the second parameter. If a type initializer is found for the given name, it will be executed.
	An array of lasso_type_t parameters can be passed to the type initializer.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param typeName The name of the type to instantiate.
	@param paramCount The number of lasso_type_t parameters to pass to the initializer.
	@param paramsArray An array of parameters to pass to the type initializer.
	@param outType A pointer to the newly instantiated type.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAlloc ( lasso_request_t token, const char * typeName, int paramCount, lasso_type_t * paramsArray, lasso_type_t * outType );
/**
	@brief Allocates a new type instance.
	
	The name of the type to allocate is signified by the second parameter. If a type initializer is found for the given name, it will be executed.
	An array of lasso_type_t parameters can be passed to the type initializer.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param typeName The name of the type to instantiate.
	@param paramCount The number of lasso_type_t parameters to pass to the initializer.
	@param paramsArray An array of parameters to pass to the type initializer.
	@param outType A pointer to the newly instantiated type.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocW ( lasso_request_t token, const UChar * typeName, int paramCount, lasso_type_t * paramsArray, lasso_type_t * outType );
/**
	@brief Allocates a custom type within a type initializer.
	
	This function is used within lasso_tag_funcs that were registered as being a type initializer (flag_typeInitializer).
	It initializes a blank custom type and sets the type's name to the provided value.
	The new type does not yet have a lineage and has no members added to it.
	New data or tag members should be added using lasso_typeAddMember.
	The new type must be the return value of the tag call, set via lasso_returnTagValue.
	Any inherited members will be added to the type after the LCAPI call returns.
	
 	@warning Do NOT call this unless you are in a type initializer. If you are not in a type initializer, the result will be a type that will never be fully initialized.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outCustom The resulting newly allocated custom type.
	@param name The new type's name.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocCustom( lasso_request_t token, lasso_type_t * outCustom, const char * name );
/**
	@brief Allocates a custom type within a type initializer.
	
	This function is used within lasso_tag_funcs that were registered as being a type initializer (flag_typeInitializer).
	It initializes a blank custom type and sets the type's name to the provided value.
	The new type does not yet have a lineage and has no members added to it.
	New data or tag members should be added using lasso_typeAddMember.
	The new type must be the return value of the tag call, set via lasso_returnTagValue.
	Any inherited members will be added to the type after the LCAPI call returns.
	
 	@warning Do NOT call this unless you are in a type initializer. If you are not in a type initializer, the result will be a type that will never be fully initialized.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outCustom The resulting newly allocated custom type.
	@param name The new type's name.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocCustomW( lasso_request_t token, lasso_type_t * outCustom, const UChar * name );

/**
	@brief Allocate a new type instance based on the given type instance.
	
	The given type's tag members will be referenced in the new type.
	No data members are added to the new type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param proto The original type instance whose tag members will be referenced.
	@param out The resulting new type instance.	
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocFromProto( lasso_request_t token, lasso_type_t proto, lasso_type_t * out );

/**
	@brief Allocate a new type with the given name. 
	
	The type does not have to have been registered as a type initializer or registered at all.
	The new type will have no tag or data members, but those may be added using the appropriate LCAPI call at any time.
	If no parent type name is provided (a NULL pointer  or empty string is passed in), type null will be the default.
	If a parent type name is provided, it must have been a previously registered type initializer.
	onCreate will be called for the parent and any of its parents, but will not be called for the new type instance itself.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name the name for the new type instance. The name will be automatically tokenized to produce the namespace for the given typename.
	@param parentTypeName The name of the parent type.
	@param out The new type instance.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocOneOff( lasso_request_t token, const char * name, const char * parentTypeName, lasso_type_t * out );
/**
	@brief Allocate a new type with the given name. 
	
	The type does not have to have been registered as a type initializer or registered at all.
	The new type will have no tag or data members, but those may be added using the appropriate LCAPI call at any time.
	If no parent type name is provided (a NULL pointer  or empty string is passed in), type null will be the default.
	If a parent type name is provided, it must have been a previously registered type initializer.
	onCreate will be called for the parent and any of its parents, but will not be called for the new type instance itself.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name the name for the new type instance. The name will be automatically tokenized to produce the namespace for the given typename.
	@param parentTypeName The name of the parent type.
	@param out The new type instance.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocOneOffW( lasso_request_t token, const UChar * name, const UChar * parentTypeName, lasso_type_t * out );


// ************************************************************************ //
// ****************** I N S T A N C E  L I F E T I M E ******************** //
// ************************************************************************ //

/**
	@brief Add 1 to an instance's reference count.
	
	Type instances each have a reference count.
	Types created or otherwise referenced or received during an LCAPI call can be assumed to have a reference count of at least 1.
	If client code wishes to utilize any type instance beyond the end of the LCAPI call, it should add 1 to the instance's reference count.
	This is called "referring to" an instance.
	When client code has finished utilizing any "referred to" instance, it should subtract 1 from the instance's reference count using lasso_typeFree.
	When a type instance's reference count reaches zero, the instance is destroyed.
*/
LCAPICALL osError lasso_typeRefer( lasso_request_t token, lasso_type_t toRefer );
// synonymous with lasso_typeRefer. here for compatability
LCAPICALL osError lasso_typeDetach( lasso_request_t token, lasso_type_t toDetach );

/**
	@brief Subtract 1 from an instance's reference count

	Once the instance's reference count reaches zero, it is destroyed and the lasso_type_t is no longer valid.
	The lasso_request_t token may be NULL if it is not available for any reason.
	@warning ONLY call this if you have previously called lasso_typeRefer using the same type instance or the result may be a double free.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param inType The type instance which should be unreferred to.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeFree ( lasso_request_t token, lasso_type_t inType);

/**
	@brief Remove the LCAPI call's reference to a type instance.

	The LCAPI garbage collection system usually holds a reference to any used type instances.
	Calling this will remove that reference, which may result in the instance being destroyed at that point.
	Usage of a type pointer after calling this is not permitted unless the type has been "referred" to previously using lasso_typeRefer.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param inType The type instance to operate on.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeCollect( lasso_request_t token, lasso_type_t inType);

/**
	@name Allocating Built-in Type Instances
	
	The following functions allocate instances of specific built-in types. 
	The lasso_request_t token may be null. If it is null, the allocated types are created as "detached" and must be manually freed using lasso_typeFree.
*/
/** @{ */
/**
	Allocates a new instance of type null.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outNull The resulting new type instance.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocNull ( lasso_request_t token, lasso_type_t * outNull );
/**
	Allocates a new instance of type string.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outString The resulting new type instance.
	@param value The UTF-8 data from which to copy for the new string instance.
	@param length The length of the UTF-8 data in characters.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocString ( lasso_request_t token, lasso_type_t * outString, const char * value, int length );
/**
	Allocates a new instance of type string using the specified conversion method.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outString The resulting new type instance.
	@param value The source data from which to copy for the new string instance.
	@param length The length of the source data in bytes.
	@param conv The encoding of the source data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocStringConv ( lasso_request_t token, lasso_type_t * outString, const char * value, int length, const char * conv );
/**
	Allocates a new instance of type string.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outString The resulting new type instance.
	@param value The UTF-16 data from which to copy for the new string instance.
	@param length The length of the UTF-16 data in characters.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocStringW ( lasso_request_t token, lasso_type_t * outString, const UChar * value, int length );
/**
	Allocates a new instance of type integer.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outInteger The resulting new type instance.
	@param value The integer value which the new type instance will hold.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocInteger ( lasso_request_t token, lasso_type_t * outInteger, osInt64 value );
/**
	Allocates a new instance of type decimal.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outDecimal The resulting new type instance.
	@param value The double value which the new type instance will hold.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocDecimal ( lasso_request_t token, lasso_type_t * outDecimal, double value );
/**
	Allocates a new instance of type decimal.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outDecimal The resulting new type instance.
	@param value The double value which the new type instance will hold.
	@param precision The decimal precision that the new type instance will output to.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocDecimal2 ( lasso_request_t token, lasso_type_t * outDecimal, double value, int precision );
/**
	Allocates a new instance of type pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outPair The resulting new type instance.
	@param first The instance to use for the first member of the pair. A reference to the instance will be made.
	@param second The instance to use for the second member of the pair. A reference to the instance will be made.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocPair ( lasso_request_t token, lasso_type_t * outPair, lasso_type_t first, lasso_type_t second );
/**
	Allocates a new hard reference to a type instance. The new instance will point to the original.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outRef The resulting new type instance.
	@param referenced The instance to be referenced.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocReference( lasso_request_t token, lasso_type_t * outRef, lasso_type_t referenced );
/**
	Allocates a new instance of type tag.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outTag The resulting new type instance.
	@param nativeTagFunction The LCAPI lasso_tag_func which will be called when the tag is used.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocTag ( lasso_request_t token, lasso_type_t * outTag, lasso_tag_func nativeTagFunction );
/**
	Allocates a new instance of type tag from the given source text.
	
	If the source text is UTF-8, it must contain a BOM or it will be treated as the default platform encoding.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outTag The resulting new type instance.
	@param source The source text which will be compiled to become the body of the tag.
	@param length The length of the source text in characters.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocTagFromSource ( lasso_request_t token, lasso_type_t * outTag, const char * source, int length );
/**
	Allocates a new instance of type array.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outArray The resulting new type instance.
	@param count The number of new array elements to insert. The number of items in the \a elements parameter.
	@param elements A pointer to an array of lasso_type_t which will become the elements of the array. Each item will be referenced.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocArray ( lasso_request_t token, lasso_type_t * outArray, int count, lasso_type_t * elements );
/**
	Allocates a new instance of type map.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outMap The resulting new type instance.
	@param count The number of new map elements to insert. Half the number of items in the \a elements parameter.
	@param elements A pointer to an array of lasso_type_t which will become the elements of the map. Each even numbered item will be treated as the key for for following element.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocMap ( lasso_request_t token, lasso_type_t * outMap, int count, lasso_type_t * elements );
/**
	Allocates a new instance of type boolean.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outBool The resulting new type instance.
	@param inValue The boolean value which the new type instance will hold.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAllocBoolean( lasso_request_t token, lasso_type_t * outBool, bool inValue );
/** @} */

/**
	@name Getting or Setting Values of Built-in Type Instances
	
	The following functions get and set the data on a previously created built-in type instance.
	When getting a value, the source type instance will not be altered.
	When setting a value, the provided type instance is converted, if required.
	The lasso_request_t token may be null.
*/
/** @{ */
/**
	Retrieves character data from a type instance.
	
	If the type is a string instance, the data will be converted to UTF-8.
	If the type is a bytes instance, the data will be provided unaltered.
	For any other type, the result will be the same as converting the type into a string and returning the data as UTF-8.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance from which to retrieve the character data.
	@param val A pointer to a lasso_value_t struct in which to store the result. The data will be stored in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetString( lasso_request_t token, lasso_type_t type, auto_lasso_value_t * val );
/**
	Retrieves character data from a type instance using the specified conversion method.
	If the special conversion method of "BINARY" is used, and the source type is a string, the resulting data will be UTF-16 data.
	If the special conversion method of "BINARY" is used, and the source type is bytes, the resulting data will be provided unaltered.
	For any other type, if the special conversion method of "BINARY" is used, the result will be the same as converting the instance 
	to a string and retrieving the UTF-16 data.

	The remaining possible values for the conversion method are any of the character encoding methods supported by ICU or any of the converters 
	stored in the '__external_converters__' global variable.
	
	In all cases, the \a nameSize portion of the resulting value struct will be the number of bytes in the result data.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance from which to retrieve the character data.
	@param val A pointer to a lasso_value_t struct in which to store the result. The data will be stored in the \a name portion of the struct.
	@param conv The encoding method to use when transforming the Unicode string data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetStringConv( lasso_request_t token, lasso_type_t type, auto_lasso_value_t * val, const char * conv );
/**
	Retrieves Unicode character data from a type instance.
	
	For any type other than string, the result will be the same as converting the type into a string and returning the data as UTF-16.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance from which to retrieve the character data.
	@param val A pointer to a lasso_value_w_t struct in which to store the result. The data will be stored in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetStringW( lasso_request_t token, lasso_type_t type, auto_lasso_value_w_t * val );
/**
	Retrieves the integer value from a type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance from which to retieve the integer value.
	@param out A pointer to the resulting 64-bit integer.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetInteger( lasso_request_t token, lasso_type_t type, osInt64 * out );
/**
	Retrieves the decimal value from a type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance from which to retieve the decimal value.
	@param out A pointer to the resulting double.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetDecimal( lasso_request_t token, lasso_type_t type, double * out );
/**
	Retrieves the boolean value from a type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance from which to retieve the boolean value.
	@param out A pointer to the resulting boolean.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetBoolean( lasso_request_t token, lasso_type_t type, bool * out );
/**
	Converts the type instance into a string and sets the value.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance to set.
	@param val The source data in UTF-8 encoding.
	@param len The length of the source UTF-8 data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetString( lasso_request_t token, lasso_type_t type, const char * val, int len );
/**
	Converts the type instance into a string and sets the value.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance to set.
	@param val The source data.
	@param len The length of the source data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetStringW( lasso_request_t token, lasso_type_t type, const UChar * val, int len );
/**
	Converts the type instance using the specified source data and conversion method.
	If the special conversion method of "BINARY" is used, the type instance will be converted into a bytes type, the contents of which will be the unaltered source data.
	
	The remaining possible values for the conversion method are any of the character encoding methods supported by ICU or any of the converters 
	stored in the '__internal_converters__' global variable.
	
	If any of the built-in ICU conversion methods are used, the provided instance will be converted into a string type.
	If any other conversion methods are used, the resulting type of the provided instance can can vary.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance to be converted.
	@param val A pointer the source data to be used for the conversion.
	@param len The length of the source data in bytes.
	@param conv The conversion method to use when transforming type instance.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetStringConv( lasso_request_t token, lasso_type_t type, const char * val, int len, const char * conv );
/**
	Appends additional data to the string instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance which will be appended to.
	@param val The source data to append in UTF-8 encoding.
	@param len The number of bytes of source data to be appended.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAppendString( lasso_request_t token, lasso_type_t type, const char * val, int len );
/**
	Appends additional data to the string instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance which will be appended to.
	@param val The source data to append.
	@param len The number of characters of source data to be appended.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAppendStringW( lasso_request_t token, lasso_type_t type, const UChar * val, int len );
/**
	Sets the integer value of a type instance, converting the type if necessary.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance which will be set to the provided integer.
	@param val The integer value to use.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetInteger( lasso_request_t token, lasso_type_t type, osInt64 val );
/**
	Sets the decimal value of a type instance, converting the type if necessary.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance which will be set to the provided double.
	@param val The double value to use.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetDecimal( lasso_request_t token, lasso_type_t type, double val );
/**
	Sets the decimal value of a type instance, converting the type if necessary.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance which will be set to the provided integer.
	@param val The integer value to use.
	@param precision The decimal precision which the decimal instance will output to.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetDecimal2( lasso_request_t token, lasso_type_t type, double val, int precision );
/**
	Sets the boolean value of a type instance, converting the type if necessary.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance which will be set to the provided boolean.
	@param val The boolean value to use.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetBoolean( lasso_request_t token, lasso_type_t type, bool val );
/**
	Retrieves the size of the provided array instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param array The array instance.
	@param len A pointer in which to store the resulting size.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_arrayGetSize( lasso_request_t token, lasso_type_t array, int * len );
/**
	Retrieves the specified element of the provided array instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param array The array instance.
	@param index The zero based index at which to retreive.
	@param out The pointer in which to store the result.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_arrayGetElement( lasso_request_t token, lasso_type_t array, int index, lasso_type_t * out );
/**
	Sets the specified element of the provided array instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param array The array instance.
	@param index The zero based index at which to set. If the index is greater than the array's current size or is less than zero, the new item is added to the end of the array.
	@param elem The type instance which will be placed at the specified index. The instance will be referenced.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_arraySetElement(lasso_request_t token, lasso_type_t array, int index, lasso_type_t elem );
/**
	Removes the specified element from the provided array instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param array The array instance.
	@param index The zero based index at which to remove.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_arrayRemoveElement( lasso_request_t token, lasso_type_t array, int index );
/**
	Retrieves the size of the provided map instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param mp The map instance.
	@param len A pointer in which to store the resulting size.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_mapGetSize( lasso_request_t token, lasso_type_t mp, int * len );
/**
	Searches for the specified key in the provided map instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param mp The map instance.
	@param key The key to search for within the map.
	@param out The pointer in which to store the result.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_mapFindElement( lasso_request_t token, lasso_type_t mp, lasso_type_t key, lasso_type_t * out );
/**
	Retrieves the specified element of the provided map instance by index.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param mp The map instance.
	@param index The zero based index at which to retreive.
	@param outPair The pointer in which to store the result. The result will be a pair whose first member is the key and second member is the associated element.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_mapGetNthElement( lasso_request_t token, lasso_type_t mp, int index, lasso_type_t * outPair );
/**
	Sets the specified element of the provided map instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param mp The map instance.
	@param key The instance of the key which will be set.
	@param value The value of the element to associate with the provided key.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_mapSetElement( lasso_request_t token, lasso_type_t mp, lasso_type_t key, lasso_type_t value );
/**
	Sets the specified element of the provided map instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param mp The map instance.
	@param key The instance of the key which will be set.
	@param value The value of the element to associate with the provided key.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_mapSetElementW( lasso_request_t token, lasso_type_t mp, const UChar * key, lasso_type_t value );
/**
	Removes the specified element from the map based on the provided key.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param mp The map instance.
	@param key The key to be removed.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_mapRemoveElement( lasso_request_t token, lasso_type_t mp, lasso_type_t key );
/**
	Retrieves the first member from the provided pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param pr The pair instance.
	@param out The pointer in which to store the result.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_pairGetFirst( lasso_request_t token, lasso_type_t pr, lasso_type_t * out );
/**
	Retrieves the second member from the provided pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param pr The pair instance.
	@param out The pointer in which to store the result.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_pairGetSecond( lasso_request_t token, lasso_type_t pr, lasso_type_t * out );
/**
	Sets the first member of the provided pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param pr The pair instance.
	@param frst The instance to use as the first member. The instance will be referenced.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_pairSetFirst( lasso_request_t token, lasso_type_t pr, lasso_type_t frst );
/**
	Sets the second member of the provided pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param pr The pair instance.
	@param scnd The instance to use as the second member. The instance will be referenced.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_pairSetSecond( lasso_request_t token, lasso_type_t pr, lasso_type_t scnd );
/** @} */

// Custom types can have a pointer value associated with them.
// The value can be used for whatever purpose the programmer needs.
//LCAPICALL osError lasso_typeSetCustomPtr( lasso_type_t the_custom_type, void * the_ptr );
//LCAPICALL osError lasso_typeGetCustomPtr( lasso_type_t the_custom_type, void ** the_ptr );

/*
 * Twiddle...
 */

/**
	@brief Convert a type instance into byte data.
	
	Converts a Lasso type INTO single-byte or binary data using the specific encoding method.
	This is the basis of table-level automatic encoding and translation.
	The default for all database, column, table names should be "iso8859-1".
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param instance The instance which will be converted.
	@param encoding The encoding conversion method.
	@param outVal A pointer to a lasso_value_t struct in which to store the result. The result will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_handleExternalConversion(lasso_request_t token, lasso_type_t instance, const char * encoding, auto_lasso_value_t * outVal);
/**
	@brief Converts byte data into a type instance.

	Converts a single-byte or binary representation of a Lasso type BACK INTO an instance of that type using the specified encoding method.
			
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param src The source data.
	@param srcLen The number of bytes in the source data.
	@param encoding The encoding conversion method.
	@param closestLassoType The type code for the closest applicable Lasso type. This is used as a hint to the converter.
	@param outType The resulting Lasso type instance.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_handleInternalConversion(lasso_request_t token, const char * src, unsigned int srcLen, const char * encoding, LP_TypeDesc closestLassoType, lasso_type_t * outType);
/**
	@brief Changes the inheritence structure of a type.
	
	Sets newParent to be the new parent of child
	Any parent that child currently has will be destroyed
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param child The child instance which will have its inheritance changed.
	@param newParent The new parent for the child.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeInheritFrom( lasso_request_t token, lasso_type_t child, lasso_type_t newParent );
/**
	@brief Retrieves the parent of the provided type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param from The child instance from which to retrieve the parent.
	@param outParent A pointer in which to return the parent.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_typeGetParent( lasso_request_t token, lasso_type_t from, lasso_type_t * outParent );
/**
	@brief Retrieves the child of the provided instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param from The parent instance from which to retrieve the child.
	@param outChild A pointer in which to return the child.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetChild( lasso_request_t token, lasso_type_t from, lasso_type_t * outChild );

/**
	@brief Adds a C function to be called after onDestroy.
	
	Adds a user provided C function which will be called after onDestroy has been invoked on the instance.
	
	@param the_custom_type The custom type instance which will have the C function added.
	@param dtor The C function to add. The C function must accept one parameter, which is the type instance in question.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetCustomDtor( lasso_type_t the_custom_type, void (*dtor)(lasso_type_t the_custom_type) );
/**
	@brief Adds a member to a type instance.
	
	If the new member is a tag, it will be added to the tag members for the type. Otherwise, the new member will be added as a data member.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param to The type instance to which the new member will be added.
	@param named The name for the new member.
	@param member The new member to add.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAddMember( lasso_request_t token, lasso_type_t to, const char * named, lasso_type_t member );
/**
	@brief Adds a member to a type instance.
	
	If the new member is a tag, it will be added to the tag members for the type. Otherwise, the new member will be added as a data member.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param to The type instance to which the new member will be added.
	@param named The name for the new member.
	@param member The new member to add.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAddMemberW( lasso_request_t token, lasso_type_t to, const UChar * named, lasso_type_t member );
/**
	@brief Adds a tag member to a type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param to The type instance to which the new tag member will be added.
	@param named The name for the new member.
	@param member The new tag member to add.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAddTagMember( lasso_request_t token, lasso_type_t to, const char * named, lasso_type_t member );
/**
	@brief Adds a data member to a type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param to The type instance to which the new data member will be added.
	@param named The name for the new member.
	@param member The new data member to add.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAddDataMember( lasso_request_t token, lasso_type_t to, const char * named, lasso_type_t member );
/**
	@brief Adds a tag member to a type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param to The type instance to which the new tag member will be added.
	@param named The name for the new member.
	@param member The new tag member to add.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAddTagMemberW( lasso_request_t token, lasso_type_t to, const UChar * named, lasso_type_t member );
/**
	@brief Adds a data member to a type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param to The type instance to which the new data member will be added.
	@param named The name for the new member.
	@param member The new data member to add.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAddDataMemberW( lasso_request_t token, lasso_type_t to, const UChar * named, lasso_type_t member );
/**
	@brief Retrieves a member from a type instance.
	
	Members are searched for by name with tag members being searched first.
	Data members are searched if no tag is found with the given name.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param from The type instance in which to search.
	@param named The name of the member to look for.
	@param out A pointer to a type instance in which to store the found member.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetMember( lasso_request_t token, lasso_type_t from, const char * named, lasso_type_t * out );
/**
	@brief Retrieves a member from a type instance.
	
	Members are searched for by name with tag members being searched first.
	Data members are searched if no tag is found with the given name.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param from The type instance in which to search.
	@param named The name of the member to look for.
	@param out A pointer to a type instance in which to store the found member.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeGetMemberW( lasso_request_t token, lasso_type_t from, const UChar * named, lasso_type_t * out );
/**
	@brief Allows storage of an opaque pointer value.
	
	The pointer member is given and name and is stored as an integer in the type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param self The type instance in which to add the pointer member.
	@param name The name for the new member.
	@param data The pointer value which will be added.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_setPtrMember(lasso_request_t token, lasso_type_t self, const char * name, void * data);
/**
	@brief Allows storage of an opaque pointer value.
	
	The pointer member is given and name and is stored as an integer in the type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param self The type instance in which to add the pointer member.
	@param name The name for the new member.
	@param data The pointer value which will be added.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_setPtrMemberW(lasso_request_t token, lasso_type_t self, const UChar * name, void * data);
/**
	@brief Retrieves a previously added pointer value.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param self The type instance from which to retrieve.
	@param name The name of the previously added pointer value.
	@param data A pointer to a void * in which to store the found value.s
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getPtrMember(lasso_request_t token, lasso_type_t self, const char * name, void ** data);
/**
	@brief Retrieves a previously added pointer value.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param self The type instance from which to retrieve.
	@param name The name of the previously added pointer value.
	@param data A pointer to a void * in which to store the found value.s
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getPtrMemberW(lasso_request_t token, lasso_type_t self, const UChar * name, void ** data);
/**
	@brief Sets the flag for the given type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance on which to set the flag.
	@param flag The flag to set.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeSetFlag(lasso_request_t token, lasso_type_t type, int flag);
/**
	@brief Tests the flag for the given type instance.
	
	Returns osErrNoErr if the flag is set for that type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance to be tested.
	@param flag The flag to test for.
	@return If the flag is set, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeHasFlag(lasso_request_t token, lasso_type_t type, int flag);
/**
	@brief Removes the flag for the given type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param type The type instance.
	@param flag The flag to remove from the type instance.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeClearFlag(lasso_request_t token, lasso_type_t type, int flag);

// lasso_typeGetProperties has two uses. If the targetType parameter is non-null, it is used to get all the data and tag members from a given type. 
// They are returned as a pair of arrays in the outPair value.
// The first element of each pair is the map of data members for the type. The second element is the map of tag members.
// Each element in the array represents the members of each type the targetType inherits from (its parent)
//
// If the targetType parameter is NULL, lasso_typeGetProperties will return an array containing the variable maps for the currently active request.
LCAPICALL osError lasso_typeGetProperties ( lasso_request_t token, lasso_type_t targetType, lasso_type_t * outPair );
/**
	@brief Returns the type instance that the current tag call was a member of.
	
	This is used in member tags of custom types to return the target of the current call.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param self A pointer in which to store the resulting type instance.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagSelf ( lasso_request_t token, lasso_type_t * self );
/**
	@brief Returns osErrNoErr if the calling tag has the same 'self' as the current tag.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@return Returns osErrNoErr if the calling tag has the same 'self' as the current tag.
*/
LCAPICALL osError lasso_tagIsSelfCall(lasso_request_t token);
/**
	@brief Returns the name of the target type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param target The type instance to get the name of.
	@param outName The resulting name which will be stored in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/	
LCAPICALL osError lasso_typeGetName( lasso_request_t token, lasso_type_t target, auto_lasso_value_t * outName );
/**
	@brief Returns the name of the target type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param target The type instance to get the name of.
	@param outName The resulting name which will be stored in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/	
LCAPICALL osError lasso_typeGetNameW( lasso_request_t token, lasso_type_t target, auto_lasso_value_w_t * outName );
/**
	@brief Returns the namespace of the target type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param target The type instance to get the namespace of.
	@param outName The resulting name which will be stored in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/	
LCAPICALL osError lasso_typeGetNamespace( lasso_request_t token, lasso_type_t target, auto_lasso_value_t * outName );
/**
	@brief Returns the namespace of the target type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param target The type instance to get the namespace of.
	@param outName The resulting name which will be stored in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/	
LCAPICALL osError lasso_typeGetNamespaceW( lasso_request_t token, lasso_type_t target, auto_lasso_value_w_t * outName );
/**
	@brief Tests to see if a type is an instance of another type.
	
	Test by type code.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param target The target instance to test.
	@param type The type code to test for.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeIsA( lasso_request_t token, lasso_type_t target, LP_TypeDesc type );
/**
	@brief Tests to see if a type is an instance of another type.
	
	Test by type name.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param target The target instance to test.
	@param typeName The type name to test for.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeIsA2( lasso_request_t token, lasso_type_t target, const char * typeName );
/**
	@brief Tests to see if a type is an instance of another type.
	
	Test by type name.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param target The target instance to test.
	@param typeName The type name to test for.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeIsA2W( lasso_request_t token, lasso_type_t target, const UChar * typeName );
/**
	@brief Tests to see if a type is an instance of another type.
	
	Test based on another type instance.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param target The target instance to test.
	@param type The type name to test for.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeIsA3( lasso_request_t token, lasso_type_t target, lasso_type_t type );

/**
	@brief Specifies the return value of the current LCAPI tag call.

	Any type can be returned. A tag can only have one return value. Setting another return value will overwrite the previous.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param value The value to return.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_returnTagValue( lasso_request_t token, lasso_type_t value );
/**
	@brief Return a boolean value from the current LCAPI tag call.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param b The boolean value to return.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_returnTagValueBoolean( lasso_request_t token, bool b );
/**
	@brief Return an integer value from the current LCAPI tag call.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param i The integer value to return.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_returnTagValueInteger( lasso_request_t token, osInt64 i );
/**
	@brief Return a string value from the current LCAPI tag call.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param p The character data to be returned. The data should be in UTF-8 encoding.
	@param l The length of the character data to return.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_returnTagValueString( lasso_request_t token, const char * p, int l );
/**
	@brief Return a string value from the current LCAPI tag call.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param p The character data to be returned.
	@param l The length of the character data to return.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_returnTagValueStringW( lasso_request_t token, const UChar * p, int l );
/**
	@brief Return a decimal value from the current LCAPI tag call.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param d The double value to return.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_returnTagValueDecimal( lasso_request_t token, double d );
/**
	@brief Return binary data from the current LCAPI tag call.
	
	When using this, the result of the LCAPI call will be a bytes type.
	This can be called as many times as needed and new data will be appended to any previous data.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param data The binary data to return.
	@param length The number of bytes to return.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_returnTagValueBytes( lasso_request_t token, const char * data, int length );
/* Synonym for lasso_returnTagValueBytes */
LCAPICALL osError lasso_outputTagBytes( lasso_request_t token, const char * data, int length );
/**
	@brief Provides direct access to the LCAPI tag's return value.
	
	In many cases, directly manipulating this value can yield better performance than lasso_returnTagValue.
	The resulting type instance will start out as type null.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param outValue A pointer inwhich to place the resulting return value type instance.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagReturnValue( lasso_request_t token, lasso_type_t * outValue );

/**
	@brief Provides access to a defined variable.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param key The name of the variable to access.
	@param value The resulting variable value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getVariable2( lasso_request_t token, const char * key, lasso_type_t * value );
/**
	@brief Provides access to a defined variable.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param key The name of the variable to access.
	@param value The resulting variable value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getVariable2W( lasso_request_t token, const UChar * key, lasso_type_t * value );
/**
	@brief Sets the specified variable's value.
	
	The variable does not have to already exist.
		
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param key The name of the variable to set.
	@param value The value to set. 
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_setVariable2( lasso_request_t token, const char * key, lasso_type_t value );
/**
	@brief Sets the specified variable's value.
	
	The variable does not have to already exist.
		
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param key The name of the variable to set.
	@param value The value to set. 
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_setVariable2W( lasso_request_t token, const UChar * key, lasso_type_t value );

/**
	@brief Retrieves a parameter that was passed to the LCAPI tag call.

	The parameter index is zero based.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramIndex The zero based index of the desired parameter.
	@param result The resulting parameter value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getTagParam2( lasso_request_t token, int paramIndex, lasso_type_t * result );

/**
	@brief Retrieves a parameter that was passed to the LCAPI tag call.

	The parameter must have been specified by keyword.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramName The keyword of the desired parameter.
	@param result The resulting parameter value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_findTagParam2( lasso_request_t token, const char * paramName, lasso_type_t * result );
/**
	@brief Retrieves a parameter that was passed to the LCAPI tag call.

	The parameter must have been specified by keyword.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramName The keyword of the desired parameter.
	@param result The resulting parameter value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_findTagParam2W( lasso_request_t token, const UChar * paramName, lasso_type_t * result );

/**
	@brief Execute the contents of a container tag.
	
	Tags become containers by registering with the flag_typeContainerTag flag.
	The resulting data will be of type string.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result The resulting data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_childrenRun( lasso_request_t token, lasso_type_t * result );

/**
	@brief Execute a given tag.
	
	The tag can be run given a specific name and an array of parameters. 
	The tag's return value can be accessed.
	If the tag is a member tag, the instance of which it is a member can be passed using the final parameter.
	The params, returnValue and optionalTarget parameters may all be NULL.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name The name for the tag.
	@param tagType The tag instance to be executed.
	@param paramCount The number of parameters passed in the \a params array.
	@param params The array of parameters for the tag. May be NULL.
	@param returnValue A pointer in which to store the return value.
	@param optionalTarget The target instance for a member tag call. May be NULL if the tag is not a member tag.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeRunTag ( lasso_request_t token, const char * name, lasso_type_t tagType, int paramCount, lasso_type_t * params, lasso_type_t * returnValue, lasso_type_t optionalTarget );
/**
	@brief Execute a given tag.
	
	The tag can be run given a specific name and an array of parameters. 
	The tag's return value can be accessed.
	If the tag is a member tag, the instance of which it is a member can be provided using the final parameter.
	The params, returnValue and optionalTarget parameters may all be NULL.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name The name for the tag.
	@param tagType The tag instance to be executed.
	@param paramCount The number of parameters passed in the \a params array.
	@param params The array of parameters for the tag. May be NULL.
	@param returnValue A pointer in which to store the return value.
	@param optionalTarget The target instance for a member tag call. May be NULL if the tag is not a member tag.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeRunTagW ( lasso_request_t token, const UChar * name, lasso_type_t tagType, int paramCount, lasso_type_t * params, lasso_type_t * returnValue, lasso_type_t optionalTarget );
/**
	@brief Searches for a tag.
	
	The search method is the same as would be used to find a tag in a script. Returns osErrNoErr if the tag is found.
	The final parameter can be used to disable searching for the tag in any on-demand libraries.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name The tag name to search for.
	@param outTag The resulting tag, if found.
	@param disableOnDemand Used to disable the searching of on-demand libraries.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_findTagW( lasso_request_t token, const UChar* name, lasso_type_t * outTag, bool disableOnDemand );
/**
	@brief Searches for a tag.
	
	The search method is the same as would be used to find a tag in a script. Returns osErrNoErr if the tag is found.
	The final parameter can be used to disable searching for the tag in any on-demand libraries.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name The tag name to search for.
	@param outTag The resulting tag, if found.
	@param disableOnDemand Used to disable the searching of on-demand libraries.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_findTag( lasso_request_t token, const char * name, lasso_type_t * outTag, bool disableOnDemand );

// Set, Get and Clear various flags belonging to a tag instance
LCAPICALL osError lasso_tagSetFlag( lasso_request_t token, lasso_type_t tagType, int flag );
LCAPICALL osError lasso_tagHasFlag( lasso_request_t token, lasso_type_t tagType, int flag );
LCAPICALL osError lasso_tagClearFlag( lasso_request_t token, lasso_type_t tagType, int flag );

// functions related to querying a tag's parameter and return value types
LCAPICALL osError lasso_tagGetReturnType(lasso_request_t token, lasso_type_t tagType, auto_lasso_value_t * result);
LCAPICALL osError lasso_tagGetReturnTypeW(lasso_request_t token, lasso_type_t tagType, auto_lasso_value_w_t * result);

LCAPICALL osError lasso_tagSetReturnType(lasso_request_t token, lasso_type_t tagType, const char * typeName);
LCAPICALL osError lasso_tagSetReturnTypeW(lasso_request_t token, lasso_type_t tagType, const UChar * typeName);

LCAPICALL osError lasso_tagGetExpectedParamCount(lasso_request_t token, lasso_type_t tagType, int * count);
LCAPICALL osError lasso_tagGetExpectedParam(lasso_request_t token, lasso_type_t tagType, int index, auto_lasso_value_t * result, bool * required);
LCAPICALL osError lasso_tagGetExpectedParamW(lasso_request_t token, lasso_type_t tagType, int index, auto_lasso_value_w_t * result, bool * required);

/**
	@brief Performs an assignment of one type to another.
	
	The result will be the same as if the following had been executed in LDML: #left_hand_side = #right_hand_side
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param left_hand_side The left hand side of the operation.
	@param right_hand_side The right hand side of the operation.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeAssign( lasso_request_t token, lasso_type_t left_hand_side, lasso_type_t right_hand_side );
/**
	@brief Transfer the data from one type to another type.
	
	After the call, victim will still be valid, but will be of type null, while the thief will contain all the data previously contained by the victim.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param thief The thief which will take the value of the victim.
	@param victim The victim which will lose its value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_typeStealValue( lasso_request_t token, lasso_type_t thief, lasso_type_t victim );

/**
	@brief Creates and runs a new LCAPI call on the given lasso_tag_func. 

	If there is already an active request on the current thread, the lasso_tag_func will be run within that thread's context. 
	If there is no currently active request on the current thread, a new request will be created and run based off the global context.
	The tag_action_t parameter will be passed to the lasso_tag_func and can be used to signal or pass information to the function.
	
	@param func The lasso_tag_func which will be called.
	@param action User data to be passed to the tag.
	@param unused unused.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_runRequest( lasso_tag_func func, tag_action_t action, int unused );

/**
	@brief Register a constant value.
	
	Constants can be called just like tags, but the resulting value will only have a single instance and will have both it's type and value frozen.
	This is usually called at the same time that tags are registered.
	
	@param namespaceName The namespace for the constant.
	@param name The name for the constant.
	@param val The value for the constant.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_registerConstant2( const char * namespaceName, const char * name, lasso_type_t val );
/**
	@brief Register a constant value.
	
	Constants can be called just like tags, but the resulting value will only have a single instance and will have both it's type and value frozen.
	This is usually called at the same time that tags are registered.
	
	@param namespaceName The namespace for the constant.
	@param name The name for the constant.
	@param val The value for the constant.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_registerConstant2W( const UChar * namespaceName, const UChar * name, lasso_type_t val );
/**
	@brief Register a constant value.
	
	Constants can be called just like tags, but the resulting value will only have a single instance and will have both it's type and value frozen.
	This is usually called at the same time that tags are registered.
	
	@param name The name for the constant.
	@param val The value for the constant.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_registerConstant( const char * name, lasso_type_t val );
/**
	@brief Register a constant value.
	
	Constants can be called just like tags, but the resulting value will only have a single instance and will have both it's type and value frozen.
	This is usually called at the same time that tags are registered.
	
	@param name The name for the constant.
	@param val The value for the constant.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_registerConstantW( const UChar * name, lasso_type_t val );

/**
	@brief Works just like the include_raw tag.
	
	Includes file data based on the given path. The path is resolved based on the current request. 
	Full paths, or paths outside of the request's webserver folder are not permitted.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param path The path to the file.
	@param out The resulting data which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_includeRaw( lasso_request_t token, const char * path, auto_lasso_value_t * out );

/**
	@brief Works just like the include_raw tag.
	
	Returns file data based on the given path. The path is resolved based on the current request. 
	Full paths, or paths outside of the request's webserver folder are not permitted.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param path The path to the file in UTF8.
	@param out The resulting data.
	@param modTime The modification time of the document. NULL is permitted if this value is not required.
	@param ifModifiedSince If the document was not modified since this date, osErrPreconditionFailed is returned and no document content is provided.
	@param unused Pass NULL for this value.
	@param flags Flags to control the behavior of the document search
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_includeRaw2( lasso_request_t token, const char * path, lasso_type_t * out, time_t * modTime, time_t * ifModifiedSince, void ** unused, int flags );

/**
	@brief Changes the given type into a hard reference.
*/
#define MAKE_REF(X) lasso_typeAllocReference( token, &X, X )

/*********************************************************
	TAG MODULE FUNCTIONS
**********************************************************/
/**
	@brief Fetches the tag instance that triggered this call.
		
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result The resulting tag, or NULL if there is no current tag (such as while within a datasource).
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTag( lasso_request_t token, lasso_type_t * result );
/**
	@brief Fetches the name of the tag that triggered this call.
	
	For example, in the case of: [my_tag: ...] the resulting value would be "tag".
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result The resulting tag name which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagName( lasso_request_t token, auto_lasso_value_t * result);
/**
	@brief Fetches the name of the tag that triggered this call.
	
	For example, in the case of: [my_tag: ...] the resulting value would be "tag".
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result The resulting tag name which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagNameW( lasso_request_t token, auto_lasso_value_w_t * result);

/**
	@brief Fetches the namespace name of the tag that triggered this call.
	
	For example, in the case of: [my_tag: ...] the resulting value would be "my_". A tag's namespace is persistent and will remain the same even if the tag is subsequently imported into another namespace.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result The resulting tag namespace which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagNamespace( lasso_request_t token, auto_lasso_value_t * result);
/**
	@brief Fetches the namespace of the tag that triggered this call.
	
	For example, in the case of: [my_tag: ...] the resulting value would be "my_". A tag's namespace is persistent and will remain the same even if the tag is subsequently imported into another namespace.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result The resulting tag namespace which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagNamespaceW( lasso_request_t token, auto_lasso_value_w_t * result);
/**
	@brief Fetches the number of parameters that were passed to the tag.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result A pointer in which to store the result.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagParamCount( lasso_request_t token, int * result );
/**
	@brief Fetches the indicated parameter that was sent to the tag.
	
	Parameter indexes start at zero. If the parameter was specified as a keyword/value pair, the keyword will be placed in the \a name portion of the struct and the value in the \a data portion.
	If the parameter was provided without a keyword, the \a name portion of the struct will be NULL.
	
	All parameters values will be converted to string, regardless of the original type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramIndex The zero based index of the desired parameter.
	@param result The resulting parameter data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagParam( lasso_request_t token, int paramIndex, auto_lasso_value_t * result );
/**
	@brief Fetches the indicated parameter that was sent to the tag.
	
	Parameter indexes start at zero. If the parameter was specified as a keyword/value pair, the keyword will be placed in the \a name portion of the struct and the value in the \a data portion.
	If the parameter was provided without a keyword, the \a name portion of the struct will be NULL.
	
	All parameters values will be converted to string, regardless of the original type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramIndex The zero based index of the desired parameter.
	@param result The resulting parameter data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTagParamW( lasso_request_t token, int paramIndex, auto_lasso_value_w_t * result );
/**
	@brief Returns osErrNoErr if the param WAS defined.
	
	Any other result means it wasn't defined. Only parameters specified with a keyword should be searched for.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramName The name of the keyworded parameter to search for.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_tagParamIsDefined( lasso_request_t token, const char * paramName );
/**
	@brief Returns osErrNoErr if the param WAS defined.
	
	Any other result means it wasn't defined. Only parameters specified with a keyword should be searched for.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramName The name of the keyworded parameter to search for.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_tagParamIsDefinedW( lasso_request_t token, const UChar * paramName );
/**
	@brief Retrieves a parameter that was passed to the LCAPI tag call.

	The parameter must have been specified by keyword. 
	The keyword will be placed in the \a name portion of the struct and the value in the \a data portion. 
	If only a keyword was provided, the \a data portion will be NULL.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramName The keyword of the desired parameter.
	@param result The resulting parameter value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_findTagParam( lasso_request_t token, const char * paramName, auto_lasso_value_t * result );
/**
	@brief Retrieves a parameter that was passed to the LCAPI tag call.

	The parameter must have been specified by keyword. 
	The keyword will be placed in the \a name portion of the struct and the value in the \a data portion. 
	If only a keyword was provided, the \a data portion will be NULL.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param paramName The keyword of the desired parameter.
	@param result The resulting parameter value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_findTagParamW( lasso_request_t token, const UChar * paramName, auto_lasso_value_w_t * result );

// flags for param sorting
const int FLAG_PARAM_REQUIRED = 0x01;


struct lasso_param_desc_t
{
	const UChar * name;
	const UChar * type;
	int flags;
};

// terminate the names list with a NULL ->name member
// sorted will have the same number of elements as names did.
// NULL items will indicate that the items were not provided by the caller
// extra *may* be allocated and will contain all the parameters that were given after the req/opt requirements have been met
// both resulting arrays will be deleted by Lasso after the call returns
osError lasso_tagSortNamedParams( lasso_request_t token, const lasso_param_desc_t names[], lasso_type_t ** sorted, lasso_type_t ** extra );

/*
	Deprecated.
*/
LCAPICALL osError lasso_outputTagData2( lasso_request_t token, const char * data, int length );
LCAPICALL osError lasso_outputTagData( lasso_request_t token, const char * data );
LCAPICALL osError lasso_outputTagData2W( lasso_request_t token, const UChar * data, int length );
LCAPICALL osError lasso_outputTagDataF( lasso_request_t token, const char * format, ... );
/**
	@brief Allows access to the result (HTTP) header.

	The data will be a copy of the current HTTP header data.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result The header data which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getResultHeader( lasso_request_t token, auto_lasso_value_t * result );
/**
	@brief Allows access to the result (HTTP) header.

	The data will be a copy of the current HTTP header data.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param result The header data which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getResultHeaderW( lasso_request_t token, auto_lasso_value_w_t * result );
/**
	@brief Sets the result (HTTP) header.
	
	The data will be set as-is, so make sure it is properly formatted. The data should not include the final trailing CRLF pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param header The header data to use.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_setResultHeader( lasso_request_t token, const char * header );
/**
	@brief Sets the result (HTTP) header.
	
	The data will be set as-is, so make sure it is properly formatted. The data should not include the final trailing CRLF pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param header The header data to use.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_setResultHeaderW( lasso_request_t token, const UChar * header );
/*
	@brief Appends the supplied data to the header.
	
	The data will be set as-is, so make sure it is properly formatted. The data should not include the final trailing CRLF pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param header The header data to append.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_addResultHeader( lasso_request_t token, const char * data );
/*
	@brief Appends the supplied data to the header.
	
	The data will be set as-is, so make sure it is properly formatted. The data should not include the final trailing CRLF pair.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param header The header data to use.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_addResultHeaderW( lasso_request_t token, const UChar * data );
/**
	@brief Retrieves a cookie value set by the client.
		
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param named The name of the cookie to look for.
	@param value The resulting cookie value which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getCookieValue( lasso_request_t token, const char * named, auto_lasso_value_t * value );
/**
	@brief Retrieves a cookie value set by the client.
		
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param named The name of the cookie to look for.
	@param value The resulting cookie value which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getCookieValueW( lasso_request_t token, const UChar * named, auto_lasso_value_w_t * value );

/* The following are all deprecated. */
LCAPICALL osError lasso_getVariableCount( lasso_request_t token, const char * named, int * count );
LCAPICALL osError lasso_getVariableCountW( lasso_request_t token, const UChar * named, int * count );
LCAPICALL osError lasso_getVariable( lasso_request_t token, const char * named, int index, auto_lasso_value_t * value );
LCAPICALL osError lasso_getVariableW( lasso_request_t token, const UChar * named, int index, auto_lasso_value_w_t * value );
LCAPICALL osError lasso_setVariable( lasso_request_t token, const char * named, const char * value, int index );
LCAPICALL osError lasso_setVariableW( lasso_request_t token, const UChar * named, const UChar * value, int index );
LCAPICALL osError lasso_removeVariable( lasso_request_t token, const char * named, int index );
LCAPICALL osError lasso_removeVariableW( lasso_request_t token, const UChar * named, int index );

/**
	@name Datasource Module Functions
*/
/** @{ */
/**
	@brief Adds a datasource result value.
	
	Datasource actions which require returning multiple values can use this to add those values.
	For example, this call can be used to add the name of a datasource that the module services or the names of the tables in a particular datasource.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param data The data to add. Data must be in ISO8859-1 encoding.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_addDataSourceResult( lasso_request_t token, const char * data );
/**
	@brief Adds a datasource result value.
	
	Datasource actions which require returning multiple values can use this to add those values.
	For example, this call can be used to add the name of a datasource that the module services or the names of the tables in a particular datasource.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param data The data to add. Data must be in UTF-8 encoding.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_addDataSourceResultUTF8( lasso_request_t token, const char * data );
/**
	@brief Gets the currently specified database name and associated data.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param t The resulting database name.
	@param useHostDefault The setting which specifies whether the hosts username/password combo should be used in the absence of a database specific combo.
	@param usernamepassword The username/password combo for the database.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getDataSourceName( lasso_request_t token, auto_lasso_value_t * t, bool * useHostDefault, auto_lasso_value_t * usernamepassword );
/**
	@brief Gets the currently specified table name.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param t The resulting table name which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTableName( lasso_request_t token, auto_lasso_value_t * t );
/**
	@brief Gets the currently specified table encoding. This is the encoding as set in SiteAdmin for the current table.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param t The resulting table encoding which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getTableEncoding( lasso_request_t token, auto_lasso_value_t * t );
/**
	@brief Gets the currently specified schema name.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param schema The resulting schema name which will be placed in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getSchemaName( lasso_request_t token, auto_lasso_value_t * schema);
/**
	@brief Returns the host that maintains the current database.
	
	The host name will be placed in the \a name portion of the struct while the port will be placed in the \a data portion.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param host The resulting host data.
	@param usernamepassword The resulting username/password combo.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getDataHost( lasso_request_t token, auto_lasso_value_t * host, auto_lasso_value_t * usernamepassword );
/**
	@brief Returns the host that maintains the current database.
	
	The host name will be placed in the \a name portion of the struct while the port will be placed in the \a data portion.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param host The resulting host data.
	@param defaultSchema The resulting default schema name which will be placed in the \a name portion of the struct.
	@param usernamepassword The resulting username/password combo.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getDataHost2( lasso_request_t token, auto_lasso_value_t * host, auto_lasso_value_t * defaultSchema, auto_lasso_value_t * usernamepassword );
/**
	@brief Returns the "extra" information associated with the current host.
	
	The data, if any, will be placed in the \a name portion of the struct.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param data The resulting extra  host data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getDataHostExtra( lasso_request_t token, auto_lasso_value_t * data );
/**
	@brief Indicates if the host is dynamic.
		
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param wasDyn Will be set to true if the host was dynamic and false if it was not.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getDataHostIsDynamic( lasso_request_t token, bool * wasDyn );

/**
	@brief The number of rows that should be skipped in the found set.
	
	-skiprecords
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param recs The resulting -skiprecords value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getSkipRows( lasso_request_t token, int * recs );
/**
	@brief The maximum number of rows in the found set to return.
	
	-maxrecords
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param recs The resulting -maxrecords value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getMaxRows( lasso_request_t token, int * recs );
/**
	@brief Gets the currently specified record id (FileMaker specific).
	
	-keyvalue
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param id The resulting numeric -keyvalue value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getRowID( lasso_request_t token, osInt64 * id );
/**
	@brief Sets the currently specified record id (FileMaker specific).
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param id The numeric row id.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_setRowID( lasso_request_t token, osInt64 id );
/**
	@brief Gets the name and the value of the currently specified primary key column.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param v The resulting key name and value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getPrimaryKeyColumn( lasso_request_t token, auto_lasso_value_t * v );
/**
	@brief Gets the name and the value of the primary key column specified by index.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param index The zero based index.
	@param v The resulting key name and value.
	@param desc A pointer to a LP_TypeDesc in which to place the original type of the input data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getPrimaryKeyColumn2( lasso_request_t token, int index, auto_lasso_value_t * v, LP_TypeDesc * desc );
/**
	@brief Gets the name and the value of the primary key column specified by index, preserving the original type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param index The zero based index.
	@param keyName The name of the key field will be stored in the \a name member.
	@param type The value portion of the key field in its original type.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getPrimaryKeyColumn3( lasso_request_t token, int index, auto_lasso_value_t * keyName, lasso_type_t * type );
/**
	@brief Gets the number of key columns.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param index A pointer to an int in which the number of key columns will be placed.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getPrimaryKeyColumnCount( lasso_request_t token, int * count );
/**
	@brief Gets the number of input columns for this database action.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param count The resulting number of input columns.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getInputColumnCount( lasso_request_t token, int * count );
/**
	@brief Gets the number of sort columns for this database action.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param count The resulting number of sort columns.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getSortColumnCount( lasso_request_t token, int * count );
/**
	@brief Gets an individual input column by index.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param num the zero based index of the column to retrieve.
	@param v The resulting value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getInputColumn( lasso_request_t token, int num, auto_lasso_value_t * v );
/**
	@brief Gets an individual input column by index.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param num the zero based index of the column to retrieve.
	@param v The resulting value.
	@param desc A pointer to a LP_TypeDesc in which to place the original type of the input data.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getInputColumn2( lasso_request_t token, int num, auto_lasso_value_t * v, LP_TypeDesc * desc );
/**
	@brief Gets an individual input column by index, preserving the original type.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param num the zero based index of the column to retrieve.
	@param colName The name of the column will be stored in the \a name member.
	@param type The value portion of the input column in its original type.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getInputColumn3( lasso_request_t token, int num, auto_lasso_value_t * colName, lasso_type_t * type );
/**
	@brief Gets an individual sort column by index.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param num the zero based index of the sort column to retrieve.
	@param v The resulting value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getSortColumn( lasso_request_t token, int num, auto_lasso_value_t * v );
/**
	@brief Gets an individual input column by name.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name the name of the input column to retrieve.
	@param value The resulting value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_findInputColumn( lasso_request_t token, const char * name, auto_lasso_value_t * value );
/**
	@brief Gets an individual input column by name.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name the name of the input column to retrieve.
	@param value The resulting value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_findInputColumnW( lasso_request_t token, const UChar * name, auto_lasso_value_t * value );
/**
	@brief Gets the logical operator for this database action.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param op The resulting operator code. Operator codes are declared in LassoDefines.h.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getLogicalOp( lasso_request_t token, LP_TypeDesc * op );
/**
	@brief Gets the number of return columns for this action.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param count The resulting number of return columns.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getReturnColumnCount( lasso_request_t token, int * count );
/**
	@brief Gets an individual return column by index.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param num The zero based index of the return column to retrieve.
	@param v The resulting value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getReturnColumn( lasso_request_t token, int num, auto_lasso_value_t * v );
/**
	@brief Adds information about a particular column. 
	
	Column information should be added in the order in which the columns occur inthe database.
	Column information should be added no matter what the action is. For the show action, only column information is added.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name The column name.
	@param nullOK IS a null value for this column ok?
	@param type The data type for the column as represented by a type code. osConfig.h.
	@param protection The code for the column's protection. osConfig.h.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_addColumnInfo( lasso_request_t token, const char * name, int nullOK, LP_TypeDesc type, LP_TypeDesc protection );
/**
	@brief Adds information, including valuelists, about a particular column.
	
	Column information should be added in the order in which the columns occur inthe database.
	Column information should be added no matter what the action is. For the show action, only column information is added.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param name The column name.
	@param nullOK IS a null value for this column ok?
	@param type The data type for the column as represented by a type code. osConfig.h.
	@param protection The code for the column's protection. osConfig.h.
	@param valueList An array of strings for the column's valuelist.
	@param countValueList The number of values in the value list.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_addColumnInfo2( lasso_request_t token, const char * name, int nullOK, LP_TypeDesc type, LP_TypeDesc protection, const char ** valueList, int countValueList );
/**
	@brief Add the column data for the next result row.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param columns An array of column values.
	@param dataSizes An array of column value lengths. Every column in \a columns should have an associated length.
	@param numColumns The number of items in the \a columns and \a dataSizes arrays.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_addResultRow( lasso_request_t token, const char ** columns, unsigned long * dataSizes, int numColumns );
/**
	@brief Add the column data for the next result row. Column data is represented as Lasso types.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param cols An array of column values.
	@param num The number of items in the \a cols array.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_addResultRow2( lasso_request_t token, lasso_type_t* cols, int num );

/**
	@brief Adds a new result set to the datasource results.
	
	At the onset of each LCAPI datasource call, there is an initial blank result set created. This is the first result set. 
	After a datasource has added all the data for the first result set, if there is a second result set, the datasource should call
	lasso_addResultSet to start a new set and then proceed to populate the column information and data for that set as normal. 
	This should be repeated for each result set. lasso_addResultSet should be called for each result set after the first. 
	Calling lasso_addResultSet before populating the first result set will result in the first set being empty.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_addResultSet( lasso_request_t token );
/**
	@brief Sets the number of rows found in the query. 
	
	This will not always be the same value as the number of rows added with the lasso_addResultRow call as the skip recs and max recs parameters effect that number.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param num the number of found rows.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_setNumRowsFound( lasso_request_t token, int num );
/**
	@brief Returns the name the current datasource module was registered with.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param val The resulting datasource module name which will be palced in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getDataSourceModuleName( lasso_request_t token, auto_lasso_value_t * val );
/**
	@brief Allows a datasource to set a prepared statement pointer.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param ptr The pointer value to set for later retrieval.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_setDSPreparedPtr( lasso_request_t token, void * ptr );
/**
	@brief Allows a datasource to get a previously set prepared statement pointer.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param ptr The pointer, if it has been presiously set, will be placed in this parameter.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.
*/
LCAPICALL osError lasso_getDSPreparedPtr( lasso_request_t token, void ** ptr );

/**
	@brief Called to access the current datasource connection.
	
	Datasource connections are set using the lasso_setDSConnection function.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param conn A pointer in which to store the result.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL osError lasso_getDSConnection(lasso_request_t token, lasso_dsconnection_t * conn);
/**
	@brief Called to set the current connection for the datasource.
	
	May recurse to deliver the datasourceCloseConnection message if there is already a valid lasso_dbconnection_t.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param conn The connection value to set.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_setDSConnection(lasso_request_t token, lasso_dsconnection_t conn);
/**
	@brief Called to set the statement for the current action.
	
	Datasources must call this to support the action_statement tag.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param stat The statement value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_setActionStatement(lasso_request_t token, const char * stat);
/**
	@brief Called to set the statement for the current action.
	
	Datasources must call this to support the action_statement tag.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param stat The statement value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_setActionStatementW(lasso_request_t token, const UChar * stat);
/**
	@brief Used to check for the -statementonly inline param.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param out A pointer in which to store the result.	
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getIsStatementOnly(lasso_request_t token, bool * out);
/** @} */

/**
	@name Semaphore & Thread Functions
	
	Semephores are a means by which access to shared resources can be control amongst several threads of execution.
	Each semephore must have a unique name. Always release semaphores as soon as possible.
*/
/** @{ */
LCAPICALL osError lasso_createSem( lasso_request_t token, const char * name );
LCAPICALL osError lasso_destroySem( lasso_request_t token, const char * name );
LCAPICALL osError lasso_acquireSem( lasso_request_t token, const char * name );
LCAPICALL osError lasso_releaseSem( lasso_request_t token, const char * name );

typedef void (*thread_done_func)(void *);
LCAPICALL osError lasso_addThreadDoneFunc(thread_done_func, void * arg);

/** @} */

/**
	@brief Processes the supplied buffer.
	
	The buffer should consist of plain text, bracketed Lasso tags or <?LassoScript ?> pairs.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param buffer The buffer holding the script data.
	@param output The resulting data which will be stored in the \a name portion of the struct.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_processBuffer( lasso_request_t token, const char * buffer, auto_lasso_value_t * output );
/* Synonym */
LCAPICALL osError lasso_formatBuffer( lasso_request_t token, const char * buffer, auto_lasso_value_t * output );

/* ********************************************************
	STORAGE FUNCTIONS
********************************************************* */

/*
  These calls can be used to have Lasso store arbitrary persistent data. This is an excellent way to store preferences.
 */
/* 
   Returns osErrNoErr if the data, specified by key, exists.
   The length of the stored data can be ruented in the length parameter.
   This parameter may be NULL if the length is not required.
 */
LCAPICALL osError lasso_storeHasData( lasso_request_t token, const char * key, unsigned int * length );

/*
  Adds the data to Lasso's storage.
  key is the unique identifier for the data.
 */
LCAPICALL osError lasso_storePutData( lasso_request_t token, const char * key, const void * data, unsigned int length );

/*
  Fetches that has been stored.
  key is the unique identifier for the data.
  The data will be returned in the "data" member of the auto_lasso_value_t parameter.
  The length of the resulting data is stored in the length parameter. This param may NOT be NULL.
 */
LCAPICALL osError lasso_storeGetData( lasso_request_t token, const char * key, auto_lasso_value_t * value, unsigned int * length );

/**
	@name Security Functions
*/
/** @{ */
/**
	@brief Checks the current request for administrator privileges.
	
	Returns osErrNoErr if the current user has administrator privs.
  	This is useful for doing administrative activities or file accesses which only the administrator should be able to do.
  	
  	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_isAdministrator( lasso_request_t token );
/**
	@brief Checks the current request for the lowest privilege level.
	
	Returns osErrNoErr if the current user is the anonymous user.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_isAnonymous( lasso_request_t token );

/**
	@brief Checks the current user's group membership.
	
	Returns osErrNoErr if the current user is in any of the specified groups.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param names An array of group names to check. This array should be terminated by a NULL value.
	@return If the function succeeds, and the user is a member of one of the groups, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_isInGroups( lasso_request_t token, const char ** names );

/**
	@brief Switches the current request to the given user.
	
	A proper password must be provided. The switch will persist until the request is finished or
	is switches again.
	Pass "anonymous" with no password to switch to lowest privileges.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param username The username.
	@param password The password.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_switchToUser( lasso_request_t token, const char * username, const char * password );

/**
	@brief Various types of ways to access files.
*/
typedef enum secure_operation_t
{
	kInspectFiles, /**< Query size, etc. */
	kCreateFiles, /**< Create. */
	kDeleteFiles, /**< Delete. */
	kCopyFiles, /**< Copy. */
	kMoveFiles, /**< Move. */
	kReadFiles, /**< Read. */
	kWriteFiles, /**< Write. */
	kAccessOutsideRoot, /**< Outside of web root. */
	kExecuteFile /**< Indicates read access and that the file's extension is approved for script execution. */

} secure_operation_t;

/**
	@brief Tests for file access permissions.
	
	Returns osErrNoErr if the given operation is permitted by Lasso security.
	The data perameter should be the full path to the file using the *internal pathname* format.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param op The file operation.
	@param data The path name.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_operationAllowed( lasso_request_t token, secure_operation_t op, void * data );
/** @} */
const int kMaxPathLen = 1024;
typedef char osPathname[1024];

/**
	@brief Converts an internal pathname into a platform-specific pathname.

	@param inInternalPath The internal pathname.
	@param outPlatformPath The resulting platform specific pathname.
*/
LCAPICALL void lasso_getPlatformSpecificPath( const char* inInternalPath, osPathname outPlatformPath );

/**
	@brief Converts a platform-specific pathname into an internal pathname.
	
	@param inPlatformPath the platform specific path name.
	@param outInternalPath The resulting internal pathname.
*/
LCAPICALL void lasso_getInternalPath( const char* inPlatformPath, osPathname outInternalPath );

/**
	@brief Converts a relative path into a fully qualified from-the-server-root path.
	
	If the path is already from the root it won't be changed.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param inRelativePath The relative path.
	@param outFullyQualified the resulting fully qualified path.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL void lasso_fullyQualifyPath( lasso_request_t token, const char * inRelativePath, osPathname outFullyQualified );

/**
	@brief Converts a partial path into a full, from the file system root, internal path.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param inPath The path to resolve.
	@param outFullPath The resulting full path.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.			
*/
LCAPICALL void lasso_resolvePath( lasso_request_t token, const char * inPath, osPathname outFullPath );

/**
	@brief Converts a partial path into a full, from the file system root, internal path which is guaranteed to be within the web root.

	Resulting path must name an item within the web-root or a blank string is returned.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param inPath The path to resolve.
	@param outFullPath The resulting full path.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.			
*/
LCAPICALL void lasso_resolveIncludePath( lasso_request_t token, const char * inPath, osPathname outFullPath );

/**
	@brief Converts an internal path into a full platform specific path.
	
	Takes a path, such as one which may have been passed to any Lasso tag, and converts it into a full platform specific path name.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param relativeOrFullInternalPath The input path to convert.
	@param fullPlatformPath The resulting converted path.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.	
*/
LCAPICALL void lasso_internalToFullPlatformPath( lasso_request_t token, const char * relativeOrFullInternalPath, osPathname fullPlatformPath );

/**
	@brief Tests the given internal path to see if it is a full path from the filesystem root.
	
	@param path The path to test.
	@return True if the given internal path is a full path from the filesystem root.
*/
LCAPICALL bool lasso_isFullInternalPath(const char * path);
/**
	@brief Toggles how Lasso resolves file paths.
	
	Toggles whether Lasso will resolve the entire given file path (/blah/file.lasso) through the webserver (the default behavior - true)
	OR
	whether Lasso will resolve "/" through the server and then prepend that to the given file path (false).

	The default behavior can conflict with some usages of mod_rewrite
	As of LP7, this setting is global
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param to The toggle.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_setAutoResolveFullPaths(lasso_request_t token, bool to);
/**
	@brief Retreives Lasso's current file resolution mode.
	
	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param get The pointer in which to store the result value.
	@return If the function succeeds, osErrNoErr is returned, otherwise, an error code indicating the problem is returned.		
*/
LCAPICALL osError lasso_getAutoResolveFullPaths(lasso_request_t token, bool * get);

/**
	@name Logging Functions
*/
/** @{ */
/**
	@brief The various log levels.
	
	Each log level can be configured to go to zero or more destinations.
*/
typedef enum log_level_t
{
	LOG_LEVEL_CRITICAL, /**< Critical error. */
	LOG_LEVEL_WARNING, /**< Warning. */
	LOG_LEVEL_DETAIL, /**< Detail. */

	LOG_LEVEL_ALWAYS, /**< Always printed to window (only). Used internally in a few places. */

	LOG_LEVEL_SQL, /**< SQL statement for a database action. */
	LOG_LEVEL_DEPRECATED /**< Deprecated tag used. */
    
} log_level_t;

/*
	Various destinations for logged information to go.
	OR them together to set the preference if the information is to go to various destinations.
*/
/** @brief Printed to the console 

*/
const int flag_logDestinationConsole =	0x00000001; 
/** @brief Written to the LassoErrors.txt file 

*/
const int flag_logDestinationFile =		0x00000002; 
/** @brief Inserted into the errors table 

*/
const int flag_logDestinationDatabase =	0x00000004; 

/**
	@brief Changes the system-wide log destination preference.
	
	Multiple destinations can be OR'd together.
*/
LCAPICALL osError lasso_logSetDestination( log_level_t msgLevel, int dest );

/**
	@brief Log a message.
	
	The message goes to the prefered destination for the message level.
	Messages are limited to 2048 bytes in length.
	Messages sent to the console are limited to 512 bytes in length.
	Messages sent to the database are limited to a little less than 2048 bytes since the 
		total length of the sql statement used to insert the message is limited to 2048 bytes.
*/
LCAPICALL osError lasso_log( log_level_t msgLevel, const char * fmt, ... );
/** @} */
// This will cause Lasso's connection listening thread, which is normally the main application thread, to be split off into it's own thread
// This has no effect when running as an NT service
// The parameter is a function to call as the new "main loop"
// If some other entity has already called this and given its own mainloop function
//	the call will fail
// This can only be called at startup (before connections are accepted)
// Don't call this
typedef void (*capi_mainloop)(void);
LCAPICALL osError lasso_setMainLoop(capi_mainloop x);

// These calls shouldn't be used unless you are loading the LassoService framework
// Initialize a new instance of Lasso given the CLI arguments
// Has no effect if Lasso is already initialized
// TODO: list the CLI arguments
LCAPICALL osError lasso_initialize(int argc, char ** argv);
// Shutdown an instance of Lasso
LCAPICALL osError lasso_terminate();
// Called to have Lasso's connector listening thread start listening
// This call will not return until Lasso is instructed to stop listening
// for connector connections
LCAPICALL osError lasso_run();

// Get the site id
LCAPICALL osError lasso_getSiteID(lasso_request_t token, osInt64 * outID);
// Get the site name
LCAPICALL osError lasso_getSiteName(lasso_request_t token, auto_lasso_value_t * outName);
LCAPICALL osError lasso_getSiteNameW(lasso_request_t token, auto_lasso_value_w_t * outName);

// takes 1 or more sql statements, delimited by semi colon, and, upon successive calls, 
// will return the start and end of each individual statement in sequence
// used in SQL DS modules
LCAPICALL bool lasso_sqlFindStat( const char * stat, int * start, int * finish, int maxLen );

/**
	@brief Document resolution
	
	resolve_doc_func is a user provided function which, when properly registered using \a lasso_registerDocumentResolver,
	is called in an attempt to transform a path name into document content.

	@param token The request token which should be passed to subsequent LCAPI functions for the duration of the current service function call.
	@param context The user provided context that was given when the resolver was registered.
	@param inUTF8Path The path of the requested document.
	@param resultContent The resulting content of the document. The content may be of any type. For example, if the path pointed to an image file, the resultingContent could be a bytes type, or possibly an image type depending on the operation of the document resolver.
	@param lastModDate The modification of the document file. If modification dates are not supported by the particular document resolver, the initial value passed in to the func should not be modified.
	@param ifModifiedSince The modification date against which to match the result document's modification date. If the document was not modified since this date, the document resolver should return osErrPreconditionFailed and should not read the file's content or return anything in the resultContent parameter.
	@param unused Pass NULL for this value. It is used, just internally
	@param flags Some flags to give a hint to the resolvers of the purpose of this doc
*/
typedef osError (*resolve_doc_func)(lasso_request_t token, void * context, const char * inUTF8Path, lasso_type_t * resultContent, time_t * lastModDate, time_t * ifModifiedSince, void ** unused, int flags);
const int RESOLVE_FLAG_RAW = 0x01; // this doc is supposed to be raw bytes. if a resolver can't satisfy that, it won't say it found the file

/**
	@brief Document resolver priority	
*/
typedef enum resolver_priority { prio_low, prio_high };

/**
	@brief Register a document resolver

	Registers a document resolver function.

	@param token The request token. If provided, the registeration of this resolver will only be in effect for the remainder of the current request. If provided, the registration will be in effect globally for this site and can only be removed manualy using \a lasso_deregisterDocumentResolver.
	@param func The resolve_doc_func which will be called to resolve a document.
	@param context A user provided value which will be passed to the resolve_doc_func for each invocation.
	@param prio The priority for this resolver. If the prio is prio_high, this resolver will be placed at the front of the resolver list and so will be called before the resolvers behind it. If a resolver returns osErrNoErr, no further resolvers will be called. If the priority is not prio_high, the resolver will be placed at the end of the list.
*/
LCAPICALL osError lasso_registerDocumentResolver(lasso_request_t token, resolve_doc_func func, void * context, resolver_priority prio);

/**
	@brief Removes a document resolver.

	Removes a specific document resolver from the list.

	@param token The request token. If provided, the deregisteration of this resolver will only be in effect for the remainder of the current request. If provided, the resolver will be in removed from the global resolver list for this site.
	@param func The previously registered func.
	@param context The same context value that the resolver was originally registered with. This must be provided and must match the original value.
*/
LCAPICALL osError lasso_deregisterDocumentResolver(lasso_request_t token, resolve_doc_func func, void * context);

#ifdef __cplusplus
}
#endif

#endif

