#ifndef _LASSOPROTO_H_
#define _LASSOPROTO_H_

#include "osConfig.h"
#include <sys/types.h>

// Lasso Service Protocol
// All messages to and from the Lasso Service will begin with the LPCommandBlock
// declared below. Additional data may be required for each command.
// Commands may be overlapped, but should not be segmented.

#ifdef __cplusplus
extern "C"
{
#endif

// command types
typedef enum LPCommand
{
	cmdProtoErr = 0,	// Indicates that an error has occurred in the use of the protocol
					// The sequence number will indicate the offending command
					// DATA REQUIRED: four byte integer indicating the error code
					//				any additional data will be a textual description of that went wrong
					// SENT BY: LassoService
					// REPLY: none
	
	cmdLogin,		// First reaquired command to open a session
					// Included authentication information
					// DATA REQUIRED: null terminated username and password
					//				(password doesn't need to be terminated since the fDataSize will indicate the total size for both strings)
					// SENT BY: client
					// REPLY: cmdLoginRep
					
	cmdLoginRep,	// DATA REQUIRED: one byte indicating true (1) or false (0) that the login was successful
					// SENT BY: LassoService
					// REPLY: none
					
	cmdLogout, 		// DATA REQUIRED: none
					// SENT BY: client
					// REPLY: none
	
	cmdOpenReq,		// DATA REQUIRED: none
					// SENT BY: client
					// REPLY: cmdOpenReqRep
					
	cmdOpenReqRep, 	// DATA REQUIRED: four byte integer indicating the id of the opened request
					//				a negative id will indicate a failure in opening the request
					// SENT BY: LassoService
					// REPLY: none
					
	cmdCloseReq,	// DATA REQUIRED: none
					// SENT BY: LassoService or client
					//			sent by LassoService when there is no more data to be sent to the web browser
					//			the request id of the block indicates the id of the request that is closed
					// REPLY: none
	
	cmdGetParam,	// DATA REQUIRED: RequestParamKeyword as defined in RequestParams.h
					//					Then a four byte integer indicating the size
					//					of the data for the argument
					//					Multiple params may follow
					// SENT BY: LassoService
					// REPLY: cmdGetParamRep
	
	cmdGetParamRep,	// DATA REQUIRED: Four bytes indicating the size of the character data
					//					for the requested param
					//					If multiple params were requested, the data for each
					//					param should follow in the original order
					// SENT BY: client
					// REPLY: none
					
	cmdPushData,	// DATA REQUIRED: the data that should be sent to the web browser
					// SENT BY: LassoService
					// REPLY: none 
	
	cmdLogMsg		// DATA REQUIRED: text message for Lasso to log to the console window (if any)
					// SENT BY: client
					// REPLY: none
	
};

typedef osInt32 LPRequestID;
typedef osUInt32 LPSequenceNum;

// Lasso Proto Command Block
// used for sending all commands to and from LassoService
// EVERYTHING IS IN NETWORK BYTE ORDER
typedef struct _LPCommandBlock
{
	osInt32 fCmd; 		// the command
							// commands are enumerated above
	
	LPSequenceNum fSeqNum; 	// the sequence number for the command. should increment for each command
							// replies will have the same sequence number as the original command
							
	LPRequestID fReqID;	// the id of the request attached to the command
				// this should match what is returned by LassoService in the reply to the open request (cmdOpenReq) command

	osInt32 fResultCode; // the result of the command. used if the command is a reply
					// error codes are defined in osErrors.h
	
	osUInt32 fDataSize; // size of the data to follow (may be zero)
	// additional command specific data may follow
	// the size of the data is identified by the fDataSize member
} LPCommandBlock;

#ifdef __cplusplus
}
#endif

#endif

