#ifndef LASSO_WEBSTAR
#include "osConfig.h"
#endif
#include "POSTReader.h"
#include <string.h>
#include <time.h>
#include <ctype.h>
#include "osErrors.h"

#if !defined(WIN32) && !defined(WIN64)
#include <sys/stat.h>
#endif

static const char * kMultiPartForm = "multipart/form-data";
static const char * kBoundary = "boundary";

static const char * kContentDisposition = "Content-Disposition";
static const char * kContentType = "Content-Type";

static const char * kUploadName = "-upload.realname=";
static const char * kTmpUploadName = "-upload.name=";
static const char * kFileType = "-upload.type=";
static const char * kFileSize = "-upload.size=";
static const char * kUploadFieldName = "-upload.fieldname=";
static const char * kUpload2FieldName = "-upload2.fieldname=";

static const char * kLassoTempPrefix = "lasso-tmp";
static const char * kUploadExtension = ".uld";

//#define DUMP_POST

#if defined(WIN32) || defined(WIN64)
#define _tempnam tempnam
#endif

#if defined(WIN32) || defined(WIN64)
#define strncasecmp _strnicmp
#endif

// works like strstr but doesn't stop at null termination
// pat should be null terminated
inline char * memmem( char * start, char * pat, unsigned int len )
{
	int patLen = strlen(pat);
	for ( unsigned int i = 0; i < len; ++i )
	{
		if ( memcmp(start+i, pat, patLen) == 0 )
		{
			return start + i;
		}
	}
	return NULL;
}

LP8POSTReader::LP8POSTReader( const char * contentType )
: fMaxFileSize(-1), fMulti(false), fState(stateNone), fGotFile(false)
{
	if ( contentType && strstr( contentType, kMultiPartForm ) != NULL ) // only one body
	{
		const char* tmp = NULL;

		fMulti = true;
		fState = stateBoundary;

		//Need to check if boundary data really there before buffering
		if(	(tmp = strstr( contentType, kBoundary )) != NULL )
		{
			tmp += strlen( kBoundary ) + 1;
			fBoundary += "--";
			fBoundary.append( tmp, strlen(tmp) );
		}
	}
	srand((unsigned int)time(NULL));
}

LP8POSTReader::~LP8POSTReader()
{
    for ( std::list<BodySpec*>::iterator it = fBody.begin(), end = fBody.end(); it != end; ++it )
    {
		delete (*it);
    }
}

// returns true if the field name matches
// tests for a match up until a colon in against
inline bool 
IsField( const char * match, const char * against )
{
	unsigned int pos = 0;
	while ( against[pos] != ':' )
	{
		if ( tolower(match[pos]) != tolower(against[pos]) )
			return false;
		++pos;
	}
	return true;
}

inline bool
IsBoundaryStart( const char * tst, const char * bndry )
{
	return strncmp(bndry, tst, strlen(bndry)) == 0;
}

inline bool
CouldBeBoundaryStart( const char * tst, int length, const char * bndry )
{
	int bndryLen = strlen(bndry);
	if (bndryLen <= length)
		return IsBoundaryStart(tst, bndry);
	
//	FILE * theTest = fopen( "/var/tmp/FILEUPLOAD.log", "ab" );
//	fprintf(theTest, "**Testing broken boundary %d/%s/%s**\n", length, std::string(tst, length).c_str(), bndry);
//	fclose(theTest);
	for (int x = 0; x < length; ++x)
	{
		if (tst[x] != bndry[x])
			return false;
	}
	return true;
}

inline bool
PullValue( const char * named, const char * from, unsigned int maxLen, std::string & putInto )
{
	unsigned int tmp = strlen(named);
	while( maxLen > 0 )
	{
		while( maxLen > 0 && strncasecmp(from, named, tmp) != 0 )
		{
			++from;
			--maxLen;
		}
		if ( maxLen > 0 ) // match so far
		{
			if ( maxLen >= tmp )
			{
				from += tmp;
				maxLen -= tmp;
				if ( *from == '=' )
				{
					from += 2;
					maxLen -= 2;

					unsigned int len = 0;
					while( len < maxLen && from[len] != '"' )
						++len;
					putInto.append(from, len);
					return true;
				}
			}
			++from;
			--maxLen;
		}
	}
	return false;
}

inline void
GetTempFileName( std::string & io )
{
	char* tmpName = NULL;

	tmpName = tempnam( NULL, kLassoTempPrefix );

	io = tmpName;
	
	char tmp[127];
	sprintf(tmp, "%ld%ld", ((long)tmpName)%10000, (long)abs(rand())%10000);
	
	io += tmp;
	io += kUploadExtension;

	free(tmpName);
}

inline void 
OpenTempFile( LP8POSTReader::BodySpec & spec )
{
	spec.fFile = fopen( spec.fTmpFileName.c_str(), "w+b" );
}

#define THROW_IF(Y) { if (Y) throw osErrAssert; }
#define ADVANCE(X) { THROW_IF( ((unsigned int)(X)) > length); data += (X); length -= (X); }
#define ADVANCE_PAST_FIELD(X) { ADVANCE(strlen(X)); ADVANCE(1); while(isspace(*data)) ADVANCE(1); }
#define LINE_END 2
const char CR = '\r';
const char LF = '\n';

#define BUFFER_DATA() do { 	char * tmp = new char[length];\
							memcpy(tmp, data, length);\
							fBuffer = "";\
							fBuffer.append(tmp, length);\
							delete [] tmp;\
							ADVANCE(length);\
							clearBuffer = false;\
						} while(0)

// returns position that the EOL starts (pos of \r)
inline int ProbeEOL( const char * data, unsigned int length )
{
	for ( unsigned int x = 0; x < length; ++x )
	{
		if ( data[x] == CR && data[x+1] == LF )
			return x;
	}
	return -1; // no EOL
}

int LP8POSTReader::AddToBuffer( const char* data, unsigned int length )
{
#ifdef DUMP_POST
	FILE * theTest = fopen( "/var/tmp/FILEUPLOAD_DUMP.log", "ab" );
	fwrite(data, 1, length, theTest);
	fwrite("||BREAK||", 1, 9, theTest);
	fclose(theTest);
#endif
	if( IsMultiPart() )
	{
		if ( fBuffer.length() != 0 ) // work with buffer instead
		{
			fBuffer.append(data, length);
			data = fBuffer.data();
			length = fBuffer.length();
		}
		
		bool clearBuffer = true; // once we run out of chars, clear fBuffer
		
		while( length > 0 )
		{
			switch(fState)
			{
			case stateDone:
			case stateNone:
				return -1;
			case stateBoundary:
				{
					if ( length < fBoundary.length()+2 ) // not a full boundary line; buffer
					{
						BUFFER_DATA();
					}
					else
					{
						ADVANCE(fBoundary.length());
						if ( data[0] == '-' && data[1] == '-' ) // DONE
						{
							fState = stateDone;
							ADVANCE(2);
						}
						else
						{
							fState = stateHeader;
							fBody.push_back(new BodySpec());
						}
						if ( fState != stateDone )
						{
	 						ADVANCE(LINE_END);
						}
						else
						{
							ADVANCE(length); // just get rid of it
						}
					}
				}
				break;
			case stateHeader:
				{
					int eolPos = ProbeEOL(data, length);
					if ( eolPos == -1 ) // no eol
					{
						// broken header data
						BUFFER_DATA();
						
						// else we just appended more data
						// yet still have not reached eol
						// keep going
					}
					else
					{
						BodySpec & spec = *fBody.back();
						if ( eolPos != 0 ) // eolPos is zero if we had a break between the crlfcrlf pairs
						{
							// we are interested in Content-Disposition
							// and Content-Type
							if ( IsField(kContentDisposition, data) )
							{
								// we want the name= part
								PullValue("name", data, eolPos, spec.fFieldName);
								// we want the filename= part (if it's there)
								PullValue("filename", data, eolPos, spec.fFileName);

//								FILE * theTest = fopen( "/var/tmp/FILEUPLOAD.log", "ab" );
//								fprintf(theTest, "Found file %s/%s\n", spec.fFieldName.c_str(), spec.fFileName.c_str());
//								fclose(theTest);
							}
							else if ( IsField(kContentType, data) )
							{
								unsigned int advLen = strlen(kContentType) + 2;
								spec.fContentType.append(data + advLen, eolPos - advLen);
							}
							// else don't care?
							ADVANCE(eolPos);
							ADVANCE(LINE_END);
						}
						if ( (eolPos == 0 || length) && data[0] == CR && data[1] == LF ) // data starting!!
						{
							ADVANCE(LINE_END);
							// if we have a file name, set that up
							if ( spec.fFileName.length() > 0 )
							{
								GetTempFileName(spec.fTmpFileName);
								OpenTempFile(spec);
								fState = stateFile;
							}
							else // regular data
							{
								fState = stateNameValue;
							}
						}
					}
				}
				break;
			case stateNameValue:
				{
					BodySpec & spec = *fBody.back();
					while( length )
					{
						// see if we are at the boundry start
						// note we won't want the CRLF immediately preceeding it
						if ( data[0] == CR )
						{
							if ( length == 1 ) // buffer
							{
								BUFFER_DATA();
								continue;
							}
							else if ( data[1] == LF )
							{
								// if we have some dashes but no finishing CRLF we buffer it until the next send
								if ( IsBoundaryStart(data+2, fBoundary.c_str()) )
								{
									ADVANCE(LINE_END);
									fState = stateBoundary;
									break;
								}
								else if ( length-2 < fBoundary.length() )
								{
									int tempLen = length-2;
									int tempIndex = 2;
									while( tempLen && data[tempIndex] == fBoundary[tempIndex-2] )
									{
										--tempLen;
										++tempIndex;
									}
									if ( tempLen == 0 ) // all we found were dashes; buffer it
									{
										BUFFER_DATA();
										continue;
									}
								}
							}
						}
						// cat the data
						spec.fFieldValue.append(data, 1);
						ADVANCE(1);
					}
				}
				break;
			case stateFile:
				{
					BodySpec & spec = *fBody.back();
					while( length )
					{
						// see if we are at the boundry start
						// note we won't want the CRLF immediately preceeding it
						if ( data[0] == CR )
						{							
							if ( length == 1 ) // buffer
							{
								BUFFER_DATA();
								continue;
							}
							else if ( data[1] == LF )
							{
								if ( IsBoundaryStart(data+2, fBoundary.c_str()) )
								{
									ADVANCE(LINE_END);
									fState = stateBoundary;
									FCLOSE(spec.fFile);
#ifndef WIN32
									chmod(spec.fTmpFileName.c_str(), S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
#endif
									spec.fFile = MY_INVALID_FILE;
									break;
								}
								else
								{
									if (CouldBeBoundaryStart(data+2, length-2, fBoundary.c_str()))
									{
										BUFFER_DATA();
										continue;
									}
								}
							}
						}
						if ( spec.fFile != MY_INVALID_FILE )
						{
							unsigned int writeLen = 0; 
							while( writeLen < length )
							{
								if (data[writeLen] == CR && data[writeLen+1] == LF)
								{
									if (IsBoundaryStart(data+writeLen+2, fBoundary.c_str()))
										break;
									if (length-writeLen-2 && CouldBeBoundaryStart(data+writeLen+2, length-writeLen-2, fBoundary.c_str()))
									{
//										FILE * theTest = fopen( "/var/tmp/FILEUPLOAD.log", "ab" );
//										fprintf(theTest, "**Found broken boundary %d**\n", length-writeLen-2);
//										fclose(theTest);
										break;
									}
								}
								++writeLen;
							}
							// cat the data
							spec.fFileSize += fwrite( data, sizeof(char), writeLen, spec.fFile );
							if (writeLen == length)
								fBuffer = "";
							if ( fMaxFileSize > -1 && int(spec.fFileSize) > fMaxFileSize ) // exceeded
							{
								FCLOSE(spec.fFile);
								spec.fFile = MY_INVALID_FILE;
								UNLINK(spec.fTmpFileName.c_str());
								spec.fTmpFileName = "";
								spec.fContentType = "UPLOAD SIZE EXCEEDED";
								spec.fFileSize = 0;
								ADVANCE(writeLen);
								continue;
							}
							ADVANCE(writeLen);
							fGotFile = true;
							continue;
						}
						ADVANCE(1);
					}					
				}
				break;
			}
		}
		if (clearBuffer)
		{
			fBuffer = "";
		}
	}
	else
		fBuffer.append( data, length );

	return fState;
}

const std::string* LP8POSTReader::GetPostArgs( )
{
	if( IsMultiPart() )
	{
		std::list<BodySpec*>::iterator it = fBody.begin();
	
		fBuffer = "";

		while( it != fBody.end() )
		{
			if ( fBuffer.length() > 0 )
				fBuffer += "&";
			std::string temp;

			if( (*it)->fFileName.length() > 0 ) //a file
			{
				temp = (*it)->fFileName;
				EncodeURL( &temp );
				fBuffer += kUploadName + temp;
				
				fBuffer += "&";
				temp = (*it)->fTmpFileName;
				EncodeURL( &temp );
				fBuffer += kTmpUploadName + temp;

				fBuffer += "&";

				temp = (*it)->fContentType;
				EncodeURL( &temp );
				fBuffer += kFileType + temp;

				fBuffer += "&";

				char tmp[256];
				int tmpInt = (int)(*it)->fFileSize;
				sprintf(tmp, "%d", tmpInt);
				fBuffer += kFileSize;
				fBuffer += tmp;

				fBuffer += "&";

				fBuffer += kUploadFieldName;
				temp = (*it)->fFieldName;
				EncodeURL(&temp);
				fBuffer += temp;
				
				// this is a temp fix to present the file's field name so it's accessible in LP5
				fBuffer += "&";
				fBuffer += kUpload2FieldName;
				fBuffer += temp;
			}
			else
			{
				if ((*it)->fContentType.length() > 0)
				{
					fBuffer += "-contenttype=";
					temp = (*it)->fContentType;
					EncodeURL( &temp );
					fBuffer += temp;
					fBuffer += "&";
				}
				temp = (*it)->fFieldName;
				EncodeURL( &temp );
				fBuffer += temp;
				fBuffer += "=";
				temp = (*it)->fFieldValue;
				EncodeURL( &temp );
				fBuffer += temp;
			}
			++it;
		}
	} 
	else
	{
		if( fBuffer.length() < 1 )
			return NULL;
	}

	return &fBuffer;
}
/*
void LP8POSTReader::GetTMPFilePaths( std::vector<std::string>& tmppaths )
{
}
*/

bool LP8POSTReader::IsUploadedFile( const char * path ) const
{
	return strstr(path, kLassoTempPrefix) != NULL;
}

void LP8POSTReader::Reset()
{
	fBuffer = "";
	fBoundary = "";
    for ( std::list<BodySpec*>::iterator it = fBody.begin(), end = fBody.end(); it != end; ++it )
    {
	delete (*it);
    }
	fBody.clear();
}

inline bool ShouldEncode( char x )
{
	return (x <= 32) || (x >= 127) || (x == '$' || x == '&' || x == '+' || x == ',' || x == '/' || x == ':' || x == ';' || x == '=' || x == '?' || x == '@' ) ||
		(x == '\'' || x == '"') || (x == '<' || x == '>') || (x == '#') || (x == '%') || (x == '{' || x == '}' || x == '|' || x == '\\' || x == '^' || x == '~' || x == '[' || x == ']' || x == '`' )
		;
}

bool LP8POSTReader::EncodeURL( std::string* data )
{	
	if( data->empty() )
		return false;
	
	for ( int i = 0; i < (int)data->length(); i++ )
	{
		if ( ShouldEncode((*data)[i]) )
		{
			char s[32];
				
			sprintf( s, "%%%x", (unsigned char)(*data)[i] );
			int len = strlen(s);
			/*
			if ( len > 3 )
			{
				s[1] = s[len-2];
				s[2] = s[len-1];
				s[3] = '\0';
			}
			// low-low ascii chars only have one hex char. we will add a '0'
			else */
			if (len == 2)
			{
				s[2] = s[1];
				s[1] = '0';
				s[3] = '\0';
			}
			
			data->erase( i, 1 );
			data->insert( i, s ); 
		}
	}
	
	return true;
}
