/******************************************************************************
 * File:		/osAPI/osDebug.h
 * Date:		Unknown
 * Author:		Unknown
 *
 * Copyright (c) 2000 Blue World Communications, Inc.  All rights reserved.
 *
 * Notes:		Header file for cross-platform debugging and error-handling 
 *				using the built-in OS features.
 *
 * History:
 *		07.07.2000 - Created a version of DEBUGSTRING_FILE_LINE to handle a 
 *					 EOL character so that Statements to the debug window with
 *					 MSVC6.0 would have a line break an not just run together
 *					 like they used to. - CJS
 *		07.03.2000 - Modified formatting of file and added some comments- CJS
 *
 *
 *****************************************************************************/
#ifndef _OSDEBUG_H_
#define _OSDEBUG_H_




// Includes:
//
#include "osConfig.h"					// Lasso's platform configuration
#include "osUtilities.h"				// Platform utilities
#include "osErrors.h"

#if MACHINE_PC
#include <windows.h>					// Windows API
#endif

#if MACHINE_MAC
#include <Threads.h>					// Threads on the Mac
#endif

#if MACHINE_UNIX
#include <pthread.h>					// POSIX threads
typedef pthread_t osThreadID;
#endif

#if MACHINE_LINUX
#include <signal.h>
#endif

#include "osAPIExport.h"			// NOTE: this must be the last include

// debug includes
#if DEBUG_BUILD
#include <cstring>
#include <cstdio>
#endif

#include "LassoCAPI.h"

// debug functions
#if DEBUG_BUILD

	EXPORT void osDebugBreak();		// breaks into the debugger

#endif

// debug macros
#if DEBUG_BUILD
#  ifndef ASSERT
#include <assert.h>
	#define ASSERT( test )						assert(test) //if (!(test)) DEBUGSTRING_FILE_LINE_EOL( test )
#  endif
	#define ASSERTISVALID()						AssertIsValid()

#if MACHINE_PC
	#define ASSERT_NT( test )					ASSERT( WinIsNT4OrBetter() ? (int)(test) : 1 )
	#define AsmDebugBreak()						__asm { int 3 }
	#define DEBUGBREAK()						AsmDebugBreak()
#define DEBUGSTRING( str )						{ lasso_log(LOG_LEVEL_CRITICAL, "%s", str); } //{ OutputDebugString( str ); DEBUGBREAK(); }
	// 07.07.2000 - Added this modified version of DEBUG_FILE_LINE to properly 
	// handle "\n" at the end of the debug string. - CJS
	#define DEBUGSTRING_FILE_LINE_EOL( str ) \
	{ \
		char * MACROBUFFER = new char[1024]; \
		sprintf( MACROBUFFER, "Assertion (%s) failed in \"%s\" on line %d\n", \
				 #str, __FILE__, __LINE__); \
		DEBUGSTRING(MACROBUFFER); \
		delete [] MACROBUFFER; \
	}	// end DEBUGSTRING_FILE_LINE_EOL
	#define DEBUGSTRING_FILE_LINE( str )		{ char * MACROBUFFER = new char[1024]; sprintf(MACROBUFFER, "Assertion (%s) failed in \"%s\" on line %d", #str, __FILE__, __LINE__); DEBUGSTRING(MACROBUFFER); delete [] MACROBUFFER; }
#endif

#if MACHINE_MAC
	#define DEBUGBREAK()						Debugger()
	#define DEBUGSTRING( str )					{ StringPtr MACROMSG = new unsigned char[256]; std::memmove( MACROMSG + 1, str, std::strlen( str ) ); MACROMSG[0] = std::strlen( str ); DebugStr( MACROMSG ); delete [] MACROMSG; }
	#define DEBUGSTRING_FILE_LINE( str )		{ char * MACROBUFFER = new char[1024]; std::sprintf(MACROBUFFER, "Assertion (%s) failed in \"%s\" on line %d", #str, __FILE__, __LINE__); DEBUGSTRING(MACROBUFFER); delete [] MACROBUFFER; }
#endif

#if MACHINE_UNIX
#	include <signal.h>
	#define DEBUGBREAK()						raise(SIGTRAP)
	#define DEBUGSTRING( str )					{ printf(str); DEBUGBREAK(); }
	#define DEBUGSTRING_FILE_LINE( str )		{ char * MACROBUFFER = new char[1024]; sprintf(MACROBUFFER, "Assertion (%s) failed in \"%s\" on line %d", #str, __FILE__, __LINE__); DEBUGSTRING(MACROBUFFER); delete [] MACROBUFFER; }
	// 07.07.2000 - Added this modified version of DEBUG_FILE_LINE to properly 
	// handle "\n" at the end of the debug string. - CJS
	#define DEBUGSTRING_FILE_LINE_EOL( str ) \
	{ \
		char * MACROBUFFER = new char[1024]; \
		sprintf( MACROBUFFER, "Assertion (%s) failed in \"%s\" on line %d\n", \
				 #str, __FILE__, __LINE__); \
		DEBUGSTRING(MACROBUFFER); \
		delete [] MACROBUFFER; \
	}	// end DEBUGSTRING_FILE_LINE_EOL
#ifdef NT_POSIX
	#define ASSERT_NT( test )					ASSERT( WinIsNT4OrBetter() ? (int)(test) : 1 )
	#undef DEBUGSTRING
	#define DEBUGSTRING( str )					{ OutputDebugString( str ); DEBUGBREAK(); }
#endif
#endif

	#define DEBUGBREAKIF( test )				if ( (test) )		DEBUGBREAK()
	#define STEPPATH()							DEBUGBREAK()
#else
#  ifndef ASSERT
	#define ASSERT( test )						// nothing
#  endif
	#define ASSERTISVALID()						// nothing
	#define ASSERT_NT( test )					// nothing
	#define DEBUGSTRING_FILE_LINE( str )		// nothing
	#define DEBUGBREAK()						// nothing
	#define DEBUGBREAKIF( test )				// nothing
	#define STEPPATH()							// nothing
	#define DEBUGSTRING( str )					// nothing
#endif

#if !DEBUG_BUILD
#  ifndef ASSERT
	#define ASSERT( test )						// nothing
#  endif
    #define ASSERTISVALID()						// nothing
    #define ASSERT_NT( test )					// nothing
    #define DEBUGSTRING_FILE_LINE( str )		// nothing
    #define DEBUGBREAK()						// nothing
    #define DEBUGBREAKIF( test )				// nothing
    #define STEPPATH()							// nothing
    #define DEBUGSTRING( str )					// nothing
#endif
// error handling

// redefined the osThreadID type because this file cannot include osThreads.h
#if MACHINE_PC && !defined(USE_POSIX)
typedef DWORD			osThreadID;
#endif

#if MACHINE_MAC
typedef ThreadID		osThreadID;
#endif
/*
#if MACHINE_UNIX && !defined(NT_POSIX)
typedef pthread_t		osThreadID;
#endif*/

// call this at startup
EXPORT void osInitOSError();
EXPORT osError osGetLastError();								// returns the last error that occurred
EXPORT void osSetLastError( osError newErr );					// sets the last error to a new value
EXPORT void osSetLastErrorIf( bool test, osError newErr );		// if test is true sets last error to newErr, otherwise sets last error to osErrNoErr
EXPORT void osClearLastError();									// sets the last error to osErrNoErr
EXPORT const UChar * osGetErrorMessage( osError error );			// returns a text string that describes the error
EXPORT const UChar * osGetLastErrorMessage();						// returns a text string describing the last error received
EXPORT void osRemoveErrorForThread( osThreadID threadID );		// removes the error entry in the error hashmap for the passed-in thread

typedef struct
{
	osError			errorCode;
	const UChar *			errorMessage;
} ErrorInfo;

// error messages
const ErrorInfo kErrorMessages[ kNumErrors ] = 
{
	{	osErrNoErr,						LPW(L"No error")										},
	{	osErrAssert,					LPW(L"Internal program assertion")					},

	// stream errors
	{	osErrStreamReadError,			LPW(L"Error reading from stream")						},
	{	osErrStreamWriteError,			LPW(L"Error writing to stream")						},

	// memory errors
	{	osErrMemory,					LPW(L"Memory error")									},
	{	osErrInvalidMemoryObject,		LPW(L"Invalid memory object")							},
	{	osErrOutOfMemory,				LPW(L"Out of memory")									},
	{	osErrOutOfStackSpace, 			LPW(L"Out of stack space")							},
	{	osErrCouldNotLockMemory,		LPW(L"Could not lock memory")							},
	{	osErrCouldNotUnlockMemory,		LPW(L"Could not unlock memory")						},
	{	osErrCouldNotDisposeMemory,		LPW(L"Could not dispose memory")						},

	// file errors
	{	osErrFile,						LPW(L"File error")									},
	{	osErrFileInvalid,				LPW(L"The file object is invalid")					},
	{	osErrFileInvalidAccessMode,		LPW(L"Invalid access mode")							},
	{	osErrCouldNotCreateOrOpenFile,	LPW(L"Could not create or open file")					},
	{	osErrCouldNotCloseFile,			LPW(L"Could not close file")							},
	{	osErrCouldNotDeleteFile,		LPW(L"Could not delete file")							},
	{	osErrFileNotFound,				LPW(L"File not found")								},
	{	osErrFileAlreadyExists,			LPW(L"File already exists")							},
	{	osErrFileCorrupt,				LPW(L"The file is corrupt")							},
	{	osErrVolumeDoesNotExist,		LPW(L"Volume does not exist")							},
	{	osErrDiskFull,					LPW(L"Disk full")										},
	{	osErrDirectoryFull,				LPW(L"Directory full")								},
	{	osErrIOError,					LPW(L"I/O error")										},
	{	osErrInvalidPathname,			LPW(L"Invalid pathname")								},
	{	osErrInvalidFilename,			LPW(L"Invalid filename")								},
	{	osErrFileLocked,				LPW(L"File locked")									},
	{	osErrFileUnlocked,				LPW(L"File unlocked")									},
	{	osErrFileIsOpen,				LPW(L"File is already open with write permission")	},
	{	osErrFileIsClosed,				LPW(L"File is closed")								},
	{	osErrBOF,						LPW(L"Beginning of file reached")						},
	{	osErrEOF,						LPW(L"End of file reached")							},
	{	osErrCouldNotWriteToFile,		LPW(L"Could not write to file")						},
	{	osErrCouldNotReadFromFile,		LPW(L"Could not read from file")						},

	// resource errors
	{	osErrResNotFound,				LPW(L"Resource not found")							},
	{	osErrResource,					LPW(L"Resource error")								},

	// networking errors
	{	osErrNetwork,					LPW(L"Network error")									},

	// security violation errors
	{	osErrInvalidUsername, 			LPW(L"Invalid user name")								},
	{	osErrInvalidPassword, 			LPW(L"Invalid password")								},
	{	osErrInvalidDatabase, 			LPW(L"Invalid database")								},
	{	osErrNoPermission, 				LPW(L"No permission")									},
	{	osErrFieldRestriction, 			LPW(L"Field restriction")								},
	
	// contiguous web errors
	{	osErrWebAddError,				LPW(L"Add error")										},
	{	osErrWebUpdateError, 			LPW(L"Update error")									},
	{	osErrWebDeleteError, 			LPW(L"Delete error")									},
	
	// general errors
	{	osErrInvalidParameter,			LPW(L"An invalid parameter was passed to a tag")		},
	{	osErrOverflow,					LPW(L"Overflow: some memory passed to a function was too small to hold the results")	},
	{	osErrNilPointer,				LPW(L"A pointer was nil when it should not have been")	},
	{	osErrUnknownError,				LPW(L"Unknown error")									},
	{	osErrLoopAborted,				LPW(L"A looping tag was aborted")						},
	{	osErrSyntaxError,				LPW(L"Bad syntax used in a format file; parsing of the file was aborted")				},	

	{	osErrDivideByZero,				LPW(L"Divide by zero") },	
	{	osErrIllegalInstruction,		LPW(L"Illegal instruction") },

	{	osErrTagNotFound,				LPW(L"Tag not found") },	
	{	osErrVarNotFound,				LPW(L"Variable not found") },
	
	{	osErrAborted,					LPW(L"Execution was aborted.") },	
	{	osErrFailure,					LPW(L"Execution failed.") },

	{	osErrPreconditionFailed,		LPW(L"Precondition failed") },
	{	osErrPostconditionFailed,		LPW(L"Postcondition failed") },
	{	osErrCriteriaNotMet,			LPW(L"Criteria not met") },
	{	osErrIllegalUseOfFrozenInstance, LPW(L"Illegal attempt to modify a frozen value or type")	},
	// non contiguous web errors
	{	osErrWebRequiredFieldMissing,	LPW(L"Value missing for required field")													},
	{	osErrWebRepeatingRelatedField,	LPW(L"Repeating related fields are not supported")										},
	{	osErrWebNoSuchObject,			LPW(L"No records found")																	},
	{	osErrWebTimeout,				LPW(L"Timeout")																			},
	{	osErrWebActionNotSupported,		LPW(L"Action not supported")																},
	{	osErrConnectionInvalid,		LPW(L"The specified database was not found")												},
	{	osErrWebModuleNotFound,			LPW(L"The module was not found")															},
	{	osErrHTTPFileNotFound, 			LPW(L"The file was not found")															},
	{	osErrDatasourceError, 			LPW(L"A data source error has occurred")													}
};	// end kErrorMessages


// call osGetOSError to get a cross-platform error from a platform-specific error
#if MACHINE_PC
EXPORT osError osGetOSError( osError defaultError = osErrUnknownError );
#endif

#if MACHINE_MAC
EXPORT osError osGetOSError( OSErr macError, osError defaultError = osErrUnknownError );
#endif

#if MACHINE_UNIX
EXPORT osError osGetOSError( osError defaultError = osErrUnknownError );
#endif

// exception-handling macros

#define osThrowIf( test )			{	if ( test )				throw osException("osErrAssert", osErrAssert);		}
#define osThrowIfZero( test )		{	if ( (test) == 0 )		throw osException("osErrAssert", osErrAssert);		}
#define osThrowIfNot( test )		{	if ( !(test) )			throw osException("osErrAssert", osErrAssert);		}
#define osThrowIfNonZero( test )	{	if ( (test) != 0 )		throw osException("osErrAssert", osErrAssert);		}
#define osThrowIfErr( test )		osThrowIfNonZero( test )
#define osThrowIfNil( test )		osThrowIfZero( test )
#define osThrowIfMemErr( test )		{	if ( (test) == 0 )		throw osException("osErrMemory", osErrMemory);		}


#endif	// _OSDEBUG_H_




///////////////////////////////// END OF FILE /////////////////////////////////

