<?lassoscript

	// [Encode_QuotedPrintable] - Encodes email bodies using quoted printable algorithm.
	//
	// -Data = (string, the text to be encoded)
	// -Charset = (optional, desired character set after decoding, defaults to 'utf-8')
	//
	// The tag encodes any characters within the ASCII range 0x20 to 0x7e except 0x5f and 0x3d
	// (that is all standard ASCII characters from space to tilde except for underscore and the
	// equal sign).  Lines are automaticaly truncated to less than 76 characters and joined with
	// \r\n.
	//

	define_tag: 'quotedprintable',
		-priority = 'replace',
		-required = 'data',
		-optional = 'charset',
		-namespace = 'encode_';

		if: (local_defined: 'charset') && (#charset != 'utf-8');
			fail_if: (#charset != 'iso-8859-1') && !(string_validcharset: #charset), -9956, '[Encode_QuotedPrintable] Expected a valid character set but got ' + #charset + '.';
			local: 'bytes' = (bytes: #data, #charset);
		else;	
			local: 'bytes' = (bytes: #data);
		/if;
		#bytes = (encode_quotedprintablebytes: #bytes);		
		return: (string: #bytes, 'iso-8859-1');

	/define_tag;


	// [Decode_QuotedPrintable] - Decodes email bodies encoded using quoted printable algorithm.
	//
	// -Data = (string, the text to be decoded)
	// -Charset = (optional, character set of encoded text, defaults to 'utf-8')
	//
	
	define_tag: 'quotedprintable',
		-priority = 'replace',
		-required = 'data',
		-optional = 'charset',
		-namespace = 'decode_';

		if: !(local_defined: 'charset');
			local: '_charset' = 'utf-8';
		else: (#charset == 'utf-8') || (#charset='iso-8859-1');
			local: '_charset' = #charset;
		else: !(string_validcharset: #charset);
			local: '_charset' = 'iso-8859-1';
		else;
			local: '_charset' = #charset;
		/if;
		local: 'bytes' = (bytes: #data);
		#bytes = (decode_quotedprintablebytes: #bytes);
		return: (string: #bytes, #_charset);

	/define_tag;


	// [Encode_QHeader] - Encodes email headers using Q (quoted printable) algorithm
	//
	// -Name = (string, the name of the header to be encoded)
	// -Value = (string, the value of the header to be encoded)
	// -Charset = (optional, desired character set after decoding, defaults to 'utf-8')
	//
	// The tag encodes any characters within the ASCII range 0x20 to 0x7e except 0x5f and 0x3d
	// (that is all standard ASCII characters from space to tilde except for underscore and the
	// equal sign).  Lines are automaticaly truncated to less than 76 characters and joined with
	// \r\n.
	//
	// The output is =?charset?Q?text?= with any characters that require encoding changed to =xx 
	// entities within the text.  If no encoding is required then the charset and Q prefix are 
	// not included.  Individual words are encoded.  Maxlength specifies the maximum length of header
	// lines.  With no maxlength no header folding will be performed.
	//

	define_tag:'qheader',
		-Priority='Replace',
		-required='name', -type='string',
		-required='value', -type='string',
		-optional='charset',
		-optional='maxlength',
		-namespace = 'encode_';

		local: '_name' = #name;
		#_name->trim;
		#_name->(removetrailing: ':');

		local: '_value' = #value;
		#_value->trim;
		#_value = (string_replaceregexp: #_value, -find='[\r\n\t]+', -replace=' ');
		
		local: '_max' = (integer: (local: 'maxlength'));
		
		if: ( #_value->(beginswith: '=?') && #_value->(endswith: '?=') );
			return: #_name + ': ' + #_value;
		/if;

		if: (local_defined: 'charset') && (#charset == 'none');
			local: 'bytes' = (bytes: #_value);
			local: 'prefix' = '';
			local: 'suffix' = '';
		else: (local_defined: 'charset') && (#charset != 'utf-8');
			fail_if: (#charset != 'iso-8859-1') && !(string_validcharset: #charset), -9956, '[Encode_QHeader] Expected a valid character set but got ' + #charset + '.';
			local: 'array' = #_value->(split: ' ');
			local: 'bytes' = bytes;
			iterate: #array, (local: 'temp');
				loop_count > 1 ? #bytes += ' ';
				local: 'temp2' = #temp;
				#temp2->(removetrailing: ',');
				if: (#temp2->(beginswith: '<') && #temp2->(endswith: '>')) || ((#temp2 >> '@') && (valid_email: #temp2));
					#bytes += (bytes: #temp, #charset);
				else;
					#bytes += (encode_quotedprintablebytes: (bytes: #temp, #charset));
				/if;
			/iterate;
			local: 'prefix' = '=?' + #charset + '?Q?';
			local: 'suffix' = '_?=';
		else;	
			local: 'array' = #_value->(split: ' ');
			local: 'bytes' = bytes;
			iterate: #array, (local: 'temp');
				loop_count > 1 ? #bytes += ' ';
				local: 'temp2' = #temp;
				#temp2->(removetrailing: ',');
				if: (#temp2->(beginswith: '<') && #temp2->(endswith: '>')) || ((#temp2 >> '@') && (valid_email: #temp2));
					#bytes += (bytes: #temp);
				else;
					#bytes += (encode_quotedprintablebytes: (bytes: #temp));
				/if;
			/iterate;
			local: 'prefix' = '=?utf-8?Q?';
			local: 'suffix' = '_?=';
		/if;

		local: '_value' = (string: #bytes, 'iso-8859-1');
	
		// Strip out unnecessary encoding for headers
		#_value->(replace: '=66', 'f');
		#_value->(replace: '=46', 'F');
		#_value->(replace: '=2E', '.');
		#_value->(replace: '=\r\n', '');

		// check line length, can not exceed maxlength chars per line
		if: #_max <= 0;
			local: '_max' = 998;
		/if;
		local: 'max_length' = #_max - (#_name->size + 2) - #prefix->size - #suffix->size;
		local: 'array' = (string_findregexp: #_value, -find='[^ ]{0,' + (#max_length - 6) + '}(?:[^ =]{2})? ?');
		
		local: '_value' = #_name + ': ';
		local: 'length' = #_value->size;
		iterate: #array, (local: 'temp');
			local: 'add' = '';
			if: (#temp >> '=');
				local: 'endfix' = '';
				while: #temp->(beginswith: '(') || #temp->(beginswith: '"');
					#add += #temp->(get: 1);
					#temp = (_substring: #temp, 2);
				/while;
				while: #temp->(endswith: ')') || #temp->(endswith: '"') || #temp->(endswith: ' ');
					#endfix = #temp->(get: #temp->size) + #endfix;
					#temp = (_substring: #temp, 1, #temp->size - 1);
				/while;
				#add += #prefix + #temp + #suffix + #endfix;
			else;
				#add += #temp;
			/if;
			#length += #add->size;
			if: #length <= #max_length;
				#_value += #add;
			else;
				#_value += '\r\n ' + #add;
				local: 'length' = #add->size + 1;
				local: 'max_length' = #_max;
			/if;
		/iterate;		
		return: @#_value;

	/define_tag;


	// [Decode_QHeader] - Decodes email header values using Q (quoted printable) algorithm
	//
	// -Data = (string, the text to be decoded)
	//

	define_tag:'qheader',
		-Priority='Replace',
		-required='data', -type='string',
		-namespace = 'decode_';

		local: 'output' = (string_replaceregexp: #data, -find='\r\n\\s+', -replace='');
		
		if: (#data !>> '=?') || (#data !>> '?q?') || (#data !>> '?=');
			return: @#output;
		/if;
		
		local: 'output' = (string_replaceregexp: #output, -find='\\?=\\s+=\\?', -replace='?= =?');
		local: 'parse' = (string_findregexp: #output, -find='=\\?([a-zA-Z0-9_-]+?)\\?Q\\?(.*?)\\?= ?');

		if: #parse->size == 0;
			return: @#output;
		/if;
		
		loop: -from=1, -to=#parse->size, -by=3;
			local: 'token' = #parse->(get: loop_count);
			local: 'charset' = #parse->(get: loop_count + 1);
			local: 'value' = #parse->(get: loop_count + 2);

			if: (string_validcharset: #charset);
				local: 'newvalue' = (decode_quotedprintable: (string_replace: #value, -find='_', -replace=' '), #charset);
				#output->(replace: #token, #newvalue);
			/if;
		/loop;
		
		return: @#output;

	/define_tag;


	// [Encode_BHeader] - Encodes email header values using B (base64) algorithm
	//
	// -Name = (string, the name of the header)
	// -Value = (string, the value of the header)
	// -Charset = (string, optional charset, defaults to utf-8)
	// -MaxLength = (integer, maximum line length)

	define_tag:'bheader',
		-Priority='Replace',
		-required='name', -type='string',
		-required='value', -type='string',
		-optional='charset',
		-optional='maxlength',
		-namespace = 'encode_';

		local: '_name' = #name;
		#_name->trim;
		#_name->(removetrailing: ':');

		local: '_value' = #value;
		#_value->trim;
		#_value = (string_replaceregexp: #_value, -find='[\r\n\t]', -replace=' ');
		
		local: '_max' = (integer: (local: 'maxlength'));
		
		if: ( #_value->(beginswith: '=?') && #_value->(endswith: '?=') );
			return: #_name + ': ' + #_value;
		/if;

		if: (local_defined: 'charset') && (#charset == 'none') || ((string_findregexp: -find = '[^\\x09\\x0A\\x0D\\x20-\\x7E]', #_value)->size == 0);
			local: 'bytes' = (bytes: #_value);
			local: 'prefix' = '';
			local: 'suffix' = '';
		else: (local_defined: 'charset') && (#charset != 'utf-8');
			fail_if: (#charset != 'iso-8859-1') && !(string_validcharset: #charset), -9956, '[Encode_BHeader] Expected a valid character set but got ' + #charset + '.';
			local: 'bytes' = (encode_base64: (bytes: #_value, #charset));
			local: 'prefix' = '=?' + #charset + '?B?';
			local: 'suffix' = '_?=';
		else;	
			local: 'bytes' = (encode_base64: (bytes: #_value));
			local: 'prefix' = '=?utf-8?B?';
			local: 'suffix' = '?=';
		/if;

		local: '_value' = (string: #bytes, 'iso-8859-1');

		// check line length, can not exceed maxlength chars per line
		if: #_max <= 0;
			local: '_max' = 998;
		/if;
		local: 'max_length' = #_max - (#_name->size + 2) - #prefix->size - #suffix->size;
		local: 'array' = (string_findregexp: #_value, -find='[^ ]{0,' + (#max_length - 6) + '} ?');
		
		if: #prefix != '';
			local: '_value' = #_name + ': ' + #prefix;
			local: 'size' = #array->size;
			iterate: #array, (local: 'temp');
				#_value += #temp + ((loop_count == #size) ? #suffix | '\r\n ');
			/iterate;
		else;
			local: '_value' = #_name + ': ';
			local: 'length' = #_value->size;
			iterate: #array, (local: 'temp');
				#length += #temp->size;
				if: #length <= #max_length;
					#_value += #temp;
				else;
					#_value += '\r\n ' + #temp;
					local: 'length' = #temp->size + 1;
					local: 'max_length' = #_max;
				/if;
			/iterate;		
		/if;
		return: @#_value;

	/define_tag;
	
	// [Decode_BHeader] - Decodes email header values using B (base64) algorithm
	//
	// -Data = (string, the text to be decoded)
	//

	define_tag:'bheader',
		-Priority='Replace',
		-required='data', -type='string',
		-namespace = 'decode_';

		local: 'output' = (string_replaceregexp: #data, -find='\r\n\\s+', -replace='');
		
		if: (#data !>> '=?') || (#data !>> '?b?') || (#data !>> '?=');
			return: @#output;
		/if;
		
		local: 'output' = (string_replaceregexp: #output, -find='\\?=\\s+=\\?', -replace='?= =?');
		local: 'parse' = (string_findregexp: #output, -find='=\\?([a-zA-Z0-9_-]+?)\\?B\\?(.*?)\\?= ?');

		if: #parse->size == 0;
			return: @#output;
		/if;
		
		loop: -from=1, -to=#parse->size, -by=3;
			local: 'token' = #parse->(get: loop_count);
			local: 'charset' = #parse->(get: loop_count + 1);
			local: 'value' = #parse->(get: loop_count + 2);

			if: (string_validcharset: #charset);
				local: 'newvalue' = (bytes: (decode_base64: #value), #charset);
				#output->(replace: #token, #newvalue);
			/if;
		/loop;
		
		return: @#output;

	/define_tag;

?>
