<?LassoScript

// email_pop - POP Mail Downloader for Lasso Professional 8
//
// Quick Start - This type is designed to be used as a high level POP interface.  The basic operation is:
//
// [var: 'pop' = (email_pop: -host='mail.example.com', -username='username', -password='', -apop=true, -log=true)]
// [iterate: $pop, (var: 'msg')]
//   [var: 'msg']
//   [$pop->(delete)]
// [/iterate]
// [output: $pop->(close)]
//
// High-Level Commands - These are the recommended tags for most users
//
// [email_pop:
//   -host='mail.example.com',
//   -port=110, (optional)
//   -username='username', (if a username and password are specifed then
//   -password='password', the type will auto-authorize the POP server)
//   -apop=true, (optional, APOP will be tried before USER/PASS)
//   -log=false, (optional, log all actions to the console)
//   -timeout=15 (optional, timeout for al reads and writes 
// ]
// [email_pop->authorize:
//   -username='username',
//   -password='password',
//   -apop=true, (optional, APOP will be tried before USER/PASS)
// ]
// [email_pop->size] returns the number of messages availble for download
// [email_pop->(get: x)] returns the specified message from the server
// [email_pop->(delete: x)] deletes the specified message from the server
// [email_pop->(close)] closes the POP connection, performing any specified deletes
// [email_pop->(cancel)] closes the POP connection, but does not perform any deletes
//
// The size and get tags allow the iterate code above to be used.  The delete tag
// will delete the last message which was specified for get if no parameter is passed.
//
// Low-Level Commands - These tags allow access to the raw POP commands
//
// [email_pop->(USER: 'username')] sends a username to the server
// [email_pop->(PASS: 'username')] sends a password to the server
// [email_pop->(APOP: 'username','digest')] sends APOP authentication to the server
// [email_pop->(QUIT)] closes the connection to the server
// [email_pop->(RSET)] resets the server, cancelling any deletes
// [email_pop->(STAT] returns the status of the server
// [email_pop->(LIST)] returns a list of index and sizes for all messages on server
// [email_pop->(LIST: x)] returns the index and size for message x.
// [email_pop->(RETR: x)] retrieves message x from the server
// [email_pop->(TOP: x, y)] fetches message x from the server, getting the headers and y lines
// [email_pop->(DELE: x)] deletes message x from the server
// [email_pop->(NOOP)] a null operation which can be used to keep a connection open.
//
// Implementation - These tags are used in the implementation and shouldn't be accessed directly
//
// [email_pop->(pop_cmd:
//   -command='POP command',
//   -timeout=15, (optional timeout for this command)
//   -multi=false (if true a long response to the command (ended with .) is collected
// ]
// [email_pop->(pop_err:
//   -msg='error message',
//   -code=-1, (optional, result code)
//   -cmd='POP command' (optional)
// ]
// [email_pop->(pop_log:
//   -msg='log message',
//   -code=0, (optional, result code)
//   -cmd='POP command' (optional)
// ]
//

define_type: 'pop',
	-namespace = 'email_';
	
	local: 'pop_net' = null;
	local: 'pop_mode' = '';	// init,auth,live,done
	local: 'pop_timeout' = 15;
	local: 'pop_token' = '';
	local: 'pop_index' = 0;
	local: 'pop_err' = (array);
	local: 'pop_res' = (array);
	local: 'pop_log' = false;
	local: 'pop_debug' = false;
	local: 'pop_get' = 'retr';
	local: 'pop_ids' = map;
	local: 'pop_capa' = null;
	local: 'pop_server' = '';
	
	// Call Back Tags
	
	define_tag: 'onCreate',
			-optional='server',
			-optional='port',
			-optional='username',
			-optional='password',
			-optional='apop',
			-optional='timeout',
			-optional='log',
			-optional='debug',
			-optional='get',
			-optional='host',
			-optional='ssl';

		self->'pop_net' = (net);
		
		if: (local_defined: 'get');
			self->'pop_get' = (string: #get);
		else;
			self->'pop_get' = 'uidl';
		/if;
		
		if: (local_defined: 'log');
			self->'pop_log' = (#log === null) || (boolean: #log);
		/if;

		if: (local_defined: 'debug');
			self->'pop_debug' = (#debug === null) || (boolean: #debug);
		/if;

		if: (local_defined: 'apop');
			local: '_apop' = (#apop === null) || (boolean: #apop);
		else;
			local: '_apop' = true;
		/if;

		if: (local_defined: 'ssl');
			local: '_ssl' = (#ssl === null) || (boolean: #ssl);
		else;
			local: '_ssl' = false;
		/if;

		if: (local_defined: 'server');
			local: '_server' = (string: #server);
			#_server->trim;
		else: (local_defined: 'host');
			local: '_server' = (string: #host);
			#_server->trim;
		/if;
		
		self->'pop_server' = #_server;
		
		if: (local_defined: 'port');
			local: '_port' = (integer: #port);
		else;
			local: '_port' = 110;
		/if;

		local: '_timeout' = (integer: self->'pop_timeout');
		if: (local_defined: 'timeout');
			#_timeout = (integer: #timeout);
			if: (#_timeout < 5);
				local: '_timeout' = 5;
			/if;
			self->'pop_timeout' = (integer: #_timeout);
		/if;

		self->'pop_net'->(connect: #_server, #_port);

		if: (self->'pop_net'->(wait: #_timeout, net_waitread) == net_waitread);
			local: 'result' = self->'pop_net'->(read: 32768);
			protect;
				self->'pop_token' = (string_findregexp: #result, -find='<[^>]+>')->(get:1);
			/protect;
			self->(pop_log: -msg=#result->(split: '\r\n')->(get:1));
		else;
			self->(pop_err: -msg='The remote server timed out on connect.');
		/if;
		
		local: 'capa' = self->capabilities;
		if: (#_ssl == true);
			if: (#capa >> 'STLS');
				self->stls;
				local: 'done' = (self->'pop_res'->last->second >> '+OK');
				self->'pop_net'->(SetType: Net_TypeSSL);
			else;
				self->(pop_err: -msg='Could not open SSL communications',-cmd='STLS');
			/if;
		/if;

		if: (local_defined: 'username') && (local_defined: 'password');
			self->(authorize: -username=(local: 'username'), -password=(local: 'password'), -apop=(local: '_apop'));
		/if;
			
	/define_tag;
	
	define_tag: 'onDestroy';

		protect;
			if: (self->'pop_mode' != 'done');
				self->quit;
			/if;
			
			handle: true;
				if: (self->'pop_net'->(isa: 'net'));
					self->'pop_net'->close;
					self->'pop_net' = null;
				/if;
			/handle;
		/protect;
		
	/define_tag;


	define_tag: 'onConvert',
			-required='type';
	
	/define_tag;
	

	// Communication Commands

	define_tag: 'pop_cmd', 
			-required='command',
			-optional='timeout',
			-optional='multi',
			-optional='until';

		local: 'result' = '';
		if: (self->'pop_net'->(isa: 'net'));
			
			local: '_command' = (string: #command);
			if: !(#_command->(endswith: '\r\n'));
				#_command += '\r\n';
			/if;
			
			local: '_timeout' = (integer: self->'pop_timeout');
			if: (local_defined: 'timeout');
				#_timeout = (integer: #timeout);
			/if;
			
			local: '_multi' = (boolean: (local: 'multi'));
			local: '_until' = (string: (local: 'until'));
			
			if: (self->'pop_net'->(wait: #_timeout, net_waitwrite) == net_waitwrite);
				local: 'write' = (integer: self->'pop_net'->(write: #_command));
				if: (self->'pop_net'->(wait: #_timeout, net_waitread) == net_waitread);
					if: (#_until != '');
						local: 'result' = '';
						while: #result !>> #_until;
							#result += self->'pop_net'->(read: 1);
							if: (loop_count > 32768);
								loop_abort;
							/if;
						/while;
						self->(pop_log: -msg=#result->(split: '\r\n')->(get:1), -cmd=#command);
					else;
						local: 'result' = '';
						while: #result !>> '\r\n';
							#result += self->'pop_net'->(read: 1);
							if: (loop_count > 32768);
								loop_abort;
							/if;
						/while;
						self->(pop_log: -msg=#result->(split: '\r\n')->(get:1), -cmd=#command);
						if: (#_multi == true) && (#result !>> '\r\n.\r\n') && (#result !>> '-ERR');
							while: (#result !>> '\r\n.\r\n');
								if: (loop_count > 10000);
									self->(pop_err: -msg='Error while reading from remote server.', -cmd=#command);
									loop_abort;
								/if;
								local: 'temp' = self->'pop_net'->(read: 32768);
								self->(pop_log: -msg='... reading (' + #temp->size + ' bytes)', -cmd=#command);
								if: (#temp->size == 0);
									loop_abort;
								/if;
								#result += #temp;
							/while;
						/if;
					/if;
				else;
					self->(pop_err: -msg='The remote server timed out on read.', -cmd=#command);
				/if;
			else;
				self->(pop_err: -msg='The remote server timed out on write.', -cmd=#command);
			/if;
		else;
			self->(pop_err: -msg='Lost network connection unexpectedly.', -cmd=#command);
		/if;
		return: #result;
	
	/define_tag;

	define_tag: 'pop_err',
			-required='msg',
			-optional='code',
			-optional='cmd';
		
		if: (local_defined: 'cmd');
			local: '_cmd' = (string: #cmd);
		else;
			local: '_cmd' = '(none)';
		/if;
		
		if: (local_defined: 'code');
			local: '_code' = (integer: #code);
		else;
			local: '_code' = -1;
		/if;
		
		if: (self->'pop_err'->size > 0) && (self->'pop_err'->last == null);
			self->'pop_err'->remove;
		/if;
			
		if: (self->'pop_debug');
			self->'pop_err'->(insert: #_cmd=#msg);
			self->'pop_res'->(insert: #_cmd=#msg);
		else;
			self->'pop_err' = (array: #_cmd=#msg);
			self->'pop_res' = (array: #_cmd=#msg);
		/if;
		
		if: (self->'pop_log');
			log_detail: 'POP: ' + #_cmd + ' - ' + #msg;
		/if;
		
		fail: #_code, #msg;
		
	/define_tag;
	
	define_tag: 'pop_log',
			-required='msg',
			-optional='code',
			-optional='cmd';

		if: (local_defined: 'cmd');
			local: '_cmd' = (string: #cmd);
		else;
			local: '_cmd' = '(none)';
		/if;
		
		if: (local_defined: 'code');
			local: '_code' = (integer: #code);
		else;
			local: '_code' = -1;
		/if;
		
		if: (self->'pop_err'->size > 0) && (self->'pop_err'->last != null);
			if: (self->'pop_debug');
				self->'pop_err'->(insert: null);
			else;
				self->'pop_err' = (array: null);
			/if;
		/if;
			
		if: (self->'pop_debug');
			self->'pop_res'->(insert: #_cmd=#msg);
		else;
			self->'pop_res' = (array: #_cmd=#msg);
		/if;
		
		if: (self->'pop_log');
			log_detail: 'POP: ' + #_cmd + ' - ' + #msg;
		/if;
		
		error_seterrormessage: (error_noerror);
		error_seterrorcode: (error_noerror: -errorcode);
		
	/define_tag;
	
	
	// Low Level Commands

	define_tag: 'user',
			-optional='username';
		if: (local_defined: 'username');
			return: self->(pop_cmd: 'USER ' + #username);
		else;
			self->(pop_err: -msg='No username specified.',-cmd='USER');
		/if;
	/define_tag;

	define_tag: 'pass',
			-optional='password';
		if: (local_defined: 'password');
			return: self->(pop_cmd: 'PASS ' + #password);
		else;
			self->(pop_err: -msg='No password specified.',-cmd='PASS');
		/if;
	/define_tag;

	define_tag: 'apop',
			-required='username',
			-required='password';
		if: (self->'pop_token' >> '<');
			local: 'digest' = (encrypt_md5: self->'pop_token' + #password);
			return: self->(pop_cmd: 'APOP ' + #username + ' ' + #digest);
		else;
			self->(pop_err: -msg='No token found.',-cmd='APOP');
		/if;
	/define_tag;

	define_tag: 'auth',
			-required='username',
			-required='password',
			-required='method';
		if: #method == 'CRAM-MD5';
			local: 'result' = self->(pop_cmd: 'AUTH CRAM-MD5');
			if: #result->(beginswith: '+ ');
				#result = (_substring: #result, 3);
				local: 'token' = (string: (decode_base64: #result));
				local: 'digest' = (encode_base64: #username + ' ' + (encrypt_crammd5: #password, #token));
				return: self->(pop_cmd: #digest);
			else;	
				self->(pop_err: -msg='Invalid response for ' + #method + ' method.',-cmd='AUTH');
			/if;
		else: #method == 'DIGEST-MD5';
			local: 'result' = self->(pop_cmd: 'AUTH DIGEST-MD5');
			if: #result->(beginswith: '+ ');
				#result = (_substring: #result, 3);
				local: 'digest' = (email_digestchallenge: (decode_base64: #result));
				#digest->(insert: 'username' = #username);
				#digest->(insert: 'password' = #password);
				#digest->(insert: 'digest-uri' = 'pop/' + self->'pop_server');
				#digest->(insert: 'cnonce' = (encrypt_md5: lasso_uniqueid));
				#digest->(insert: 'nc' = '00000001');
				local: 'response' = (encode_base64: (email_digestresponse: #digest));
				return: self->(pop_cmd: #response);
			else;	
				self->(pop_err: -msg='Invalid response for ' + #method + ' method.',-cmd='AUTH');
			/if;
		else;
			self->(pop_err: -msg='Method ' + #method + ' not found.',-cmd='AUTH');
		/if;

			local: 'digest' = (encrypt_md5: self->'pop_token' + #password);
			return: self->(pop_cmd: 'APOP ' + #username + ' ' + #digest);
	/define_tag;

	define_tag: 'quit';
		self->'pop_mode' = 'done';
		return: self->(pop_cmd: 'QUIT')->trim &;
	/define_tag;

	define_tag: 'rset';
		return: self->(pop_cmd: 'RSET')->trim &;
	/define_tag;

	define_tag: 'stat';
		return: self->(pop_cmd: 'STAT')->trim &;
	/define_tag;

	define_tag: 'list',
			-optional='msg';
		if: (local_defined: 'msg');
			return: self->(pop_cmd: 'LIST ' + (integer: #msg))->trim &;
		else;
			return: self->(pop_cmd: 'LIST', -multi=true)->trim &;
		/if;
	/define_tag;

	define_tag: 'uidl',
			-optional='msg';
		if: (local_defined: 'msg');
			local: 'result' = 
			return: self->(pop_cmd: 'UIDL ' + (integer: #msg))->trim &;
		else;
			return: self->(pop_cmd: 'UIDL', -multi=true)->trim &;
		/if;
	/define_tag;

	define_tag: 'retr',
			-required='msg';
		return: self->(pop_cmd: 'RETR ' + (integer: #msg), -multi=true)->trim &;
	/define_tag;

	define_tag: 'top',
			-required='msg',
			-optional='size';
		return: self->(pop_cmd: 'TOP ' + (integer: #msg) + ' ' + (integer: (local: 'size')), -multi=true)->trim &;
	/define_tag;

	define_tag: 'dele',
			-required='msg';
		return: self->(pop_cmd: 'DELE ' + (integer: #msg))->trim &;
	/define_tag;

	define_tag: 'noop';
		return: self->(pop_cmd: 'NOOP')->trim &;
	/define_tag;

	define_tag: 'capa';
		return: self->(pop_cmd: 'CAPA', -multi=true)->trim &;
	/define_tag;

	define_tag: 'stls';
		return: self->(pop_cmd: 'STLS', -until='\r\n')->trim &;
	/define_tag;


	// High Level Commands
	
	define_tag: 'authorize',
			-required='username',
			-required='password',
			-optional='apop';

		local: 'done' = false;
		if: (local_defined: 'apop');
			local: '_apop' = (#apop === null) || (boolean: #apop);
		else;
			local: '_apop' = true;
		/if;
		if: self->'pop_capa' >> 'sasl';
			local: 'methods' = self->'pop_capa'->(find: 'sasl');
		else;
			local: 'methods' = (array: 'LOGIN','PLAIN');
		/if;
		if: (#done == false) && (#methods >> 'DIGEST-MD5');
			self->(auth: -username=#username, -password=#password, -method='DIGEST-MD5');
			local: 'done' = (self->'pop_res'->last->second >> '+ ');
		/if;
		if: (#done == false) && (#methods >> 'CRAM-MD5');
			self->(auth: -username=#username, -password=#password, -method='CRAM-MD5');
			local: 'done' = (self->'pop_res'->last->second >> '+OK');
		/if;
		if: (#done == false) && (#_apop && (self->'pop_token' >> '<'));
			self->(apop: -username=#username, -password=#password);
			local: 'done' = (self->'pop_res'->last->second >> '+OK');
		/if;
		if: (#done == false);
			self->(user: #username);
			self->(pass: #password);
			local: 'done' = (self->'pop_res'->last->second >> '+OK');
		/if;
		if: #done == false;
			self->(pop_err: -msg='Could not authenticate user.',-cmd='PASS');
		else;
			self->'pop_mode' = 'ready';
		/if;
		self->'pop_ids' = map;
		self->'pop_index' = 0;
	/define_tag;

	define_tag: 'size';
		protect;
			return: (integer: (string_findregexp: self->(pop_cmd: 'STAT'), -find='[0-9]+')->(get:1));
		/protect;
		return: 0;
	/define_tag;
	
	define_tag: 'get',
			-optional='msg';
		if: (local_defined: 'msg') && !(string_isdigit: #msg);
			if: (self->'pop_ids' >> #msg);
				local: 'dex' = self->'pop_ids'->(find: #msg);
			else;
  				fail_if: (local_defined: 'msg'), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
			/if;
		else;
			local: 'dex' = (integer: (local: 'msg'));
		/if;
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = #dex;
		if: #dex > 0;
			if: (self->'pop_get' >> 'retr');
				return: self->(retrieve: #dex);
			else: (self->'pop_get' >> 'dele');
				return: self->(delete: #dex);
			else: (self->'pop_get' >> 'head');
				return: self->(headers: #dex);
			else: (self->'pop_get' >> 'top');
				local: 'size' = (integer: (string_replaceregexp: self->'pop_get', -find='.*(\\d+).*', -replace='\\1'));
				return: self->(retrieve: #dex, #size);
			else;
				return: self->(uniqueid: #dex);
			/if;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'retrieve',
			-optional='msg',
			-optional='size';
  		fail_if: (local_defined: 'size') && !(string_isdigit: #size), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer size.';
		if: (local_defined: 'msg') && !(string_isdigit: #msg);
			if: (self->'pop_ids' >> #msg);
				local: 'dex' = self->'pop_ids'->(find: #msg);
			else;
  				fail_if: (local_defined: 'msg'), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
			/if;
		else;
			local: 'dex' = (integer: (local: 'msg'));
		/if;
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = #dex;
		if: #dex > 0;
			if: (local_defined: 'size');
				local: 'result' = self->(top: #dex, (integer: #size));
			else;
				local: 'result' = self->(retr: #dex);
			/if;
			fail_if: (#result->(beginswith: '-ERR')), -9956, 'The tag [Email_POP->' + tag_name + '] did not find a message at provided position (' + #dex ').';
			local: 'index' = #result->(find: '\r\n');
			if: (#index > 0);
				local: 'result' = (_substring: #result, #index + 2);
			/if;
			#result->trim;
			#result->(removetrailing: '\r\n.');
			return: #result;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'headers',
			-optional='msg';
		if: (local_defined: 'msg') && !(string_isdigit: #msg);
			if: (self->'pop_ids' >> #msg);
				local: 'dex' = self->'pop_ids'->(find: #msg);
			else;
  				fail_if: (local_defined: 'msg'), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
			/if;
		else;
			local: 'dex' = (integer: (local: 'msg'));
		/if;
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = #dex;
		if: #dex > 0;
			local: 'result' = self->(top: #dex, 0);
			fail_if: (#result->(beginswith: '-ERR')), -9956, 'The tag [Email_POP->' + tag_name + '] did not find a message at provided position (' + #dex ').';
			local: 'index' = #result->(find: '\r\n');
			if: (#index > 0);
				local: 'result' = (_substring: #result, #index + 2);
			/if;
			#result->trim;
			#result->(removetrailing: '\r\n.');
			return: #result;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'delete',
			-optional='msg';
		if: (local_defined: 'msg') && !(string_isdigit: #msg);
			if: (self->'pop_ids' >> #msg);
				local: 'dex' = self->'pop_ids'->(find: #msg);
			else;
  				fail_if: (local_defined: 'msg'), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
			/if;
		else;
			local: 'dex' = (integer: (local: 'msg'));
		/if;
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = 0;
		if: #dex > 0;
			local: 'result' = self->(dele: #dex);
			fail_if: (#result->(beginswith: '-ERR')), -9956, 'The tag [Email_POP->' + tag_name + '] did not find a message at provided position (' + #dex ').';
		/if;
	/define_tag;
	
	define_tag: 'uniqueid',
			-optional='msg';
  		fail_if: (local_defined: 'msg') && !(string_isdigit: #msg), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
		local: 'dex' = (integer: (local: 'msg'));
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = #dex;
		if: #dex > 0;
			local: 'result' = self->(uidl: #dex);
			fail_if: (#result->(beginswith: '-ERR')), -9956, 'The tag [Email_POP->' + tag_name + '] did not find a message at provided position (' + #dex ').';
			local: 'array' = #result->(split: ' ');
			if: (#array->size > 2);
				local: 'id' = #array->(get: 3);
				if: !(string_isdigit: #id);
					self->'pop_ids'->(insert: #id = #dex);
				/if;
				return: @#id;
			/if;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'capabilities';
		if: self->'pop_capa' === null;
			self->'pop_capa' = map;
			local: 'results' = self->capa;
			if: #results >> '\r\n';
				local: 'array' = #results->(split: '\r\n');
			else: #results >> '\r';
				local: 'array' = #results->(split: '\r');
			else: #results >> '\n';
				local: 'array' = #results->(split: '\n');
			else;
				local: 'array' = array;
			/if;
			iterate: #array, (local: 'temp');
				if: (#temp->size > 0) && (#temp->(beginswith:'+') == false) && (#temp->(beginswith:'.') == false);
					if: #temp >> ' ';
						local: 'split' = #temp->(split: ' ');
						local: 'name' = #split->first;
						#split->(remove: 1);
						self->'pop_capa'->(insert: #name = #split);
					else;
						self->'pop_capa'->(insert: #temp = '');
					/if;
				/if;
			/iterate;
		/if;
		return: self->'pop_capa';
	/define_tag;

	define_tag: 'close';
		self->'pop_ids' = map;
		self->'pop_index' = 0;
		self->(quit);
	/define_tag;
	
	define_tag: 'cancel';
		self->'pop_ids' = map;
		self->'pop_index' = 0;
		self->(rset);
		self->(quit);
	/define_tag;
	
	define_tag: 'results';
		return: self->'pop_res';
	/define_tag;
	
	define_tag: 'errors';
		return: self->'pop_err';
	/define_tag;
	
	define_tag: 'lasterror';
		return: self->'pop_err'->last;
	/define_tag;
			
	define_tag: 'data';
		return: self->'pop_token';
	/define_tag;
	
/define_type;

	
	// email_parse
	
define_type: 'parse',
	-namespace = 'email_', -prototype;

	local: 'parse_msg';
	local: 'parse_rawhdrs';
	local: 'parse_hdrs';
	local: 'parse_body';
	local: 'parse_content_type';
	local: 'parse_content_transfer_encoding';
	local: 'parse_content_disposition';
	local: 'parse_charset';
	local: 'parse_boundary';
	local: 'parse_parts';
	local: 'parse_mode';

	define_tag: 'onCreate',
			-required='msg';
		
		fail_if: !(#msg->(isa: 'string') || #msg->(isa: 'bytes') || #msg->(isa: 'compose')), -9956, '[Email_Parse] expects a string or bytes parameter.';
		fail_if: (#msg == ''), -9956, '[Email_Parse] expects a non-empty email message to parse.';
		
		if: #msg->(isa: 'compose');
			self->'parse_msg' = (local: 'msg')->data;
		else;
			self->'parse_msg' = (local: 'msg');
		/if;
		
		// Normalize if the email doesn't have proper line endings
		local: 'line_ending' = '\r\n';
		if: (self->'parse_msg' !>> '\r\n');
			if: (self->'parse_msg' >> '\n');
				local: 'line_ending' = '\n';
			else: (self->'parse_msg' >> '\r');
				local: 'line_ending' = '\r';
			/if;
		/if;

		local: 'index' = self->'parse_msg'->(find: #line_ending + #line_ending);
		if: #index > 0;
			self->'parse_rawhdrs' = (_substring: self->'parse_msg', 1, #index);
		else;
			self->'parse_rawhdrs' = self->'parse_msg';
		/if;
		self->'parse_rawhdrs'->trim;
		self->'parse_hdrs' = (_chop_mimes: (string: self->'parse_rawhdrs', 'iso-8859-1'));
		
		if: #index >= 0;
			local: 'temp_body' = (_substring: self->'parse_msg', #index + (#line_ending->size * 2));
		else;
			local: 'temp_body' = (_substring: self->'parse_msg', #index);
		/if;

		self->'parse_content_type' = self->(header: 'content-type');
		self->'parse_content_transfer_encoding' = self->(header: 'content-transfer-encoding');
		self->'parse_content_disposition' = self->(header: 'content-disposition');

		self->'parse_mode' = self->(content_type: 'type');

		if: (self->'parse_mode' >> 'multipart');
		
			// Decompose Multi-Part Message

			self->'parse_boundary' = '--' + self->boundary;
			if: (self->'parse_boundary' != '--');
				local: 'index' = #temp_body->(find: self->'parse_boundary');
				local: 'endex' = #temp_body->(find: (self->'parse_boundary' + '--'));
				if: (#index > 0);
					if: (#index > 1);
						self->'parse_body' = (_substring: #temp_body, 1, #index - 1);
					else;
						self->'parse_body' = '';
					/if;
					if: (#endex > #index);
						#temp_body = (_substring: #temp_body, 1, #endex - 1);
					/if;
					#temp_body = (_substring: #temp_body, #index + self->'parse_boundary'->size);
					self->'parse_parts' = #temp_body->(split: (self->'parse_boundary' + #line_ending));
				else;
					self->'parse_body' = '';
					self->'parse_parts' = (array: #temp_body);
				/if;
			else;
				self->'parse_body' = '';
				self->'parse_parts' = (array: #temp_body);
			/if;
			
		else;
		
			// Simple Message
			self->'parse_body' = @#temp_body;
			self->'parse_charset' = self->charset;
			self->'parse_parts' = null;
		
		/if;
	/define_tag;
	
	define_tag: 'onConvert';
		local: 'output' = '';
		local: 'temp' = '';
		(#temp := self->date) ? #output += 'Date: ' + #temp + '\r\n';
		(#temp := self->(from: -safeemail)) ? #output += 'From: ' + #temp + '\r\n';
		(#temp := self->(to: -safeemail)) ? #output += 'To: ' + #temp + '\r\n';
		(#temp := self->(cc: -safeemail)) ? #output += 'Cc: ' + #temp + '\r\n';
		(#temp := self->(bcc: -safeemail)) ? #output += 'Bcc: ' + #temp + '\r\n';
		(#temp := self->subject) ? #output += 'Subject: ' + #temp + '\r\n';
		(#temp := self->content_type) ? #output += 'Content-Type: ' + #temp + '\r\n';
		(#temp := self->content_transfer_encoding) ? #output += 'Content-Transfer-Encoding: ' + #temp + '\r\n';
		(#temp := self->content_disposition) ? #output += 'Content-Disposition: ' + #temp + '\r\n';
		(#temp := self->'parse_parts'->size) ? #output += 'Parts: ' + #temp + '\r\n';
		(#output != '') ? #output += '\r\n';
		#output += self->body;
		return: #output;
	/define_tag;
	
	define_tag: 'raw';
		return: self->'parse_msg';
	/define_tag;
			
	define_tag: 'data';
		return: self->'parse_msg';
	/define_tag;
			
	define_tag: 'headers';
		return: self->'parse_hdrs';
	/define_tag;
	
	define_tag: 'rawheaders';
		return: self->'parse_rawhdrs';
	/define_tag;
	
	define_tag: 'header',
			-required='name',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail',
			-optional='nodecode';
			
		local: 'headers' = self->'parse_hdrs'->(find: #name);
		local: 'array' = (array);

		iterate: #headers, (local: 'temp');
			local: 'output' = #temp->second;
			if: !(local_defined: 'nodecode');
				if: (#output >> '=?') && (#output >> '?q?') && (#output >> '?=');
					local: 'output' = (decode_qheader: #output);
				/if;
				if: (#output >> '=?') && (#output >> '?b?') && (#output >> '?=');
					local: 'output' = (decode_bheader: #output);
				/if;
			/if;
			if: (local_defined: 'comment');
				local: 'output' = (email_extract: #output, -comment);
			else: (local_defined: 'extract');
				local: 'output' = (email_extract: #output);
			else: (local_defined: 'safeemail');
				local: 'output' = (email_safeemail: #output);
			/if;
			if: #output->(isa: 'array');
				#array->(merge: #output);
			else;
				#array->(insert: #output);
			/if;
		/iterate;
		
		if: (#array->size == 0);
			return: null;
		else: (#array->size == 1);
			return: #array->first;
		else: (local_defined: 'join') && (#join !== null);
			return: #array->(join: #join);
		else;
			return: #array;
		/if;
		
	/define_tag;
	
	define_tag: 'content_type',
			-optional='fld';
		if: !(local_defined: 'fld') || (#fld == '');
			return: self->'parse_content_type';
		else: (#fld == 'type');
			local: 'temp' = (string_findregexp: self->'parse_content_type', -find='[0-9a-zA-Z_\\-]+(?=/)');
			(#temp->size > 0) ? return: #temp->first;
		else: (#fld == 'subtype');
			local: 'temp' = (string_findregexp: self->'parse_content_type', -find='(?<=/)[0-9a-zA-Z_\\-]+');
			(#temp->size > 0) ? return: #temp->first;
		else;
			local: 'temp' = (string_findregexp: self->'parse_content_type', -find='(?<=' + #fld + '=")[^"]+(?=")', -ignorecase);
			(#temp->size > 0) ? return: #temp->first;
			local: 'temp' = (string_findregexp: self->'parse_content_type', -find='(?<=' + #fld + '=)[^;\r\n\t ]+', -ignorecase);
			(#temp->size > 0) ? return: #temp->first;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'content_transfer_encoding';
		return: self->'parse_content_transfer_encoding';
	/define_tag;
	
	define_tag: 'content_disposition';
		return: self->'parse_content_disposition';
	/define_tag;
	
	define_tag: 'mode';
		return: self->(content_type: 'type');
	/define_tag;
	
	define_tag: 'boundary';
		return: self->(content_type: 'boundary');
	/define_tag;
	
	define_tag: 'charset';
		return: self->(content_type: 'charset');
	/define_tag;
	
	define_tag: 'recipients',
			-optional='join';
		local: 'array' = (Email_FindEmails: self->(header: 'to') + ' ' + self->(header: 'cc') + ' ' + self->(header: 'bcc'));
		if: (local_defined: 'join') && (#join !== null);
			return: #array->(join: #join);
		else;
			return: #array;
		/if;
	/define_tag;
	
	define_tag: 'to',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail';
		if: !(local_defined: 'join');
			local: 'join' = ', ';
		/if;
		if: (local_defined: 'safeemail');
			return: self->(header: 'to', -join=#join, -safeemail);
		else: (local_defined: 'comment');
			return: self->(header: 'to', -join=#join, -comment);
		else: (local_defined: 'extract');
			return: self->(header: 'to', -join=#join, -extract);
		else;
			return: self->(header: 'to', -join=#join);
		/if;
	/define_tag;
	
	define_tag: 'from',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail';
		if: !(local_defined: 'join');
			local: 'join' = ', ';
		/if;
		if: (local_defined: 'safeemail');
			return: self->(header: 'from', -join=#join, -safeemail);
		else: (local_defined: 'comment');
			return: self->(header: 'from', -join=#join, -comment);
		else: (local_defined: 'extract');
			return: self->(header: 'from', -join=#join, -extract);
		else;
			return: self->(header: 'from', -join=#join);
		/if;
	/define_tag;
	
	define_tag: 'subject';
		return: self->(header: 'subject', -join=', ');
	/define_tag;
	
	define_tag: 'cc',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail';
		if: !(local_defined: 'join');
			local: 'join' = ', ';
		/if;
		if: (local_defined: 'safeemail');
			return: self->(header: 'cc', -join=#join, -safeemail);
		else: (local_defined: 'comment');
			return: self->(header: 'cc', -join=#join, -comment);
		else: (local_defined: 'extract');
			return: self->(header: 'cc', -join=#join, -extract);
		else;
			return: self->(header: 'cc', -join=#join);
		/if;
	/define_tag;
	
	define_tag: 'bcc',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail';
		if: !(local_defined: 'join');
			local: 'join' = ', ';
		/if;
		if: (local_defined: 'safeemail');
			return: self->(header: 'bcc', -join=#join, -safeemail);
		else: (local_defined: 'comment');
			return: self->(header: 'bcc', -join=#join, -comment);
		else: (local_defined: 'extract');
			return: self->(header: 'bcc', -join=#join, -extract);
		else;
			return: self->(header: 'bcc', -join=#join);
		/if;
	/define_tag;
	
	define_tag: 'date';
		local: 'formats' = (array: 
			'%a %-d %b %Y %-H:%-M:%-S %z',
			'%a , %b %-d %Y %H:%M:%S %z',
			'%-d %b %Y %T %z',
			'%_d %b %Y %T %z',
			'%a, %_d %b %Y %H:%M:%S %z',
			'%a, %_d %b %Y %H:%M:%S %Z',
			'%-d-%b-%Y %T %z',
			'%_d-%b-%Y %T %z');
		local: 'date' = self->(header: 'date');
		iterate: #formats, (local: 'format');
			protect;
				local: 'output' = (date_gmttolocal: (date: #date, -format=#format));
				if: (#output != '');
					return: @#output;
				/if;
			/protect;
		/iterate;
		protect;
			local: 'output' = (date_gmttolocal: (date: #date));
			if: (#output != '');
				return: @#output;
			/if;
		/protect;
		return: @#date;
	/define_tag;
	
	define_tag: 'body',
			-optional='type',
			-optional='preamble',
			-optional='array';
		if: (self->'parse_parts'->size == 0) || ((local_defined: 'preamble') && (#preamble !== false));
		
			local: 'output' = self->'parse_body';
			if: (self->'parse_content_transfer_encoding' >> 'quoted-printable');
				local: 'output' = (decode_quotedprintable: (string: #output), self->'parse_charset');
			else: (self->'parse_content_transfer_encoding' >> 'base64');
				local: 'output' = (decode_base64: #output);
			else: #output->(isa: 'bytes') && (string_validcharset: self->'parse_charset');
				local: 'output' = #output->(exportstring: self->'parse_charset');
			/if;
			if: (self->'parse_content_type' >> 'text/enriched');
				// Try to convert text/enriched to HTML
				#output = (string_replaceregexp: #output, -find='<(/?)bold>', -replace='<\\1b>');
				#output = (string_replaceregexp: #output, -find='<(/?)italic>', -replace='<\\1i>');
				#output = (string_replaceregexp: #output, -find='<(/?)excerpt>', -replace='<\\1quote>');
				#output = (string_replaceregexp: #output, -find='<bigger>', -replace='<font size="+1">');
				#output = (string_replaceregexp: #output, -find='<smaller>', -replace='<font size="-1">');
				#output = (string_replaceregexp: #output, -find='<fontfamily><param>(.*?)</param>', -replace='<font face="\\1">');
				#output = (string_replaceregexp: #output, -find='<color><param>..(..),..(..),..(..)</param>', -replace='<font color="#\\1\\2\\3">');
				#output = (string_replaceregexp: #output, -find='</(?:smaller|bigger|fontfamily|color)>', -replace='</font>');
			/if;
			return: @#output;
			
		else: (self->'parse_parts'->size == 1);
		
			return: (email_parse: -msg=self->'parse_parts'->(get: 1))->(body: (local: 'type'));
		
		else;
		
			local: 'output' = array;
			local: 'options' = (array:
					-type=local('type'),
					-preamble=((local_defined: 'preamble') && (#preamble !== false)),
					-array=((local_defined: 'array') && (#array !== false)),
				);
			local: 'single' = (!(local_defined: 'array') || (#array === false));
			local: 'types' = (array);
			local: 'parts' = (array);
			iterate: self, (local: 'temp');
				#parts->(insert: @#temp);
				#types->(insert: #temp->content_type);
			/iterate;
			if: (local_defined: 'type') && (#type != '');
				iterate: #types, (local: 'temp');
					if: (#temp >> #type);
						local: 'result' = #parts->(get: loop_count)->\body->run(-owner=#parts->(get: loop_count), -params=#options);
						#single ? return: @#result;
						#result->isa('array') ? #output->insertfrom(#result->iterator) | #output->insert(#result);
					/if;
				/iterate;
				return: @#output;
			else;
				// Use text/plain parts
				if: #output->size == 0;
					iterate: #types, (local: 'temp');
						if: (#temp >> 'text/plain');
							local: 'result' = #parts->(get: loop_count)->\body->run(-owner=#parts->(get: loop_count), -params=#options);
							#single ? return: @#result;
							#result->isa('array') ? #output->insertfrom(#result->iterator) | #output->insert(#result);
						/if;
					/iterate;
				/if;
				// Use text/html parts if there are no text/plain parts
				if: #output->size == 0;
					iterate: #types, (local: 'temp');
						if: (#temp >> 'text/html');
							local: 'result' = #parts->(get: loop_count)->\body->run(-owner=#parts->(get: loop_count), -params=#options);
							#single ? return: @#result;
							#result->isa('array') ? #output->insertfrom(#result->iterator) | #output->insert(#result);
						/if;
					/iterate;
				/if;
				// Use any text/ parts if there are no text/plain or text/html parts
				if: #output->size == 0;
					iterate: #types, (local: 'temp');
						if: (#temp >> 'text/');
							local: 'result' = #parts->(get: loop_count)->\body->run(-owner=#parts->(get: loop_count), -params=#options);
							#single ? return: @#result;
							#result->isa('array') ? #output->insertfrom(#result->iterator) | #output->insert(#result);
						/if;
					/iterate;
				/if;
				// Use first part if there are no text/ parts
				if: #output->size == 0;
					local: 'result' = (#parts->(get: 1)->\body->run(-owner=#parts->(get: 1), -params=#options));
					#single ? return: @#result;
					#result->isa('array') ? #output->insertfrom(#result->iterator) | #output->insert(#result);
				/if;
			/if;
		/if;
		
		return: @#output;
		
	/define_tag;
	
	define_tag: 'size';
		return: self->'parse_parts'->size;
	/define_tag;
	
	define_tag: 'get',
			-required='count';
		fail_if: (self->'parse_parts'->size < (integer: #count)), -9956, 'The count (' + (integer: #count) + ') was greater than the number of message parts (' + self->'parse_parts'->size + ').';
		return: (email_parse: -msg=self->'parse_parts'->(get: #count));
	/define_tag;

	define_tag: 'first';
		fail_if: (self->'parse_parts'->size < 1), -9956, 'The message does not contain any parts.';
		return: (email_parse: -msg=self->'parse_parts'->(get: 1));
	/define_tag;

	define_tag: 'second';
		fail_if: (self->'parse_parts'->size < 2), -9956, 'The message does not contain more than one part.';
		return: (email_parse: -msg=self->'parse_parts'->(get: 2));
	/define_tag;

	define_tag: 'last';
		fail_if: (self->'parse_parts'->size < 1), -9956, 'The message does not contain any parts.';
		return: (email_parse: -msg=self->'parse_parts'->(get: self->'parse_parts'->size));
	/define_tag;

/define_type;


// [Email_SafeEmail]
//
// First parameter is a string containing the contents of a To/From header.
//
// This tag returns a 'safe' version of the user's address using their
// friendly name or the first part of their email address.
//

define_tag: 'SafeEmail',
		-Required='data',
		-Optional='join',
		-Priority='Replace',
		-namespace = 'email_';
	local: 'comment' = (Email_Extract: #data, -comment);
	local: 'extract' = (Email_Extract: #data);
	local: 'array' = (array);
	if: !(#comment->(isa: 'array'));
		local: 'comment' = (array: #comment);
	/if;
	if: !(#extract->(isa: 'array'));
		local: 'extract' = (array: #extract);
	/if;
	if: #comment->size >= #extract->size;
		iterate: #comment, (local: 'temp');
			if: (#temp == '') && (loop_count <= #extract->size);
				#temp = #extract->(get: loop_count);
			/if;
		/iterate;
		local: 'array' = @#comment;
	else;
		iterate: #extract, (local: 'temp');
			if: (#temp == '') && (loop_count <= #comment->size);
				#temp = #comment->(get: loop_count);
			/if;
		/iterate;
		local: 'array' = @#extract;
	/if;
	iterate: #array, (local: 'output');
		if: (#output >> '@');
			local: 'index' = #output->(find: '@');
			#output = (_substring: #output, 1, #index - 1);
		/if;
		#output->trim;
		#output->(removeleading: '"');
		#output->(removetrailing: '"');
	/iterate;
	if: (#array->size == 0);
		return: null;
	else: (#array->size == 1);
		return: #array->(get: 1);
	else: (local_defined: 'join') && (#join !== null);
		return: #array->(join: #join);
	else;
		return: #array;
	/if;
/define_tag;


// [Email_Extract]
//
// First parameter is a string containing the contents of a header.
//

define_tag: 'Extract',
		-Required='data',
		-Optional='comment',
		-Optional='join',
		-Priority='Replace',
		-namespace = 'email_';
	if: (local_defined: 'comment');
	
		// Extract parentheses
		local: 'array' = (string_findregexp: #data, -find='\\([^\\)]+\\)');
		if: #array->size > 0;
			iterate: #array, (local: 'output');
				#output->(removeleading: '(');
				#output->(removetrailing: ')');
				#output = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
			/iterate;
			#array->(removeall: '');
			if: #array->size == 1;
				return: #array->(get: 1);
			else: (local_defined: 'join') && (#join !== null);
				return: #array->(join: #join);
			else;
				return: #array;
			/if;
		/if;
		
		// Use heuristic to strip out non-comments
		local: 'array' = (string_replaceregexp: #data, -find='<[^>]+>', -replace='<>')->(split: '<>');
		if: #array->size > 0;
			iterate: #array, (local: 'output');
				#output->trim;
				#output = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
			/iterate;
			#array->(removeall: '');
			#array->(removeall: ',');
			if: (#array->size == 0);
				return: null;
			else: (#array->size == 1);
				return: #array->(get: 1);
			else: (local_defined: 'join') && (#join !== null);
				return: #array->(join: #join);
			else;
				return: #array;
			/if;
		/if;
		
		// Fallback heuristic
		local: 'output' = (string_replaceregexp: #data, -find='<[^>]+>', -replace='');
		#output->trim;
		#output = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
		return: #output;
		
	else;

		// Extract angle brackets
		local: 'array' = (string_findregexp: #data, -find='<[^>]+>');
		if: #array->size > 0;
			iterate: #array, (local: 'output');
				#output->(removeleading: '<');
				#output->(removetrailing: '>');
				#output = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
			/iterate;
			#array->(removeall: '');
			if: #array->size == 1;
				return: #array->(get: 1);
			else: (local_defined: 'join') && (#join !== null);
				return: #array->(join: #join);
			else;
				return: #array;
			/if;
		/if;

		// Use heuristic to strip out comments
		local: 'array' = (string_replaceregexp: #data, -find='\\([^\\)]+\\)', -replace='()')->(split: '()');
		if: #array->size > 0;
			iterate: #array, (local: 'output');
				#output->trim;
				#output = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
			/iterate;
			#array->(removeall: '');
			#array->(removeall: ',');
			if: (#array->size == 0);
				return: null;
			else: (#array->size == 1);
				return: #array->(get: 1);
			else: (local_defined: 'join') && (#join !== null);
				return: #array->(join: #join);
			else;
				return: #array;
			/if;
		/if;
		
		// Fallback heuristic
		local: 'output' = (string_replaceregexp: #data, -find='\\([^\\)]+\\)', -replace='');
		#output->trim;
		#output = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
		return: #output;
		
	/if;
/define_tag;

define_tag: '_substring', -required='data', -optional='start', -optional='length';
	if: #data->(isa: 'bytes');
		if: ((local: 'start') !== null) && (#start > 0);
			if: ((local: 'length') !== null) && (#length > 0);
				return: #data->(getrange: #start, #length);
			else;
				return: #data->(getrange: #start, (1 + #data->size - #start));
			/if;
		/if;
	else: #data->(isa: 'string');
		if: ((local: 'start') !== null) && (#start > 0);
			if: ((local: 'length') !== null) && (#length > 0);
				return: #data->(substring: #start, #length);
			else;
				return: #data->(substring: #start);
			/if;
		/if;
	else;
		return: (_substring: (string: #data), -start=(local: 'start'), -length=(local: 'length'));
	/if;
	return: #data;
/define_tag;

	Define_Tag: 'digestchallenge', 
			-Namespace='email_',
			-required='authorization';
	
		Local: 'output' = (map);

		define_tag: 'extract', -required='data', -required='field';
			if: (#data >> #field);
				local: 'pattern' = #field + '="([^"]*)"';
				local: 'result' = (string_findregexp: #data, -find=#pattern);
				if: #result->size > 1;
					return: #result->second;
				/if;
				local: 'pattern' = #field + '=([^, ]*)';
				local: 'result' = (string_findregexp: #data, -find=#pattern);
				if: #result->size > 1;
					return: #result->second;
				/if;
			/if;
			return: '';
		/define_tag;

		local: 'digest_realm' = (extract: #authorization, 'realm');
		local: 'digest_nonce' = (extract: #authorization, 'nonce');
		local: 'digest_qop' = (extract: #authorization, 'qop');
		local: 'digest_stale' = (extract: #authorization, 'stale');
		local: 'digest_maxbuf' = (extract: #authorization, 'maxbuf');
		local: 'digest_charset' = (extract: #authorization, 'charset');
		local: 'digest_algorithm' = (extract: #authorization, 'algorithm');
		local: 'digest_cipher' = (extract: #authorization, 'cipher');

		#output->(insert: 'realm' = #digest_realm);
		#output->(insert: 'nonce' = #digest_nonce);
		#output->(insert: 'qop' = #digest_qop);
		#output->(insert: 'stale' = #digest_stale);
		#output->(insert: 'maxbuf' = #digest_maxbuf);
		#output->(insert: 'charset' = #digest_charset);
		#output->(insert: 'algorithm' = #digest_algorithm);
		#output->(insert: 'cipher' = #digest_cipher);
	
		Return: @#output;
	
	/Define_Tag;

	Define_Tag: 'digestresponse', 
			-Namespace='email_',
			-required='digest';
	
		Define_Tag: 'quote', -Required='value';
			((String_ReplaceRegexp: #value, -Find='\\w+', -replace='') == '') ? Return: #value;
			Local: 'output' = (String: #value);
			#output->(replace: '\\', '\\\\');
			#output->(replace: '"', '\\"');
			#output = '"' + #output + '"';
			Return: @#output;
		/Define_Tag;
	
		Local: 'output' = '';
		if: #digest->(find: 'charset') != '';
			#output += 'charset=' + (quote: #digest->(find: 'charset')) + ',';
		/if;
		#output += 'username=' + (quote: #digest->(find: 'username')) + ',';
		if: #digest->(find: 'realm') != '';
			#output += 'realm=' + (quote: #digest->(find: 'realm')) + ',';
		/if;
		#output += 'nonce=' + (quote: #digest->(find: 'nonce')) + ',';
		#output += 'nc=' + (quote: #digest->(find: 'nc')) + ',';
		#output += 'cnonce=' + (quote: #digest->(find: 'cnonce')) + ',';
		#output += 'digest-uri=' + (quote: #digest->(find: 'digest-uri')) + ',';
		if: #digest !>> 'response';
			local: 'a0' = #digest->(find: 'username') + ':' + #digest->(find: 'realm') + ':' + #digest->(find: 'password');
			local: 'a1' = (cipher_digest: #a0, -digest='md5');
			#a1 += ':' + #digest->(find: 'nonce') + ':' + #digest->(find: 'cnonce');
			local: 'a2' = 'AUTHENTICATE' + ':' + #digest->(find: 'digest-uri');
			local: 'response' = (encrypt_md5: #a1) + ':' + #digest->(find: 'nonce') + ':' + #digest->(find: 'nc') + ':' + #digest->(find: 'cnonce') + ':auth:' + (encrypt_md5: #a2);
			#output += 'response=' + (quote: (encrypt_md5: #response)) + ',';
		else;
			#output += 'response=' + (quote: #digest->(find: 'response')) + ',';
		/if;
		#output += 'qop=' + (quote: 'auth') + '';
		Return: @#output;
	
	/Define_Tag;


?>
