<!--	
	Library of routines to perform low-level scheduling, analogous to old Post_Inline;
	
	Implements:
		[Event_Schedule]

	Variables:
		__Schedule_Status__ is Active/Sleep/Pause/Abort
			Active - Check schedule queue within one minute, sleep until next event 
			Sleep - Sleep until indicated time
			Pause - Pause indefinitely
			Abort - Kill the session watcher
		__Schedule_Sleep__ is date/time at which to wake up

	History:
		10/30/2003	JS	- added special handling of delays that are multiples of 1440
		10/30/2003	JS	- added async execution of events
		10/30/2003	JS	- added custom tag _include_async
		10/30/2003	JS	- removed -NoReply from include_url since that 
							still waits for the page so it doesn`t help much
	    10/1/2003   FMS - LP7 syntax update, privileged tag, new status variable, silent sleep.
	    08/13/2002  FMS - LP6 syntax update, optimized SQL statements, added logging.
	    05/30/2002  FMS - Updated with new syntax, optimized SQL statements
		12/07/2001	GLH - Changed max expire date from 3000 to 2021 to prevent math overflow
		09/07/2001	GLH - Changed to _LLEvent_Schedule; to get around security issues
		06/18/2001	GLH - Added schedule pause logic
		04/11/2001	GLH - Initial Development
-->

<!-- 
	Event_Schedule:
		Queues a schedule item for later execution.  Roughly equivalent to old Post_Inline;

	Inputs:
		-Start=(Date),			//	defaults to today
		-End=(Date),			//	defaults to never
		-URL=(executeAction),		//	URL to execute (simulate a browser hit) (Required)
	    -Repeat=(true/false),	//	Should this repeat? Default to true if -Delay passed
	    -Restart=(true/false),	//	Continue executing after restart.  Defaults to true
	    -Delay=(#Minutes),
	    -Username,				//	optional username for authentication
	    -Password;				//	optional password for authentication

	Outputs:
		None

	Side-Effects:
		Appends a record to schedule that contains all info required to execute the item
-->

<?LassoScript

// moved block inside async

Define_Tag: 'Schedule', -namespace='Event_', -Priority='replace', -Privileged,
		-Required='url',
		-Optional='start',
		-Optional='end',
		-Optional='delay',
		-Optional='repeat',
		-Optional='restart',
		-Optional='username',
		-Optional='password';

	// Create an array to store the constructed event
	local: 'event_array' = (array);

	// Insert the event
	local: 'l_url' = (string: (local: 'url'));
	#l_url->trim;
	#event_array->(insert: 'executeAction' = (local: 'l_url'));

	log_detail: 'Schedule: Queue URL <' + (local: 'l_url') + '>';

	// If no start date defined, then default to today
	if: ((local: 'start') == '');
		#event_array->(insert: 'date_time_start' = (Date)->(Format: '%Q %T'));
		#event_array->(insert: 'next_execute' = (Date)->(Format: '%Q %T'));
	else;
		#event_array->(insert: 'date_time_start' = (Date: (string: (local: 'start')))->(Format: '%Q %T'));
		#event_array->(insert: 'next_execute' = (Date: (string: (local: 'start')))->(Format: '%Q %T'));
	/If;

	// If no end date defined, then default to never
	if: ((local: 'end') == '');
		#event_array->(insert: 'date_time_end' = '3001-01-01 00:00:00');
	else;
		#event_array->(insert: 'date_time_end' = (Date: (string: (local: 'end')))->(Format: '%Q %T'));
	/if;

	// If no delay defined, then default to 0
	if: ((local: 'delay') == '');
		local: 'l_delay' = 0;
	else: ((integer: (local: 'delay')) < 0);
		local: 'l_delay' = 0;
	else;
		local: 'l_delay' = (integer: (local: 'delay'));
	/if;
	#event_array->(insert: 'delay'=#l_delay);

	// If no repeat defined, then default to true iff a delay is passed
	if: !(local_defined: 'repeat');
		if: (#l_delay > 0);
			#event_array->(insert: 'repeat'='1');
		else;
			#event_array->(insert: 'repeat'='0');
		/if;
	else: (#l_delay == 0);
		#event_array->(insert: 'repeat'='0');
	else;
		#event_array->(insert: 'repeat'=(integer: (boolean: (local: 'repeat'))));
	/if;

	// If no restart defined, then default to true
	if: !(local_defined: 'restart');
		#event_array->(insert: 'restart'='1');
	else;
		#event_array->(insert: 'restart'=(integer: (boolean: (local: 'restart'))));
	/if;

	// If no username defined, then default to ''
	#event_array->(insert: 'user'=(string: (local: 'username')));

	// If no password defined, then default to ''
	#event_array->(insert: 'pass'=(string: (local: 'password')));

	// Add the event to the database
	inline: -add, 
			-log=$__startup_loglevel__, 
			-database='site', 
			-table='schedule', 
			#event_array;
		if: (error_currenterror) != (error_noerror);
			log_detail: 'Schedule: ' + (error_currenterror) + ' (' + (error_currenterror: -errorcode) + ')';
		/if;
	/inline;
	
	// After adding an event, check the event queue immediately
	if: ((Global: '__Schedule_Status__') >> 'Sleep');
		Global: '__Schedule_Status__' = 'Active';
	/if;
	
/Define_Tag;

?>

<!--
	Schedule Code

	Actions:
		Creates new schedule table if it doesn`t exist already
		Changes the type of the executeAction column if it is not TEXT
		Creates global __Schedule_Status__ variable
		Creates global __Schedule_Sleep__ variable
		Deletes events that are not set to restart
		Starts watcher and loops waiting for events to be scheduled

	Note:
		This code is only designed to be called once from LassoStartup
		It is formatted as a compound expression so Local variables
		can be used.
		
-->

<?LassoScript
	{
		Log_Warning: 'Schedule: Initializing Event Queue';
	
		var: '__Schedule_Status__' = 'Active';
		var: '__Schedule_Sleep__' = (Date);
		
		inline: -log=$__startup_loglevel__, -database='lasso_internal', -table='global_prefs', -search, -op='eq', 'store_key'='lasso_schedule_pause';
			if: (field: 'data') == 'Y';
				var: '__Schedule_Status__'='Pause';
			else;
				var: '__Schedule_Status__'='Active';
			/if;
		/inline;
	
		// Check if the table exists
		Local: 'schedule_table_exists' = false;
		Local: 'error_table_exists' = false;
		inline: -log=$__startup_loglevel__, -Database='site', -SQL='SELECT name from sqlite_master WHERE type="table" AND name LIKE "schedule%"';
			records;
				if: (field: 'name') == 'schedule';
					#schedule_table_exists = true;
				else: (field: 'name') == 'schedule_error';
					#error_table_exists = true;
				/if;
			/records;
		/inline;
	
		if: #schedule_table_exists == false;
			var: '__need_reload__' = true;
			// Create table if it doesn't exist already
			inline: -log=$__startup_loglevel__, -Database='site',
				-SQL='CREATE TABLE schedule \
				( \
					id INTEGER, \
					date_time_start TEXT, \
					date_time_end TEXT, \
					next_execute TEXT, \
					executeAction TEXT, \
					repeat INTEGER, \
					restart INTEGER, \
					delay INTEGER, \
					user TEXT, \
					pass TEXT, \
					PRIMARY KEY (id) \
				)';
			/Inline;
		/if;

		if: #error_table_exists == false;
			var: '__need_reload__' = true;
			// Create table if it doesn't exist already
			inline: -log=$__startup_loglevel__, -Database='site',
				-SQL='CREATE TABLE schedule_error \
				( \
					id INTEGER, \
					id_schedule INTEGER, \
					date_time TEXT, \
					code TEXT, \
					error TEXT, \
					PRIMARY KEY (id) \
				)';
			/Inline;
		/if;
	
		// Delete all records that are marked as non-restart or are past end date
		inline: -log=$__startup_loglevel__, -Database='site',
			-SQL='DELETE FROM schedule WHERE (restart = 0) OR ( (date_time_end NOT LIKE "3001%") AND (strftime("%s", date_time_end) <= strftime("%s", datetime("now","localtime"))) ) OR (next_execute = "")';
		/Inline;
	
		// Push the next_execute date to the next even multiple
		// This prevents events from ever executing twice rapidly
		inline: -log=$__startup_loglevel__, -Database='site', -SQL='SELECT * FROM schedule WHERE ( (date_time_end LIKE "3001%") OR (strftime("%s", date_time_end) >= strftime("%s", datetime("now","localtime"))) ) AND (strftime("%s", next_execute) <= strftime("%s", datetime("now","localtime"))) ORDER BY next_execute';
			Records;
				local: 'next_execute' = (Date: (Field: 'next_execute'), -format='%Q %T');
				local: 'date_fix' = (Date)->(Difference: (Date)->(Format: '%Q %T'), -MinutesBetween);
				local: 'next_difference' = #next_execute->(Difference: (Date), -MinutesBetween) + #date_fix;
				
				local: 'delay' = (integer: (field: 'delay'));
	
				If: (#delay > 0);
					local: 'next_difference' = #next_difference + #delay - (#next_difference % #delay);
					#next_execute->(add: -minute=#next_difference);
					
					inline: -log=$__startup_loglevel__,  -Database='site', -SQL='UPDATE schedule SET next_execute=\"' + (Date_Format: #next_execute, -format='%Q %T') + '\" WHERE id = ' + (Field: 'id');
					/Inline;
				/If;
			/Records;
		/Inline;
		
	}->(Run);

	// The rest of the script is run asynchronously
	{
		// 
		// Start Schedule Watcher
		//

		// added by Johan Slve 10/30/2003
		// error capture by FMS 02/24/2006
		define_tag: '_include_async', -async,
			-required='url',
			-optional='username',
			-optional='password',
			-optional='id',
			-priority='replace';
			// runs include_url in a separate thread
			protect;
				
				if: (local_defined: 'username');
					include_url: #url,
						-username=(local: 'username'),
						-password=(local: 'password');
				else;
					include_url: #url;
				/if;
				
				handle: true;
					// Reading these directly does not work
					// Storing them in a local does
					local: 'code' = $__result_code__;
					local: 'msg' = $__result_message__;

					if: (local_defined: 'id') && (#code != 0);
						log_warning: 'Schedule: Error on URL <' + (local: 'url') + '> ' + #code + ' ' + #msg;
						// Record the error message
						inline: -log=$__startup_loglevel__, 
								-add='',
								-database='site', 
								-table='schedule_error',
								'id_schedule' = #id,
								'code' = #code,
								'error' = #msg,
								'date_time' = date->(format: '%Q %T');
						/inline;
					/if;
				/handle;
			/protect;
		/define_tag;

		Log_Warning: 'Schedule: Starting Event Queue';
		
		// 30 Second Delay at Startup
		Sleep: 30000; 
		
		While: ((Global: '__Schedule_Status__') != 'Abort');
			Protect;
				If: ((Global: '__Schedule_Status__') == 'Pause') || ((Global: '__Schedule_Status__') == 'Stop');
					local: 'execute' = false;
				else: ((Global: '__Schedule_Status__') == 'Sleep');
					if: (Global: '__Schedule_Sleep__')->(isa: 'date') && ((Global: '__Schedule_Sleep__') <= Date);
						local: 'execute' = true;
					else;
						local: 'execute' = false;
					/if;
				else;
					local: 'execute' = true;
				/if;
				
				If: (#execute == true);
				
					//
					// Check Event Queue
					//
				
					Inline: -log=$__startup_loglevel__, -Database='site', -SQL='SELECT id FROM schedule LIMIT 1';
						If: Found_Count > 0;
							Log_Detail: 'Schedule: Checking Event Queue';
						/If;
					/Inline;
						
					// Delete all records that are past end date.
					Inline: -log=$__startup_loglevel__, -Database='site', -SQL='DELETE FROM schedule WHERE ( (date_time_end NOT LIKE "3001%") AND (strftime("%s", date_time_end) <= strftime("%s", datetime("now","localtime"))) )';
					/Inline;
				
					// Set the sleep time to one hour
					Global: '__Schedule_Sleep__' = (Date);
					(Global: '__Schedule_Sleep__')->(Add: -Hour=1);
				
				
					// Execute all records that are not past end date, not before start date, and which have next_execute earlier than now.
					// Note: adjusting start date into the year 3001 effectively "pauses" an event.
					inline: -log=$__startup_loglevel__, -Database='site', -SQL='SELECT * FROM schedule WHERE ( (date_time_end LIKE "3001%") OR (strftime("%s", date_time_end) >= strftime("%s", datetime("now","localtime"))) ) AND ( (date_time_start NOT LIKE "3001%") AND (strftime("%s", date_time_start) <= strftime("%s", datetime("now","localtime")))) AND (strftime("%s", next_execute) <= strftime("%s", datetime("now","localtime"))) ORDER BY next_execute';
						Records;
							If: ((Global: '__Schedule_Status__') == 'Pause') || ((Global: '__Schedule_Status__') == 'Stop');
								Loop_Abort;
							/If;
							
							local: 'executeAction' = (field: 'executeAction');
							#executeAction->(trim);
				
							log_detail: 'Schedule: Executing URL <' + (local: 'executeAction') + '>';
				
							Protect;
								//  Now execute the URL (do not wait for results)
								If: (Field: 'user') == '';
									_include_async: -url=(local: 'executeAction'), -id=(field: 'id');
								Else;
									_include_async: -url=(local: 'executeAction'), -Username=(Field: 'user'), -Password=(Field: 'pass'), -id=(field: 'id');
								/If;
							/Protect;
							
							local: 'delay' = (integer: (field: 'delay'));
							local: 'repeat' = (integer: (field: 'delay'));
				
							// Delete the record if it's a one-shot
							If: (#repeat == 0) || (#delay <= 0);
								log_detail: 'Schedule: Deleting One-Time Event';
								inline: -log=$__startup_loglevel__,  -Database='site', -SQL='DELETE FROM schedule WHERE id = ' + (Field: 'id'); 
								/Inline;
							Else;
								local: 'next_execute' = (Date: (Field: 'next_execute'), -format='%Q %T');
								if: #delay % 1440 == 0;
									// handle daily, weekly etc repeats specially to make them repeat on the same time of day,
									// regardless of clock changes like daylight savings switch
									// Added by Johan Slve 10/30/2003
									local: 'next_difference'=0;
									// split #next_execute into time and date (use neutral time for the date)
									local: 'next_execute_time'= #next_execute->(format: '%T');
									local: 'next_execute_date'= (date: (#next_execute -> format: '%Q 12:00:00'));
									// add the delay to get the right date (we will disregard the time from this operation)
									#next_execute_date->(add: -minute=#delay);
									// restore the previous time value
									#next_execute = (date: (#next_execute_date -> format: ('%Q ' + #next_execute_time)));
								else;
									// normal rescheduling
									local: 'date_fix' = (Date)->(Difference: (Date)->(Format: '%Q %T'), -MinutesBetween);
									local: 'next_difference' = #next_execute->(Difference: (Date), -MinutesBetween) + #date_fix;
									local: 'next_difference' = #next_difference + #delay - (#next_difference % #delay);
									#next_execute->(add: -minute=#next_difference);
								/if;
								log_detail: 'Schedule: Rescheduling to ' + (local: 'next_execute')->(Format: '%Q %T');
								inline: -log=$__startup_loglevel__,  -Database='site', -SQL='UPDATE schedule SET next_execute=\"' + (Date_Format: #next_execute, -format='%Q %T') + '\" WHERE id = ' + (Field: 'id');
								/Inline;
								If: (global: '__Schedule_Sleep__')->(isa: 'date') && (#next_execute < (global: '__Schedule_Sleep__'));
									global: '__Schedule_Sleep__' = #next_execute;
								/if;
							/If;
							
							// 5 Second Delay Between Events
							If: (Loop_Count < Found_Count);
								Sleep: 5000; 
							/If;
							
						/Records;
					/Inline;

					if: ((Global: '__Schedule_Status__') != 'Pause') && ((Global: '__Schedule_Status__') != 'Stop');
						inline: -log=$__startup_loglevel__, -Database='site', -SQL='SELECT next_execute FROM schedule ORDER BY next_execute LIMIT 1';
							Local: 'found_count' = Found_Count;
							If: (#found_count > 0);
								local: 'next_execute' = (Date: (Field: 'next_execute'), -format='%Q %T');
								If: (global: '__Schedule_Sleep__')->(isa: 'date') && (#next_execute < (global: '__Schedule_Sleep__'));
									global: '__Schedule_Sleep__' = #next_execute;
								/if;
							/if;
						/inline;
						
						If: (global: '__Schedule_Sleep__')->(isa: 'date') && ((date_add: (date), -minute=2) < (global: '__Schedule_Sleep__'));
							global: '__Schedule_Status__' = 'Sleep';
							If: (#found_count > 0);
								Log_Detail: 'Schedule: Sleeping Until ' + (global: '__Schedule_Sleep__')->(Format: '%Q %T');
							/if;
						else;
							global: '__Schedule_Status__' = 'Active';
						/if;
					/if;

				/If;
				
			/Protect;
			
			Protect;
			
				//
				// Session Delete
				//
				
				if: !(valid_date: (global: '_session_lastexpiration'));
					if: (global: '_session_expirefrequency') >> 'hourly';
						global: '_session_lastexpiration' = date->(format: '%Q %T');
					else: (global: '_session_expirefrequency') >> 'daily';
						global: '_session_lastexpiration' = date->(format: '%Q 00:00:01');
					else: (global: '_session_expirefrequency') >> 'weekly';
						global: '_session_lastexpiration' = date->(subtract: -day=6) & (format: '%Q 00:00:01');
					/if;
				/if;
				if: (global: '_session_expirefrequency') >> 'hourly';
					if: (date_difference: date, (global: '_session_lastexpiration'), -minutesbetween) >= 59;
						session_deleteexpired;
						global: '_session_lastexpiration' = date->(format: '%Q %T');
					/if;
				else: (global: '_session_expirefrequency') >> 'daily';
					if: (date_difference: date, (global: '_session_lastexpiration'), -hoursbetween) >= 24;
						session_deleteexpired;
						global: '_session_lastexpiration' = date->(format: '%Q 00:00:01');
					/if;
				else: (global: '_session_expirefrequency') >> 'weekly';
					if: (date_difference: date, (global: '_session_lastexpiration'), -hoursbetween) >= 168;
						session_deleteexpired;
						global: '_session_lastexpiration' = date->(format: '%Q 00:00:01');
					/if;
				/if;
				
			/Protect;

			Protect;
			
				//
				// Cache Garbage Collection
				//
				
				if: global_defined('_cache_preference');
					if: !(valid_date: (global: '_cache_lastexpiration'));
						if: (global: '_cache_preference')->(find: '_cache_expirefrequency') >> 'hourly';
							global: '_cache_lastexpiration' = date->(format: '%Q %T');
						else: (global: '_cache_preference')->(find: '_cache_expirefrequency') >> 'daily';
							global: '_cache_lastexpiration' = date->(format: '%Q 00:00:01');
						else: (global: '_cache_preference')->(find: '_cache_expirefrequency') >> 'weekly';
							global: '_cache_lastexpiration' = date->(subtract: -day=6) & (format: '%Q 00:00:01');
						/if;
					/if;
					if: (global: '_cache_preference')->(find: '_cache_expirefrequency') >> 'hourly';
						if: (date_difference: date, (global: '_cache_lastexpiration'), -minutesbetween) >= 59;
							cache_maintenance;
							global: '_cache_lastexpiration' = date->(format: '%Q %T');
						/if;
					else: (global: '_cache_preference')->(find: '_cache_expirefrequency') >> 'daily';
						if: (date_difference: date, (global: '_cache_lastexpiration'), -hoursbetween) >= 24;
							cache_maintenance;
							global: '_cache_lastexpiration' = date->(format: '%Q 00:00:01');
						/if;
					else: (global: '_cache_preference')->(find: '_cache_expirefrequency') >> 'weekly';
						if: (date_difference: date, (global: '_cache_lastexpiration'), -hoursbetween) >= 168;
							cache_maintenance;
							global: '_cache_lastexpiration' = date->(format: '%Q 00:00:01');
						/if;
					/if;
				/if;
				
			/Protect;
			
			// 60 Second Delay Between Active Checks
			Sleep: 60000;
			
		/While;
		
		Log_Warning: 'Schedule: Stopping Event Queue';	

	}->(AsAsync: -Name='Schedule: Event Queue'); 
?>
