<?LassoScript

	// DNS.Lasso
	//
	// Implements:
	// [DNS_Lookup]
	// [DNS_Response]
	//
	// Copyright  2007 LassoSoft, LLC
	//
	//	[DNS_Lookup: -hostname='dns server', -name='host to lookup', -type='*|A|CNAME|MX|PTR|TXT']
	//	If the name is in the form xxx.xxx.xxx.xxx then a reverse lookup is performed.
	//	If no type is specified then * is used or PTR for a reverse lookup.
	// 	Additional parameters include -class='IN', -status to perform a status lookup, 
	//	-norecurse to turn off recursive lookups, -inverse to perform a reverse lookup.
	//
	// 	The return value depends on the type.
	// 		MX - Returns an array of pairs. Each pair has preference in the first element (lower is better) 
	//			and an address in the second element.
	//		Others - Returns array of returned values.
	//		Note an empty array is returned if no answer could be determined.
	//
	// 	Additionally, if -Format is specified then a formatted string output is returned.  -BitFormat
	//	returns a formatted display of the raw bits of the response.  -ShowResponse returns the [DNS_Response]
	// 	object directly.  Similar parameters can be used to display the query rather than execute it including
	//	-FormatQuery, -BitQuery, and -ShowQuery to return a [DNS_Response] representing the query.
	
	//	If the return value is a [DNS_Response] object. This object has the following member tags.
	//	[DNS_Response->Format] returns a formatted display of the entire response from the DNS server.
	//	[DNS_Response->BitFormat] returns a formatted display of the raw bits returned by the DNS server.
	//	[DNS_Response->Answer] returns the answer from the DNS server. This differs based on the type.
	// 		MX - Returns an array of pairs. Each pair has preference in the first element (lower is better) 
	//			and an address in the second element.
	//		Others - Returns array of returned values.
	//		Note an empty array is returned if no answer could be determined.
	//
	//  A comparison of two [DNS_Response] objects is true if the ID and Query of the responses are the same.  
	//  The comparison does not distinguish between lookups and responses so it can be used to check that a 
	//  response corresponds to the lookup which was sent.
	//
	// 	Example Usage:
	//
	//		[DNS_Lookup: -hostname='dns.example.com', -name='www.example.com', -Format]
	//		Look up all available information. Outputs formatted display.
	//
	//		[DNS_Lookup: -hostname='dns.example.com', -name='127.0.0.1']
	//		Performs reverse lookup with PTR. Returns host name or array of host names.
	//
	//		[DNS_Lookup: -hostname='dns.example.com', -name='www.example.com', -type='mx']
	//		Performs MX lookup. Returns array of pairs with best server first.
	// 

	define_type: 'response', -namespace='dns_', -priority='replace', -prototype;
	
	private;
	
		local: 'parsed' = false;
		local: 'raw' = bytes;
		local: 'answer' = null;
		
		local: 'length' = integer;
		local: 'id' = integer;		
		local: 'operation' = integer;		
		local: 'qdcount' = integer;
		local: 'ancount' = integer;
		local: 'nscount' = integer;
		local: 'arcount' = integer;
		
		local: 'qdarray' = (array);
		local: 'anarray' = (array);
		local: 'nsarray' = (array);
		local: 'ararray' = (array);
		
		local: 'domain_cache' = map;
	
		define_tag: 'consume_rdata', 
				-required='message',
				-required='type';
			local: 'output' = '';
			local: 'length' = #message->export16bits->hosttonet16;
			if: (#type == 1); // A
				iterate: bytes->(import32bits: #message->export32bits) &, (local: 'byte');
					#output += #byte + '.';
				/iterate;
				#output->(removetrailing: '.');
			else: (#type == 2) || (#type == 3) || (#type == 4) || (#type == 5) || (#type == 7) || (#type == 8) || (#type == 9) || (#type == 12);
				return: self->(consume_domain: @#message);
			else: (#type == 6); // SOA
				local: 'array' = (array);
				#array->(insert: 'mname' = self->(consume_domain: @#message));
				#array->(insert: 'rname' = (string_replaceregexp: self->(consume_domain: @#message), -find='\\.', -replace='@', -replaceonlyone));
				#array->(insert: 'serial' = #message->export32bits->hosttonet32);
				#array->(insert: 'refresh' = #message->export32bits->hosttonet32);
				#array->(insert: 'retry' = #message->export32bits->hosttonet32);
				#array->(insert: 'expire' = #message->export32bits->hosttonet32);
				#array->(insert: 'ttl' = #message->export32bits->hosttonet32);
				return: @#array;
			else: (#type == 11); // WKS
				local: 'array' = (array);
				local: 'output' = '';
				iterate: bytes->(import32bits: #message->export32bits) &, (local: 'byte');
					#output += #byte + '.';
				/iterate;
				#output->(removetrailing: '.');
				#array->(insert: #output);
				#array->(insert: #message->export8bits);
				local: 'bytes' = '';
				loop: #length - 5;
					#bytes += #message->export8bits;
				/loop;
				#array->(insert: #bytes);
				return: @#array;
			else: (#type == 13); // HINFO
				local: 'array' = (array);
				#array->(insert: self->(consume_string: @#message));
				#array->(insert: self->(consume_string: @#message));
				return: @#array;
			else: (#type == 14); // MINFO
				local: 'array' = (array);
				#array->(insert: self->(consume_domain: @#message));
				#array->(insert: self->(consume_domain: @#message));
				return: @#array;
			else: (#type == 15); // MX
				local: 'array' = (array);
				#array->(insert: #message->export16bits->hosttonet16);
				#array->(insert: self->(consume_domain: @#message));
				return: @#array;
			else: (#type == 16); // TXT
				local: 'array' = (array);
				while: #length > 0;
					#array->(insert: self->(consume_string: @#message));
					#length -= (#array->last->size + 1);
				/while;
				return: @#array;
			else: (#type == 28); // AAAA
				local: 'temp' = bytes->(import32bits: #message->export32bits) & (import32bits: #message->export32bits) & (import32bits: #message->export32bits) & (import32bits: #message->export32bits) &;
				loop: #temp->size, -by=2;
					local: 'int' = (#temp->(get: loop_count) * 256 + #temp->(get: loop_count + 1));
					#int->(setformat: -hexadecimal=true);
					#int == 0 ? #int = '0';
					#output += (string: #int) - '0x' + ':';
				/loop;
				#output->(removetrailing: ':');
				// Clean up runs of zeroes
				local('max') = 0;
				local('dex') = 0;
				local('runs') = string_findregexp(#output, -find='(?:^|:)[0:]+(?:$|:)');
				iterate(#runs, local('run'));
				  if(#run->size > #max);
					#max = #run->size;
					#dex = loop_count;
				  /if;
				/iterate;
				if(#dex > 0);
					local: 'run' = #runs->get(#dex);
					if(#run->beginswith('0'));
						#output->removeleading(#run);
						#output = '::' + #output;
					else(#run->endswith('0'));
						#output->removetrailing(#run);
						#output += '::';
					else;
						#output = string_replace(#output, -find=#run, -replace='::', -replaceonlyone);
					/if;
				/if;
			else; // NULL
				loop: #length;
					#output += bytes->(import8bits: #message->export8bits) &;
				/loop;
			/if;
			return: @#output;
		/define_tag;

		define_tag: 'consume_string', 
				-required='message';
			local: 'output' = '';
			local: 'length' = #message->export8bits;
			loop: #length; 
				#output += bytes->(import8bits: #message->export8bits) &;
			/loop;
			return: @#output;
		/define_tag;

		define_tag: 'consume_label', 
				-required='message';
			local: 'output' = '';
			local: 'length' = #message->export8bits;
			if: (#length == 0);
				return: '.';
			else: (#length >= 64) || (#length < 0);
				local: 'domain' = #message->export8bits;
				if: self->'domain_cache' >> #domain;
					return: self->'domain_cache'->(find: #domain);
				else;
					return: '.';
				/if;
			else;
				loop: #length; 
					#output += bytes->(import8bits: #message->export8bits) &;
				/loop;
			/if;
			return: @#output;
		/define_tag;

		define_tag: 'consume_domain',
				-required='message';
			local: 'cache' = (map);
			local: 'location' = #message->position - 2;
			local: 'output' = '';

			local: 'temp' = self->(consume_label: @#message);
			#output += #temp;
			#cache->(insert: #location = #temp);
			#location += #temp->size + 1;

			while: !#output->(endswith: '.');
				local: 'temp' = self->(consume_label: @#message);
				#output += '.' + #temp;
				iterate: #cache->keys, (local: 'key');
					#cache->(find: #key) += '.' + #temp;
				/iterate;
				#cache->(insert: #location = #temp);
				#location += #temp->size + 1;
			/while;
			iterate: #cache->keys, (local: 'key');
				self->'domain_cache'->(insert: #key = #cache->(find: #key));
			/iterate;
			return: @#output;
		/define_tag;

		define_tag: 'consume_message';
		
			self->'parsed' = true;
			
			local: '_error' = '';
			local: '_code' = 0;
			
			local: 'message' = self->'raw';

			
			error_push;
			protect;
				self->'length' = #message->export16bits->hosttonet16;
				self->'id' = #message->export16bits->hosttonet16;
				self->'operation' = #message->export16bits->hosttonet16;
				self->'qdcount' = #message->export16bits->hosttonet16;
				self->'ancount' = #message->export16bits->hosttonet16;
				self->'nscount' = #message->export16bits->hosttonet16;
				self->'arcount' = #message->export16bits->hosttonet16;
							
				loop: self->'qdcount';
					local: 'temp' = (map);
					#temp->(insert: 'qname' = self->(consume_domain: @#message));
					#temp->(insert: 'qtype' = #message->export16bits->hosttonet16);
					#temp->(insert: 'qclass' = #message->export16bits->hosttonet16);
					self->'qdarray'->(insert: #temp);
				/loop;
				
				loop: self->'ancount';
					local: 'temp' = (map);
					#temp->(insert: 'qname' = self->(consume_domain: @#message));
					#temp->(insert: 'qtype' = #message->export16bits->hosttonet16);
					#temp->(insert: 'qclass' = #message->export16bits->hosttonet16);
					#temp->(insert: 'ttl' = #message->export32bits->hosttonet32);
					#temp->(insert: 'rdata' = self->(consume_rdata: @#message, #temp->(find: 'qtype')));
					self->'anarray'->(insert: #temp);
				/loop;
				
				loop: self->'nscount';
					local: 'temp' = (map);
					#temp->(insert: 'qname' = self->(consume_domain: @#message));
					#temp->(insert: 'qtype' = #message->export16bits->hosttonet16);
					#temp->(insert: 'qclass' = #message->export16bits->hosttonet16);
					#temp->(insert: 'ttl' = #message->export32bits->hosttonet32);
					#temp->(insert: 'rdata' = self->(consume_rdata: @#message, #temp->(find: 'qtype')));
					self->'nsarray'->(insert: #temp);
				/loop;
	
				loop: self->'arcount';
					local: 'temp' = (map);
					#temp->(insert: 'qname' = self->(consume_domain: @#message));
					#temp->(insert: 'qtype' = #message->export16bits->hosttonet16);
					#temp->(insert: 'qclass' = #message->export16bits->hosttonet16);
					#temp->(insert: 'ttl' = #message->export32bits->hosttonet32);
					#temp->(insert: 'rdata' = self->(consume_rdata: @#message, #temp->(find: 'qtype')));
					self->'ararray'->(insert: #temp);
				/loop;
				
				handle_error;
					log_warning: '[DNS_Response] ' + error_code + ' ' + error_msg;
					local: '_error' = error_msg;
					local: '_code' = error_code;
				/handle_error;
			
			/protect;
			if: #_code != 0;
				fail: -9965, 'The DNS response could not be parsed.';
			else;
				error_pop;
			/if;
			
			if: self->'qdarray'->size > 0;
				local: 'question' = self->'qdarray'->first;
				if: #question->(find: 'qtype') == 255;
					// Nothing
				else: #question->(find: 'qtype') == 15;
					local: 'output' = (array);
					iterate: self->'anarray', (local: 'temp');
						if: (#temp->(find: 'qtype') == #question->(find: 'qtype')) && (#temp->(find: 'rdata')->(isa: 'array'));
							#output->(insert: #temp->(find: 'rdata')->first = #temp->(find: 'rdata')->second->(removetrailing: '.') &);
						/if;
					/iterate;
					#output->sort;
					self->'answer' = #output;
				else;
					local: 'output' = (array);
					iterate: self->'anarray', (local: 'temp');
						if: (#temp->(find: 'qtype') == #question->(find: 'qtype'));
							if: #temp->(find: 'rdata')->(isa: 'array');
								local: 'out' = array;
								iterate: #temp->(find: 'rdata'), (local: 'elem');
									if: #elem->(isa: 'pair');
										#out->(insert: #elem->first + ': ' + (string: #elem->second)->(removetrailing: '.') &);
									else;
										#out->(insert: (string: #elem));
									/if;
								/iterate;
								#output->(insert: #out->(join: ', '));
							else;
								#output->(insert: #temp->(find: 'rdata')->(removetrailing: '.') &);
							/if;
						/if;
					/iterate;
					self->'answer' = #output;
				/if;
			/if;

		/define_tag;
	
	/private;
	
		define_tag: 'oncreate', 
				-required='message';
			
			self->'raw' = #message;
			self->'parsed' = false;
		
		/define_tag;
			
		define_tag: 'onconvert';
			(self->'parsed' == false) ? self->consume_message;
			if: self->'answer' === null;
				return: self->format;
			else;
				return: self->'answer';
			/if;
		/define_tag;
	
		define_tag: 'format';
		
			(self->'parsed' == false) ? self->consume_message;

			local: 'dns_type_names' = (map:
				1 = 'A',
				2 = 'NS',
				3 = 'MD',
				4 = 'MF',
				5 = 'CNAME',
				6 = 'SOA',
				7 = 'MB',
				8 = 'MG',
				9 = 'MR',
				10 = 'NULL',
				11 = 'WKS',
				12 = 'PTR',
				13 = 'HINFO',
				14 = 'MINFO',
				15 = 'MX',
				16 = 'TXT',
				28 = 'AAAA',
				38 = 'A6',
				252 = 'AXFR',
				253 = 'MAILB',
				254 = 'MAILA',
				255 = '*');
		
			local: 'dns_class_names' = (map:
				1 = 'IN',
				2 = 'CS',
				3 = 'CH',
				255 = '*');

			local: 'output' = '<pre>';
			
			#output += 'Length: ' + self->'length' + '\r';
			#output += 'ID: ' + self->'id' + '\r';
			
			if: self->'operation'->(bittest: 16);
				#output += 'Type: Answer' + '\r';
			else;
				#output += 'Type: Query';
				if: self->'operation'->(bittest: 12);
					#output += ' Inverse';
				else: self->'operation'->(bittest: 13);
					#output += ' Status';
				/if;
				#output += '\r';
			/if;
			
			#output += 'Flags: ';
			local: 'del' = '';
			if: self->'operation'->(bittest: 11);
				#output += #del + 'Authoriative Answer (AA)';
				local: 'del' = ', ';
			/if;
			if: self->'operation'->(bittest: 10);
				#output += #del + 'Truncation (TC)';
				local: 'del' = ', ';
			/if;
			if: self->'operation'->(bittest: 9);
				#output += #del + 'Recursion Desired (RD)';
				local: 'del' = ', ';
			/if;
			if: self->'operation'->(bittest: 8);
				#output += #del + 'Recursion Available (RA)';
				local: 'del' = ', ';
			/if;
			#output += '\r';
			
			#output += 'Counts: ';
			local: 'del' = '';
			if: self->'qdcount' > 0;
				#output += #del + 'QD ' + self->'qdcount';
				local: 'del' = ', ';
			/if;
			if: self->'ancount' > 0;
				#output += #del + 'AN ' + self->'ancount';
				local: 'del' = ', ';
			/if;
			if: self->'nscount' > 0;
				#output += #del + 'NS ' + self->'nscount';
				local: 'del' = ', ';
			/if;
			if: self->'arcount' > 0;
				#output += #del + 'AR ' + self->'arcount';
				local: 'del' = ', ';
			/if;
			#output += '\r';

			iterate: self->'qdarray', (local: 'temp');
				#output += 'QD ' + loop_count + ': ';
				#output += #temp->(find: 'qname') + ' ';
				#output += (if_empty: #dns_type_names->(find: #temp->(find: 'qtype')), '(' + #temp->(find: 'qtype') + ')') + ' ';
				#output += (if_empty: #dns_class_names->(find: #temp->(find: 'qclass')), '(' + #temp->(find: 'qclass') + ')') + '\r';
			/iterate;
		
			iterate: self->'anarray', (local: 'temp');
				#output += 'AN ' + loop_count + ': ';
				#output += #temp->(find: 'qname') + ' ';
				#output += (if_empty: #dns_type_names->(find: #temp->(find: 'qtype')),  '(' + #temp->(find: 'qtype') + ')') + ' ';
				#output += (if_empty: #dns_class_names->(find: #temp->(find: 'qclass')), '(' + #temp->(find: 'qclass') + ')') + ' ';
				#output += #temp->(find: 'ttl') + ' ';
				if: #temp->(find: 'rdata')->(isa: 'array');
					#output += '(' + #temp->(find: 'rdata')->(join: ', ') + ')\r';
				else;
					#output += #temp->(find: 'rdata') + '\r';
				/if;
			/iterate;
		
			iterate: self->'nsarray', (local: 'temp');
				#output += 'NS ' + loop_count + ': ';
				#output += #temp->(find: 'qname') + ' ';
				#output += (if_empty: #dns_type_names->(find: #temp->(find: 'qtype')),  '(' + #temp->(find: 'qtype') + ')') + ' ';
				#output += (if_empty: #dns_class_names->(find: #temp->(find: 'qclass')), '(' + #temp->(find: 'qclass') + ')') + ' ';
				#output += #temp->(find: 'ttl') + ' ';
				#output += #temp->(find: 'rdata') + '\r';
			/iterate;
		
			iterate: self->'ararray', (local: 'temp');
				#output += 'AR ' + loop_count + ': ';
				#output += #temp->(find: 'qname') + ' ';
				#output += (if_empty: #dns_type_names->(find: #temp->(find: 'qtype')),  '(' + #temp->(find: 'qtype') + ')') + ' ';
				#output += (if_empty: #dns_class_names->(find: #temp->(find: 'qclass')), '(' + #temp->(find: 'qclass') + ')') + ' ';
				#output += #temp->(find: 'ttl') + ' ';
				#output += #temp->(find: 'rdata') + '\r';
			/iterate;
		
			#output += '</pre>';
			
			return: @#output;
			
		/define_tag;
	
		define_tag: 'data';
			return: self->'raw';
		/define_tag;
		
		define_tag: 'answer';
			(self->'parsed' == false) ? self->consume_message;
			if: self->'answer' === null;
				return: self->format;
			else;
				return: self->'answer';
			/if;
		/define_tag;
		
		define_tag: 'bitformat';
		
			local: 'input' = self->'raw';
			
			local: 'output' = '<pre>';
			local: 'size' = #input->size;
			
			local: 'string' = '';
			local: 'numeric' = '';
			iterate: #input, (local: 'integer');
				loop: 8;
					#output += (#integer->(bittest: 9 - loop_count) ? '1' | '0');
				/loop;
				#output += ' ';
				if: #integer >= 32 && #integer <= 127;
					#string += bytes->(import8bits: #integer) & + '  ';
				else: #integer == 9;
					#string += '\\t ';
				else: #integer == 10;
					#string += '\\n ';
				else: #integer == 11;
					#string += '\\v ';
				else: #integer == 12;
					#string += '\\f ';
				else: #integer == 13;
					#string += '\\r ';
				else;
					#string += '   ';
				/if;
				#numeric += #integer->(setformat: -padding=3) & + ' ';
				if: (#size == loop_count) && (#size % 4 != 0);
					#output += '         ' * (4 - #size % 4);
					#numeric += '    ' * (4 - #size % 4);
					#string += '   ' * (4 - #size % 4);
				/if;
				if: (loop_count % 4 == 0) || (#size == loop_count);
					#output += '| ' + #numeric + '| ' + #string + '|\n';
					local: 'string' = '';
					local: 'numeric' = '';
				/if;
			/iterate;
			
			#output += '</pre>';
			
			return: @#output;
		
		/define_tag;
		
		define_tag('onCompare', -required='value');
			(self->'parsed' == false) ? self->consume_message;
			if(#value->isa('dns_response'));	
				self->'id' != #value->id ? return(-1);
				self->'qdcount' != #value->qdcount ? return(-1);
				iterate(#value->qdarray, local('qd'));
					#qd >> 'qname' && #qd->find('qname') != self->'qdarray'->get(loop_count)->find('qname') ? return(-1);
					#qd >> 'qtype' && #qd->find('qtype') != self->'qdarray'->get(loop_count)->find('qtype') ? return(-1);
					#qd >> 'qclass' && #qd->find('qclass') != self->'qdarray'->get(loop_count)->find('qclass') ? return(-1);
				/iterate;
			/if;
			return(0);
		/define_tag;
		
		define_tag('id');
			(self->'parsed' == false) ? self->consume_message;
			return(self->'id');
		/define_tag;
		define_tag('qdcount');
			(self->'parsed' == false) ? self->consume_message;
			return(self->'qdcount');
		/define_tag;
		define_tag('qdarray');
			(self->'parsed' == false) ? self->consume_message;
			return(self->'qdarray');
		/define_tag;
				
	/define_type;

	define_tag: 'lookup', -namespace='dns_', -priority='replace',
			-required='name',
			-optional='type',
			-optional='class',
			-optional='norecurse',
			-optional='inverse',
			-optional='status',
			-optional='showquery',
			-optional='formatquery',
			-optional='bitquery',
			-optional='showresponse',
			-optional='format',
			-optional='bitformat',
			-optional='hostname',
			-optional='port',
			-optional='timeout';

		if: !(local_defined: 'hostname');
			local: 'hostname' = (global: '__domain_name_server__');
			if: #hostname == ''
				local: 'defaults' = dns_default;
				fail_if: #defaults->size == 0, -9965, 'Error - No DNS servers were found.  Please specify a -Hostname.';
				local: 'hostname' = #defaults->first;
			/if;
		/if;
		
		if: !(local_defined: 'port');
			local: 'port' = 53;
		/if;
		
		if: !(local_defined: 'timeout');
			local: 'timeout' = 15;
		/if;
		
		if: (local_defined: 'type');
			local: 'dns_type_codes' = (map:
				'a' = 1,
				'ns' = 2,
				'md' = 3,
				'mf' = 4,
				'cname' = 5,
				'soa' = 6,
				'mb' = 7,
				'mg' = 8,
				'mr' = 9,
				'null' = 10,
				'wks' = 11,
				'ptr' = 12,
				'hinfo' = 13,
				'minfo' = 14,
				'mx' = 15,
				'txt' = 16,
				'aaaa' = 28,
				'a6' = 38,
				'axfr' = 252,
				'mailb' = 253,
				'maila' = 254,
				'*' = 255);
			if: #dns_type_codes >> #type;
				local: 'qtype' = #dns_type_codes->(find: #type);
			else;
				fail: -9965, 'Type unrecognized for DNS lookup \"' + #type + '\".';
			/if;
		else;
			local: 'qtype' = 255;
		/if;
		
		if: (local_defined: 'class');
			local: 'dns_class_codes' = (map:
				'in' = 1,
				'cs' = 2,
				'ch' = 3,
				'*' = 255);
			if: #dns_class_codes >> #class;
				local: 'qclass' = #dns_class_codes->(find: #class);
			else;
				fail: -9965, 'Class unrecognized for DNS lookup \"' + #class + '\".';
			/if;
		else;
			local: 'qclass' = 1;
		/if;
	
		local: 'qname' = bytes;
		local: 'temp' = (string: #name);
		#temp->trim;
		local: 'array' = (string_findregexp: #temp, -find='\\d+');
		if: (#array->size == 4) && (#temp->(split: '.')->size == 4);
			local: 'temp' = #array->(get:4) + '.' + #array->(get:3) + '.' + #array->(get:2) + '.' + #array->(get:1) + '.in-addr.arpa';
			if: #qtype == 255;
				local: 'qtype' = 12; // PTR
			/if;
		/if;
		iterate: #temp->(split: '.'), (local: 'label');
			#qname->(import8bits: (math_min: #label->size, 63));
			#qname->(importstring: #label->(substring: 1,63), 'iso-8859-1');
		/iterate;
		#qname->(import8bits: 0);
			
		local: 'query' = bytes;
				
		local: 'id' = (math_random: -lower=1, -upper=65535);
		local: 'operation' = integer;
		if: (local_defined: 'inverse');
			#operation->(bitset: 12); // Inverse Query
		else: (local_defined: 'status');
			#operation->(bitset: 13); // Status
		/if;
		if: !(local_defined: 'norecurse');
			#operation->(bitset: 9); // RD Recursion Desired
		/if;
		local: 'qdcount' = 1;
		local: 'ancount' = 0;
		local: 'nscount' = 0;
		local: 'arcount' = 0;
		
		#query->(import16bits: #id->hosttonet16);
		#query->(import16bits: #operation->hosttonet16);
		#query->(import16bits: #qdcount->hosttonet16);
		#query->(import16bits: #ancount->hosttonet16);
		#query->(import16bits: #nscount->hosttonet16);
		#query->(import16bits: #arcount->hosttonet16);
		#query += #qname;
		#query->(import16bits: #qtype->hosttonet16);
		#query->(import16bits: #qclass->hosttonet16);

		local: 'query_bytes' = bytes;
		#query_bytes->(import16bits: #query->size->hosttonet16);
		#query_bytes += #query;
		
		local: 'request' = null;
		protect;
			#request = (dns_response: #query_bytes);
		/protect;
		fail_if: (#request === null), -9965, '[DNS_Lookup] failed with DNS request construction error.';
		if: (local_defined: 'showquery');
			return: #request;
		else: (local_defined: 'formatquery');
			return: #request->format;
		else: (local_defined: 'bitquery');
			return: #request->bitformat;
		/if;

		local: 'msg' = '[DNS_Lookup] failed with DNS server ' + #hostname + '.  Check the DNS server configuration in the Lasso Settings section of Site Administration.';
		
		local: 'net' = net;
		// Randomize ephemeral port
		protect;
			#net->bind(math_random(-lower=49152, -upper=65535)); 
		/protect;
		local: 'response_bytes' = '';
		local: 'result' = null;
		protect;
			local: 'result' = #net->(connect: #hostname, #port);
			handle_error;
				local: 'result' = null;
			/handle_error;
		/protect;
		fail_if: (#result != net_connectok), -9965, #msg + ' (Connect Error)';
		fail_if: (#net->(wait: (math_range: #timeout, -min=1, -max=60), net_waitwrite) != net_waitwrite), -9965, #msg + ' (Write Error)';
		local: 'result' = #net->(write: #query_bytes);
		fail_if: (#net->(wait: (math_range: #timeout, -min=1, -max=60), net_waitread) != net_waitread), -9965, #msg + ' (Read Error)';
		#response_bytes = #net->(read: 32768);
		#net->close;
		
		local: 'response' = null;
		protect;
			local: 'response' = (dns_response: #response_bytes);
		/protect;
		fail_if: (#response === null), -9965, #msg + ' (Parse Error)';

		fail_if: #response != #request, -9965, '[DNS_Lookup] failed due to ID or Query mismatch signallying possible DNS injection attempt.';

		if: (local_defined: 'showresponse');
			return: #response;
		else: (local_defined: 'format');
			return: #response->format;
		else: (local_defined: 'bitformat');
			return: #response->bitformat;
		else;
			return: #response->answer;
		/if;

	/define_tag;

?>